/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes

#include "CompareRankingDialog.h"

#include "shared/SharedStyles.h"
#include "shared/TableDelegate.h"
#include "shared/SharedObjects.h"
#include "shared/StringTools.h"
#include "shared/TournamentSettings.h"

// Qt includes
#include <QDebug>
#include <QApplication>
#include <QMessageBox>
#include <QGroupBox>
#include <QTableWidget>
#include <QHeaderView>
#include <QVBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QTimer>
#include <QScrollBar>
#include <QJsonObject>
#include <QJsonArray>

// C++ includes
#include <utility>

CompareRankingDialog::CompareRankingDialog(QWidget *parent, SharedObjects *sharedObjects)
    : TitleDialog(parent),
      m_db(sharedObjects->database()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_stringTools(sharedObjects->stringTools())
{
    setTitle(tr("Ranglistenvergleich"));

    QGroupBox *resultBox = new QGroupBox;
    resultBox->setProperty("notitle", true);
    QVBoxLayout *resultBoxLayout = new QVBoxLayout(resultBox);
    layout()->addWidget(resultBox);

    m_resultLabel = new QLabel;
    m_resultLabel->setWordWrap(true);
    resultBoxLayout->addWidget(m_resultLabel);

    m_differences = new QTableWidget;
    m_differences->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_differences->setItemDelegate(new TableDelegate(this));
    m_differences->verticalHeader()->hide();
    m_differences->setShowGrid(false);
    m_differences->setWordWrap(false);
    m_differences->setSelectionMode(QTableWidget::NoSelection);
    m_differences->setFocusPolicy(Qt::NoFocus);
    m_differences->setContextMenuPolicy(Qt::CustomContextMenu);
    resultBoxLayout->addWidget(m_differences);
    m_differences->hide();

    m_dataChangedLabel = new QLabel(tr(
        "<span style=\"%1\">Seit dem letzten Abruf hat sich die Rangliste entweder lokal oder auf "
        "dem Server geändert. Bitte die Rangliste erneut abgleichen!</span>").arg(
        SharedStyles::RED_TEXT));
    m_dataChangedLabel->setWordWrap(true);
    resultBoxLayout->addWidget(m_dataChangedLabel);
    m_dataChangedLabel->hide();

    // Button box

    addButtonBox();

    m_requestRanking = buttonBox()->addButton(QDialogButtonBox::Yes);
    m_requestRanking->setText(tr("Rangliste abfragen"));
    m_requestRanking->setFocus();
    connect(m_requestRanking, &QPushButton::clicked, this, &CompareRankingDialog::requestRanking);

    m_close = buttonBox()->addButton(QDialogButtonBox::Close);
    connect(m_close, &QPushButton::clicked, this, &QDialog::reject);

    if (checkTournamentStarted()) {
        // We have to do this via a QTimer::singleShot call, because at
        // construction time, we're not connected yet properly.
        QTimer::singleShot(0, this, &CompareRankingDialog::requestRanking);
    }
}

bool CompareRankingDialog::checkTournamentStarted()
{
    if (m_db->localTournamentStarted()) {
        m_requestRanking->setEnabled(true);
        return true;
    } else {
        m_requestRanking->setEnabled(false);
        m_differences->hide();
        m_resultLabel->setText(tr("Es wurden noch keine Ergebnisse eingegeben."));
        m_dataChangedLabel->hide();
        m_close->setFocus();
        resizeToContents();
        return false;
    }
}

void CompareRankingDialog::clientDisconnected()
{
    QApplication::restoreOverrideCursor();
    QMessageBox::critical(this, windowTitle(),
                          tr("Die Netzwerkverbindung wurde beendet. Der Dialog wird geschlossen."));
    QDialog::reject();
}

void CompareRankingDialog::requestRanking()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    m_requestRanking->setEnabled(false);
    m_close->setEnabled(false);
    emit rankingRequested();
}

void CompareRankingDialog::dataReceived()
{
    m_requestRanking->setEnabled(true);
    m_close->setEnabled(true);
    m_dataChangedLabel->hide();
}

void CompareRankingDialog::compareRanking(const QVector<Database::RankData> &remoteRanking)
{
    dataReceived();

    if (remoteRanking.count() == 0) {
        m_differences->hide();
        m_resultLabel->setText(tr("Auf dem Server wurden noch keine Ergebnisse eingegeben."));
        m_close->setFocus();
        resizeToContents();
        return;
    }

    const auto [ localRanking, success ] = m_db->ranking(-1);
    if (! success) {
        return;
    }

    QHash<QString, int> localIndex;
    for (int i = 0; i < localRanking.count(); i++) {
        localIndex[localRanking.at(i).name] = i;
    }

    // Check if the same number of results has been stored

    QHash<QString, QPair<int, int>> roundDifferences;
    for (const Database::RankData &rank : remoteRanking) {
        const int localPlayedBoogers = localRanking.at(localIndex.value(rank.name)).playedBoogers;
        if (rank.playedBoogers != localPlayedBoogers) {
            roundDifferences[rank.name] = QPair<int, int>(rank.playedBoogers, localPlayedBoogers);
        }
    }

    if (! roundDifferences.isEmpty()) {
        m_resultLabel->setText(tr(
            "<p style=\"%1\">Es wurden hier und auf dem Server nicht für alle %2 gleich viele "
            "Ergebnisse eingegeben!</p>"
            "<p>Bei folgenden %3 fehlen noch Spielstände (links stehen die eingegebenen Bobbl auf "
            "dem Server, rechts die hier eingegebenen:</p>").arg(
            SharedStyles::RED_TEXT,
            m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
            m_tournamentSettings->isPairMode() ? tr("Paaren") : tr("Spielern")));

        m_differences->show();
        m_differences->clear();
        m_differences->setColumnCount(2);
        m_differences->setHorizontalHeaderLabels( {
            m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
            tr("Eingetragene Bobbl") });
        m_differences->setRowCount(roundDifferences.count());

        QStringList sortedNames(roundDifferences.keys());
        m_stringTools->sort(sortedNames);
        int row = 0;
        for (const QString &name : std::as_const(sortedNames)) {
            m_differences->setItem(row, 0, new QTableWidgetItem(name));
            m_differences->setItem(row, 1, new QTableWidgetItem(differenceString(
                roundDifferences.value(name).first, roundDifferences.value(name).second)));
            row++;
        }

        finishDisplayDifferences();
        return;
    }

    // Check if the ranking is identical

    QHash<QString, QStringList> rankingDifferences;
    for (const Database::RankData &rank : remoteRanking) {
        const int localBoogers = localRanking.at(localIndex.value(rank.name)).boogers;
        const int localGoals = localRanking.at(localIndex.value(rank.name)).goals;
        const int localOpponentGoals = localRanking.at(localIndex.value(rank.name)).opponentGoals;

        if (rank.boogers != localBoogers
            || rank.goals != localGoals
            || rank.opponentGoals != localOpponentGoals) {

            rankingDifferences[rank.name] = QStringList {
                differenceString(rank.boogers, localBoogers),
                differenceString(rank.goals, localGoals),
                differenceString(rank.opponentGoals, localOpponentGoals) };
        }
    }

    if (! rankingDifferences.isEmpty()) {
        m_resultLabel->setText(tr(
            "<p style=\"%1\">Die lokale Rangliste ist nicht identisch mit der des Servers!</p>"
            "<p>Folgende %2 haben auf dem Server andere Werte (links stehen die Werte des Servers, "
            "rechts die hier eingegebenen):</p>").arg(
            SharedStyles::RED_TEXT,
            m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

        m_differences->show();
        m_differences->clear();
        m_differences->setColumnCount(4);
        m_differences->setHorizontalHeaderLabels( {
            m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
            tr("Bobbl"), tr("Tore"), tr("Gegn. Tore") });
        m_differences->setRowCount(rankingDifferences.count());

        QStringList sortedNames(rankingDifferences.keys());
        m_stringTools->sort(sortedNames);
        int row = 0;
        for (const QString &name : std::as_const(sortedNames)) {
            m_differences->setItem(row, 0,
                                   new QTableWidgetItem(name));
            m_differences->setItem(row, 1,
                                   new QTableWidgetItem(rankingDifferences.value(name).at(0)));
            m_differences->setItem(row, 2,
                                   new QTableWidgetItem(rankingDifferences.value(name).at(1)));
            m_differences->setItem(row, 3,
                                   new QTableWidgetItem(rankingDifferences.value(name).at(2)));
            row++;
        }

        finishDisplayDifferences();
        return;
    }

    // If we reach here, the rankings are identical

    m_resultLabel->setText(tr("<span style=\"%1\">Die lokale Rangliste ist identisch mit der des "
                              "Servers.</span>").arg(SharedStyles::GREEN_TEXT));
    m_differences->hide();
    resizeToContents();
    m_close->setFocus();
}

void CompareRankingDialog::finishDisplayDifferences()
{
    m_differences->resizeColumnsToContents();
    resizeToContents();
    resizeToTableSize();
    m_requestRanking->setFocus();
}

const QString CompareRankingDialog::differenceString(int local, int remote) const
{
    return local == remote ? QString::number(local)
                           : QStringLiteral("%1 \u2260 %3").arg(QString::number(local),
                                                                QString::number(remote));
}

void CompareRankingDialog::dataChanged()
{
    checkTournamentStarted();
    m_dataChangedLabel->setVisible(m_requestRanking->isEnabled());
    resizeToContents();
}

void CompareRankingDialog::resizeToTableSize()
{
    QApplication::processEvents();
    QTimer::singleShot(0, [this]
    {
        const auto *parentWidget = qobject_cast<QWidget *>(parent());
        const auto parentSize = parentWidget->size();

        // FIXME: If anybody ever may know how to get the correct scroll bar
        //        width/height _before_ it's actually shown, tell me :-P
        const int scrollBarWidth = std::min(m_differences->verticalScrollBar()->width(),
                                            m_differences->horizontalScrollBar()->height());

        // Step 1: Adjust the width so that all columns are shown if the parent's
        // width allows this (we don't become wider than the parent)

        const int remainingDialogWidth = width() - m_differences->width();
        const int maximumTableWidth = m_differences->horizontalHeader()->length()
                                      + m_differences->width() - m_differences->viewport()->width()
                                      + scrollBarWidth;

        resize(std::min(parentSize.width(), maximumTableWidth + remainingDialogWidth), height());

        // Step 2: Adjust the width so that all rows are shown if the parent's
        // height allows this (we don't become taller than the parent)

        const int remainingDialogHeight = height() - m_differences->height();

        int tableHeight = 0;
        for (int i = 0; i < m_differences->rowCount(); i++) {
            tableHeight += m_differences->rowHeight(i);
        }
        tableHeight += m_differences->height() - m_differences->viewport()->height();

        resize(width(), std::min(tableHeight + remainingDialogHeight, parentSize.height()));

        // Step 3: (Somehow) Center this dialog on the parent
        move(parentWidget->geometry().center() - rect().center());
    });
}
