/* Copyright (c) 2019-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes

#include "DrawOverviewPage.h"

#include "shared/SharedObjects.h"
#include "shared/Settings.h"
#include "shared/TournamentSettings.h"
#include "shared/TableDelegate.h"
#include "shared/SearchWidget.h"
#include "shared/DrawMapper.h"

// Qt includes
#include <QGroupBox>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QPushButton>
#include <QApplication>
#include <QStyle>
#include <QTableWidget>
#include <QHeaderView>
#include <QLabel>
#include <QDebug>

DrawOverviewPage::DrawOverviewPage(QWidget *parent, SharedObjects *sharedObjects,
                                   const QVector<QString> *allRegistrations,
                                   const QVector<Draw::Seat> *allDraws,
                                   const QVector<int> *incompleteTables,
                                   const QVector<int> *tableGaps,
                                   bool drawAdjustmentNeeded)
    : OptionalPage(parent),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_searchEngine(sharedObjects->searchEngine()),
      m_allRegistrations(allRegistrations),
      m_allDraws(allDraws),
      m_incompleteTables(incompleteTables),
      m_tableGaps(tableGaps)
{
    auto *box = new QGroupBox;
    box->setProperty("notitle", true);
    auto *boxLayout = new QHBoxLayout(box);
    layout()->addWidget(box);

    auto *displayLayout = new QVBoxLayout;
    boxLayout->addLayout(displayLayout);

    m_noDraw = new QLabel(tr("<i>Es wurde noch keine Auslosung eingegeben</i>"));
    m_noDraw->setAlignment(Qt::AlignCenter);
    displayLayout->addWidget(m_noDraw);

    m_isSearch = new QLabel;
    m_isSearch->hide();
    displayLayout->addWidget(m_isSearch);

    m_draw = new QTableWidget;
    m_draw->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_draw->setItemDelegate(new TableDelegate(this, TableDelegate::HighlightMode::NRows));
    m_draw->verticalHeader()->hide();
    m_draw->setShowGrid(false);
    m_draw->setWordWrap(false);
    m_draw->setSelectionMode(QTableWidget::NoSelection);
    m_draw->setFocusPolicy(Qt::NoFocus);
    m_draw->setColumnCount(3);
    m_draw->setHorizontalHeaderLabels({ tr("Tisch"), QString(), tr("Namen") });
    displayLayout->addWidget(m_draw);

    m_uncompletedTablesLabel = new QLabel;
    displayLayout->addWidget(m_uncompletedTablesLabel);

    m_adjustmentLabel = new QLabel(tr("Alle %1 haben eine Auslosung, aber die Vergabe ist nicht "
                                      "fortlaufend. <a href=\"#adjustDraw\">Auslosung bereinigen"
                                      "</a>").arg(
                                      m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                         : tr("Spieler")));
    m_adjustmentLabel->setWordWrap(true);
    m_adjustmentLabel->setTextInteractionFlags(Qt::TextBrowserInteraction);
    connect(m_adjustmentLabel, &QLabel::linkActivated,
            this, &DrawOverviewPage::showAdjustDrawDialog);
    m_adjustmentLabel->setVisible(drawAdjustmentNeeded);
    displayLayout->addWidget(m_adjustmentLabel);

    m_searchWidget = new SearchWidget(QStringLiteral("draw_page_search"), sharedObjects);
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &DrawOverviewPage::updateSearch);
    displayLayout->addWidget(m_searchWidget);

    auto *closeLayout = new QVBoxLayout;
    boxLayout->addLayout(closeLayout);

    auto *closeButton = new QPushButton(tr("Schließen"));
    closeButton->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    closeButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogCloseButton));
    connect(closeButton, &QPushButton::clicked, this, &DrawOverviewPage::closePage);
    closeLayout->addWidget(closeButton);
    closeLayout->addStretch();
}

bool DrawOverviewPage::closePage()
{
    if (m_tournamentSettings->isSinglePlayerMode()) {
        m_searchEngine->clearSearchCache(SearchEngine::SearchType::DrawnPairsDraw);
    }

    return OptionalPage::closePage();
}

void DrawOverviewPage::scrollTo(const QString &target)
{
    const auto matches = m_draw->findItems(target, Qt::MatchExactly);
    if (matches.isEmpty()) {
        // This should not happen
        return;
    }

    auto *item = matches.at(0);
    int row = item->row();
    if (! (row % 2)) {
        item = m_draw->item(++row, item->column());
        if (item == nullptr) {
            // This should not happen either
            return;
        }
    }

    const int lastRow = m_draw->rowCount() - 1;
    if (row == lastRow || row == lastRow -1) {
        m_draw->scrollToBottom();
    } else {
        m_draw->scrollToItem(item, QAbstractItemView::PositionAtCenter);
    }
}

void DrawOverviewPage::updateDraw(bool ownChange, const QString &target, bool deleted)
{
    if (ownChange) {
        if (target.isEmpty()) {
            m_draw->scrollToTop();
        } else if (deleted) {
            scrollTo(target);
        }
    }

    const auto draw = DrawMapper::mapDraw(*m_allRegistrations, *m_allDraws);
    const int rows = draw.count() * 2;
    m_searchText.resize(rows);
    QVector<bool> pairNotComplete;
    pairNotComplete.resize(rows);

    // Generate all rows we need for displaying

    int row = 0;

    if (m_tournamentSettings->isPairMode()) {
        // For fixed pairs, we simply split up the draw to two rows
        for (const auto &pairs : draw) {
            m_searchText[row] = pairs.at(0);
            m_searchText[++row] = pairs.at(2);
            row++;
        }

    } else {
        // For single players, we have to compile the pair names and update the search cache
        m_searchEngine->clearSearchCache(SearchEngine::SearchType::DrawnPairsDraw);
        const QString separator = m_settings->namesSeparator();
        pairNotComplete.resize(rows);
        for (const auto &pairs : draw) {
            for (int i = 0; i < 3; i += 2) {
                if (! pairs.at(0 + i).isEmpty() && pairs.at(1 + i).isEmpty()) {
                    pairNotComplete[row] = true;
                    m_searchText[row] = pairs.at(0 + i) + separator;
                } else if (! pairs.at(0 + i).isEmpty() && ! pairs.at(1 + i).isEmpty()) {
                    m_searchText[row] = pairs.at(0 + i) + separator + pairs.at(1 + i);
                }
                m_searchEngine->updateSearchCache(m_searchText[row],
                    SearchEngine::SearchType::DrawnPairsDraw);
                row++;
            }
        }
    }

    if (rows == 0) {
        m_noDraw->show();
        m_draw->hide();
        m_searchWidget->clear();
        m_searchWidget->hide();
        m_uncompletedTablesLabel->hide();
        m_adjustmentLabel->hide();
        return;
    } else {
        m_noDraw->hide();
        m_draw->show();
        m_draw->clearContents();
        m_draw->clearSpans();
        m_draw->setRowCount(rows);
        m_searchWidget->show();
    }

    m_draw->resizeColumnsToContents();

    // Display all tables
    for (int table = 1; table <= rows / 2; table++) {
        auto *item = new QTableWidgetItem(QString::number(table));
        const int row = (table - 1) * 2;
        m_draw->setItem(row, 0, item);
        m_draw->setSpan(row, 0, 2, 1);
    }

    // Display all pairs

    for (int row = 0; row < rows; row += 2) {
        if (m_searchText.at(row).isEmpty() && m_searchText.at(row + 1).isEmpty()) {
            auto *item = new QTableWidgetItem(tr("Noch keine Paare zugelost"));
            item->setForeground(Qt::red);
            m_draw->setItem(row, 2, item);
            m_draw->setSpan(row, 2, 2, 1);

        } else {
            m_draw->setItem(row,     1, new QTableWidgetItem(tr("Paar 1:")));
            m_draw->setItem(row + 1, 1, new QTableWidgetItem(tr("Paar 2:")));

            for (int i = 0; i < 2; i++) {
                auto *item = new QTableWidgetItem;
                if (pairNotComplete.at(row + i)) {
                    item->setText(m_searchText.at(row + i)
                                  + tr("(noch kein zweiter Spieler zugelost)"));
                    item->setForeground(Qt::red);
                } else {
                    const auto &text = m_searchText.at(row + i);
                    if (! text.isEmpty()) {
                        item->setText(text);
                    } else {
                        item->setText(tr("Fehlt noch"));
                        item->setForeground(Qt::red);
                    }
                }
                m_draw->setItem(row + i, 2, item);
            }
        }
    }

    m_draw->resizeColumnsToContents();

    QString text = QStringLiteral("<table>");

    const int incompleteTablesCount = m_incompleteTables->count();
    if (incompleteTablesCount > 0) {
        text.append(tr("<tr><td>Unvollständig besetzte Tische:</td><td>"));
        for (int i = 0; i < incompleteTablesCount; i++) {
            if (i > 0 && incompleteTablesCount > 1) {
                if (i < incompleteTablesCount - 1) {
                    text.append(tr(", "));
                } else {
                    text.append(tr(" und "));
                }
            }
            text.append(QString::number(m_incompleteTables->at(i)));
        }
        text.append(QStringLiteral("</td></tr>"));
    }

    const int tableGapsCount = m_tableGaps->count();
    if (m_tableGaps->count() > 0) {
        text.append(tr("<tr><td>Unbesetzte Tische:</td><td>"));
        for (int i = 0; i < m_tableGaps->count(); i++) {
            if (i > 0 && tableGapsCount > 1) {
                if (i < tableGapsCount - 1) {
                    text.append(tr(", "));
                } else {
                    text.append(tr(" und "));
                }
            }
            text.append(QString::number(m_tableGaps->at(i)));
        }
        text.append(QStringLiteral("</td></tr>"));
    }

    if (text == QStringLiteral("<table>")) {
        m_uncompletedTablesLabel->setVisible(false);
    } else {
        text.append(QStringLiteral("</table>"));
        m_uncompletedTablesLabel->setText(text);
        m_uncompletedTablesLabel->setVisible(true);
    }

    if (ownChange && (! target.isEmpty() && ! deleted)) {
        scrollTo(target);
    }
}

void DrawOverviewPage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_draw->rowCount(); row++) {
            m_draw->showRow(row);
        }
        m_draw->resizeColumnsToContents();
        m_isSearch->hide();
        return;
    }

    m_searchEngine->setSearchTerm(searchText);
    const SearchEngine::SearchType searchType
        = m_tournamentSettings->isPairMode() ? SearchEngine::SearchType::Default
                                             : SearchEngine::SearchType::DrawnPairsDraw;
    int matches = 0;
    const bool phoneticSearch = m_searchWidget->phoneticSearch();
    for (int row = 0; row < m_draw->rowCount(); row += 2) {
        const bool rowMatches =
               m_searchEngine->checkMatch(m_searchText.at(row    ), searchType, phoneticSearch)
            || m_searchEngine->checkMatch(m_searchText.at(row + 1), searchType, phoneticSearch);

        m_draw->setRowHidden(row,     ! rowMatches);
        m_draw->setRowHidden(row + 1, ! rowMatches);
        matches += rowMatches;
    }

    m_searchWidget->updateMatches(matches);
    m_draw->resizeColumnsToContents();
    m_isSearch->setText(tr("Suchergebnis für „%1“").arg(searchText.simplified()));
    m_isSearch->show();
}

void DrawOverviewPage::resetSearch()
{
    m_searchWidget->clear();
}

void DrawOverviewPage::drawAdjustmentNeeded(bool state)
{
    m_adjustmentLabel->setVisible(state);
}
