/* Copyright (c) 2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes

#include "BasicSettingsPage.h"

#include "shared/TournamentTemplate.h"
#include "shared/TournamentSettings.h"
#include "shared/DefaultValues.h"

// Qt includes
#include <QLabel>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QRadioButton>
#include <QSpinBox>

BasicSettingsPage::BasicSettingsPage(const TournamentTemplate::Settings &tournamentTemplate,
                                     TournamentSettings *tournamentSettings,
                                     QWidget *parent)
    : AbstractSettingsPage(tr("Grundeinstellungen"), parent),
      m_tournamentSettings(tournamentSettings)
{
    setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);

    QLabel *tournamentSettingsLabel = new QLabel(tr(
        "Hier werden die Grundparameter des Muckturniers festgelegt. Diese Einstellungen sind "
        "fix und können während des Turniers nicht mehr verändert werden."));
    tournamentSettingsLabel->setWordWrap(true);
    layout()->addWidget(tournamentSettingsLabel);

    // Tournament type

    QGroupBox *tournamentTypeBox = new QGroupBox(tr("Turniertyp"));
    QGridLayout *tournamentTypeBoxLayout = new QGridLayout(tournamentTypeBox);
    layout()->addWidget(tournamentTypeBox);

    m_fixedPairs = new QRadioButton(tr("Feste Paare"));
    m_fixedPairs->setChecked(tournamentTemplate.tournamentMode == Tournament::Mode::FixedPairs);
    tournamentTypeBoxLayout->addWidget(m_fixedPairs, 0, 0, Qt::AlignTop);

    tournamentTypeBoxLayout->addWidget(
        descriptionLabel(tr("Es werden immer zwei Spieler als festes Paar angemeldet, das das "
                            "ganze Turnier zusammen spielt.")),
        0, 1);

    m_singlePlayers = new QRadioButton(tr("Einzelne Spieler"));
    m_singlePlayers->setChecked(! m_fixedPairs->isChecked());
    tournamentTypeBoxLayout->addWidget(m_singlePlayers, 1, 0, Qt::AlignTop);
    connect(m_singlePlayers, &QRadioButton::toggled,
            this, &BasicSettingsPage::singlePlayersModeSelected);

    tournamentTypeBoxLayout->addWidget(
        descriptionLabel(tr("Es werden einzelne Spieler angemeldet.\n"
                            "Die Paare werden pro Runde ausgelost.")),
        1, 1);

    // Boogers and scores

    QGroupBox *boogersBox = new QGroupBox(tr("Bobbl und Punkte"));
    QHBoxLayout *boogerWrapperLayout = new QHBoxLayout(boogersBox);
    QGridLayout *boogerLayout = new QGridLayout;
    boogerWrapperLayout->addLayout(boogerLayout);
    layout()->addWidget(boogersBox);

    boogerLayout->addWidget(new QLabel(tr("Bobbl pro Runde:")), 1, 0);
    m_boogersPerRound = new QSpinBox;
    m_boogersPerRound->setMinimum(1);
    m_boogersPerRound->setMaximum(DefaultValues::MAXIMUM_BOOGERS_PER_ROUND);
    m_boogersPerRound->setValue(tournamentTemplate.boogersPerRound);
    boogerLayout->addWidget(m_boogersPerRound, 1, 1);

    boogerLayout->addWidget(new QLabel(tr("Punkte pro Bobbl:")), 0, 0);
    m_boogerScore = new QSpinBox;
    m_boogerScore->setMinimum(1);
    m_boogerScore->setMaximum(DefaultValues::MAXIMUM_BOOGER_SCORE);
    m_boogerScore->setValue(tournamentTemplate.boogerScore);
    boogerLayout->addWidget(m_boogerScore, 0, 1);

    boogerWrapperLayout->addStretch();

    layout()->addStretch();

    m_advancedOptionsLabel = new QLabel(tr("Für weitere Einstellungen „Erweiterte Optionen“ "
                                           "anhaken!"));
    layout()->addWidget(m_advancedOptionsLabel);
}

void BasicSettingsPage::extendedOptionsToggled(bool state)
{
    m_advancedOptionsLabel->setVisible(! state);
}

void BasicSettingsPage::saveSettings()
{
    m_tournamentSettings->setTournamentMode(m_fixedPairs->isChecked()
                                                ? Tournament::Mode::FixedPairs
                                                : Tournament::Mode::SinglePlayers);
    m_tournamentSettings->setBoogerScore(m_boogerScore->value());
    m_tournamentSettings->setBoogersPerRound(m_boogersPerRound->value());
}
