/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes
#include "MarkerDialog.h"

// Qt includes
#include <QGridLayout>
#include <QLineEdit>
#include <QLabel>
#include <QPushButton>
#include <QComboBox>
#include <QDebug>
#include <QColorDialog>
#include <QTableWidget>
#include <QMessageBox>
#include <QTimer>

// std includes
#include <math.h> // pow

MarkerDialog::MarkerDialog(QWidget *parent, const QColor &background, QTableWidget *markers)
    : TitleDialog(parent), m_background(background), m_markers(markers)
{
    addButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
    QGridLayout *settingsLayout = new QGridLayout;
    layout()->addLayout(settingsLayout);
    mainLayout()->setSizeConstraint(QLayout::SetFixedSize);

    settingsLayout->addWidget(new QLabel(tr("Name:")), 0, 0);
    m_name = new QLineEdit;
    m_name->setText(tr("Neue Markierung"));
    connect(m_name, &QLineEdit::textChanged,
            [this] { setOkButtonEnabled(! m_name->text().isEmpty()); });
    settingsLayout->addWidget(m_name, 0, 1);

    m_badContrast = new QLabel(QStringLiteral("<span style=\"color:red\">%1</span>").arg(
                                              tr("Vorsicht: Schlechter Kontrast!")));
    settingsLayout->addWidget(m_badContrast, 1, 1);
    m_badContrast->hide();

    setColor(QColor(Qt::red));

    settingsLayout->addWidget(new QLabel(tr("Farbe:")), 2, 0);
    QPushButton *chooseColorButton = new QPushButton(tr("..."));
    connect(chooseColorButton, &QPushButton::clicked, this, &MarkerDialog::chooseColor);
    settingsLayout->addWidget(chooseColorButton, 2, 1);

    settingsLayout->addWidget(new QLabel(tr("Sortierung:")), 3, 0);
    m_sorting = new QComboBox;
    m_sorting->addItem(tr("Neuer Block"), QVariant(Markers::MarkerSorting::NewBlock));
    m_sorting->addItem(tr("Fortlaufend"), QVariant(Markers::MarkerSorting::Merged));
    settingsLayout->addWidget(m_sorting, 3, 1);

    m_name->setFocus();
    QTimer::singleShot(0, m_name, &QLineEdit::deselect);
}

void MarkerDialog::chooseColor()
{
    // We have to mess with pointer here because if this dialog is closed due to a colliding
    // network change, the color dialog would persist, and if it's deleted and the color dialog
    // is an object and not a pointer, the program crashes. With the color dialog being a pointer,
    // everything is fine on deletion ;-)

    QColorDialog *dialog = new QColorDialog(m_name->palette().color(QPalette::Text), this);
    connect(dialog, &QColorDialog::colorSelected, this, &MarkerDialog::setColor);

    dialog->setWindowTitle(tr("Markierungsfarbe wählen"));
    dialog->exec();
}

void MarkerDialog::setName(const QString &text)
{
    m_name->setText(text);
}

void MarkerDialog::setColor(const QColor &color)
{
    QPalette palette;
    palette.setColor(QPalette::Text, color);
    m_name->setPalette(palette);

    if (contrastRatio(m_background, color) >= 2.3) {
        m_badContrast->hide();
    } else {
        m_badContrast->show();
    }
}

void MarkerDialog::setSorting(Markers::MarkerSorting sorting)
{
    m_sorting->setCurrentIndex(m_sorting->findData(sorting));
}

void MarkerDialog::disableSortingSelection()
{
    m_sorting->setEnabled(false);
}

qreal MarkerDialog::contrastRatio(const QColor &color1, const QColor &color2) const
{
    // This computes the contrast ratio of two colors according to WCAG 2.0
    // See https://www.w3.org/TR/2008/REC-WCAG20-20081211/#contrast-ratiodef

    qreal l1;
    qreal l2;
    if (color1.lightness() >= color2.lightness()) {
        l1 = relativeLuminance(color1);
        l2 = relativeLuminance(color2);
    } else {
        l1 = relativeLuminance(color2);
        l2 = relativeLuminance(color1);
    }

    return (l1 + 0.05) / (l2 + 0.05);
}

qreal MarkerDialog::relativeLuminance(const QColor &color) const
{
    return   0.2126 * prepareColorComponent(color.redF())
           + 0.7152 * prepareColorComponent(color.greenF())
           + 0.0722 * prepareColorComponent(color.blueF());
}

qreal MarkerDialog::prepareColorComponent(qreal component) const
{
    if (component <= 0.03928) {
        return component / 12.92;
    } else {
        return pow(((component + 0.055) / 1.055), 2.4);
    }
}

void MarkerDialog::setEditedName(const QString &name)
{
    m_editedName = name;
}

void MarkerDialog::accept()
{
    const QString name = m_name->text().simplified();
    const QString unmarked = tr("(unmarkiert)");
    if (name == unmarked) {
        if (QMessageBox::warning(this,
                windowTitle(),
                tr("<p><b>Uneindeutiger Name für die Markierung</b></p>"
                   "<p>Die Bezeichnung „%1“ ist für unmarkierte Einträge reserviert! "
                   "Soll dieser Name wirklich nochmals verwendet werden?</p>").arg(unmarked),
                QMessageBox::Yes | QMessageBox::Cancel,
                QMessageBox::Cancel) == QMessageBox::Cancel) {

            return;
        }
    }

    for (int i = 0; i < m_markers->rowCount(); i++) {
        const QString &listName = m_markers->item(i, 1)->text();
        if (listName == unmarked || (! m_editedName.isEmpty() && listName == m_editedName)) {
            continue;
        }
        if (listName.compare(name, Qt::CaseInsensitive) == 0) {
            if (QMessageBox::warning(this,
                    windowTitle(),
                    tr("<p><b>Uneindeutiger Name für die Markierung</b></p>"
                       "<p>Eine Markierung hat bereits den Namen „%1“! "
                       "Soll dieser Name wirklich nochmals verwendet werden?</p>").arg(listName),
                    QMessageBox::Yes | QMessageBox::Cancel,
                    QMessageBox::Cancel) == QMessageBox::Cancel) {

                return;
            }
            break;
        }
    }

    emit markerSaved(name, m_name->palette().color(QPalette::Text),
                     static_cast<Markers::MarkerSorting>(m_sorting->currentData().toInt()));

    QDialog::accept();
}

void MarkerDialog::setIsEditing()
{
    m_isEditing = true;
}

bool MarkerDialog::isEditing() const
{
    return m_isEditing;
}
