/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes

#include "MarkersWidget.h"
#include "MarkerDialog.h"

#include "ServerPage/Server.h"
#include "ClientPage/Client.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/Settings.h"
#include "shared/TournamentSettings.h"
#include "shared/ResourceFinder.h"
#include "shared/TableDelegate.h"
#include "shared/TitleMenu.h"
#include "shared/ChecksumHelper.h"
#include "shared/IconSizeEngine.h"
#include "shared/SharedStyles.h"
#include "shared/MinimumViewportWidget.h"
#include "shared/MinimumViewportScroll.h"

// Qt includes
#include <QVBoxLayout>
#include <QTableWidget>
#include <QDebug>
#include <QHeaderView>
#include <QGroupBox>
#include <QApplication>
#include <QMessageBox>
#include <QCheckBox>
#include <QComboBox>
#include <QTimer>
#include <QLabel>
#include <QLineEdit>
#include <QSplitter>

// C++ includes
#include <functional>
#include <utility>

namespace
{

const QVector<Markers::Type> ALL_MARKER_TYPES = {
    Markers::Default,
    Markers::Drawn,
    Markers::Absent,
    Markers::Single,
    Markers::Assigned
};

}

MarkersWidget::MarkersWidget(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings())
{
    auto *mainLayout = new QVBoxLayout(this);
    // This eliminates additional space around this widget when displayed inside the players page
    mainLayout->setContentsMargins(0, 0, 0, 0);

    // Markers list
    // ============

    auto *markersBox = new QGroupBox(tr("Markierungen"));
    auto *markersBoxLayout = new QVBoxLayout(markersBox);

    m_markers = new QTableWidget;
    m_markers->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_markers->setSelectionMode(QAbstractItemView::NoSelection);
    m_markers->setFocusPolicy(Qt::NoFocus);
    m_markers->setIconSize(sharedObjects->iconSizeEngine()->relativeSize(1.35));
    m_markers->setColumnCount(4);
    m_markers->setHorizontalHeaderLabels({ tr("n", "Anzahl"), tr("Name"), tr("Sortierung"),
                                           QString() });
    m_markers->verticalHeader()->hide();
    m_markers->setShowGrid(false);
    m_markers->setItemDelegate(new TableDelegate(this));
    m_markers->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_markers, &QWidget::customContextMenuRequested, this, &MarkersWidget::showContextMenu);
    connect(m_markers, &QTableWidget::cellDoubleClicked,
            this, &MarkersWidget::checkShowContextMenu);
    connect(m_markers, &QTableWidget::cellClicked, this, &MarkersWidget::changeVisibility);
    markersBoxLayout->addWidget(m_markers);

    // Settings part
    // =============

    m_markerSettingsWidget = new MinimumViewportWidget;
    auto *settingsLayout = new QVBoxLayout(m_markerSettingsWidget);
    settingsLayout->setContentsMargins(0, 0, 0, 0);
    m_markerSettingsWidget->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Maximum);

    // Automatic markers

    m_automaticMarkersBox = new QGroupBox(tr("Automatische Markierungen"));
    m_automaticMarkersBox->setStyleSheet(SharedStyles::GROUPBOX_NORMAL_TITLE);
    settingsLayout->addWidget(m_automaticMarkersBox);
    auto *automaticMarkersLayout = new QGridLayout(m_automaticMarkersBox);
    m_automaticMarkersBox->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Maximum);

    m_defaultMarkerLabel = new QLabel(tr("Neue Anmeldungen\nmarkieren als:"));
    m_defaultMarkerLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    automaticMarkersLayout->addWidget(m_defaultMarkerLabel, 0, 0);
    auto *defaultMarkerCombo = new QComboBox;
    m_markersComboBoxes.insert(Markers::Default, defaultMarkerCombo);
    defaultMarkerCombo->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    connect(defaultMarkerCombo, QOverload<int>::of(&QComboBox::currentIndexChanged),
            [this]
            {
                m_tournamentSettings->setMarker(
                    Markers::Default,
                    m_markersComboBoxes.value(Markers::Default)->currentData().toInt());
            });
    automaticMarkersLayout->addWidget(defaultMarkerCombo, 0, 1);

    auto *enableDrawnMarker = new QCheckBox(tr("Nach Auslosung\nmarkieren als:"));
    m_enableMarkerCheckBoxes[Markers::Drawn] = enableDrawnMarker;
    automaticMarkersLayout->addWidget(enableDrawnMarker, 1, 0);
    connect(enableDrawnMarker, &QCheckBox::toggled, this, &MarkersWidget::toggleDrawnMarker);

    auto *drawnMarker = new QComboBox;
    m_markersComboBoxes.insert(Markers::Drawn, drawnMarker);
    drawnMarker->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    connect(drawnMarker, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &MarkersWidget::drawnMarkerChanged);

    automaticMarkersLayout->addWidget(drawnMarker, 1, 1);
    drawnMarker->setEnabled(false);

    // Ignored registrations marker

    m_absentMarkerBox = new QGroupBox(tr("Abwesende Anmeldungen"));
    m_absentMarkerBox->setStyleSheet(SharedStyles::GROUPBOX_NORMAL_TITLE);
    settingsLayout->addWidget(m_absentMarkerBox);
    auto *absentMarkerBoxLayout = new QHBoxLayout(m_absentMarkerBox);
    m_absentMarkerBox->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Maximum);

    auto *enableAbsentMarker = new QCheckBox(tr("Anmeldungen mit folgender\n"
                                                "Markierung ignorieren:"));
    m_enableMarkerCheckBoxes[Markers::Absent] = enableAbsentMarker;
    connect(enableAbsentMarker, &QCheckBox::toggled,
            this, &MarkersWidget::toggleAbsentMarker);
    absentMarkerBoxLayout->addWidget(enableAbsentMarker);

    auto *absentMarker = new QComboBox;
    m_markersComboBoxes.insert(Markers::Absent, absentMarker);
    absentMarker->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    absentMarker->setEnabled(false);
    connect(absentMarker, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &MarkersWidget::absentMarkerChanged);
    absentMarkerBoxLayout->addWidget(absentMarker);

    // Singles management

    m_singlesManagementBox = new QGroupBox(tr("Allein gekommene Spieler berücksichtigen"));
    m_singlesManagementBox->setCheckable(true);
    m_singlesManagementBox->setChecked(false);
    settingsLayout->addWidget(m_singlesManagementBox);
    m_singlesManagementBox->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Maximum);

    auto *singlesManagementLayout = new QGridLayout(m_singlesManagementBox);
    connect(m_singlesManagementBox, &QGroupBox::toggled,
            this, &MarkersWidget::singlesManagementChanged);

    auto *singlesMarkerLabel = new QLabel(tr("Markierung für „Allein da“:"));
    singlesManagementLayout->addWidget(singlesMarkerLabel, 0, 0);
    auto *singleMarkerCombo = new QComboBox;
    m_markersComboBoxes.insert(Markers::Single, singleMarkerCombo);
    connect(singleMarkerCombo, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &MarkersWidget::singlesManagementChanged);
    singleMarkerCombo->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    singlesManagementLayout->addWidget(singleMarkerCombo, 0, 1, 1, 2);

    m_markSingles = new QCheckBox(tr("automatisch setzen\nwenn der Name"));
    connect(m_markSingles, &QCheckBox::toggled,
            this, &MarkersWidget::singlesManagementChanged);
    singlesManagementLayout->addWidget(m_markSingles, 1, 0);
    m_markerSettingsWidget->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);


    m_singleConditionString = new QLineEdit;
    m_singleConditionString->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    m_singleConditionString->setMinimumWidth(50);
    connect(m_singleConditionString, &QLineEdit::editingFinished,
            this, &MarkersWidget::singlesManagementChanged);
    singlesManagementLayout->addWidget(m_singleConditionString, 1, 1);

    m_singleCondition = new QComboBox;
    m_singleCondition->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_singleCondition->addItem(tr("nicht enthält"));
    m_singleCondition->addItem(tr("enthält"));
    connect(m_singleCondition, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &MarkersWidget::singlesManagementChanged);
    singlesManagementLayout->addWidget(m_singleCondition, 1, 2);

    auto *assignedMarkerlabel = new QLabel(tr("Markierung nach Zuordnung:"));
    singlesManagementLayout->addWidget(assignedMarkerlabel, 2, 0);
    auto *assignedMarkerCombo = new QComboBox;
    m_markersComboBoxes.insert(Markers::Assigned, assignedMarkerCombo);
    assignedMarkerCombo->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    connect(assignedMarkerCombo, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &MarkersWidget::singlesManagementChanged);
    singlesManagementLayout->addWidget(assignedMarkerCombo, 2, 1, 1, 2);

    // Splitter arrangement
    // --------------------

    auto *splitter = new QSplitter(Qt::Vertical);
    splitter->setHandleWidth(2);
    splitter->setStyleSheet(SharedStyles::SPLITTER_VISIBLE_VERTICAL);

    splitter->addWidget(markersBox);
    splitter->setCollapsible(0, false);
    splitter->setStretchFactor(0, 4);

    m_settingsScroll = new MinimumViewportScroll(Qt::Horizontal);
    m_settingsScroll->setFrameShape(QFrame::NoFrame);
    m_settingsScroll->setWidgetResizable(true);

    m_markerSettingsWidget->setMinimumViewport(m_markerSettingsWidget->sizeHint());
    m_settingsScroll->setMaximumHeight(m_markerSettingsWidget->sizeHint().height());

    m_settingsScroll->setWidget(m_markerSettingsWidget);

    splitter->addWidget(m_settingsScroll);
    splitter->setCollapsible(1, false);
    splitter->setStretchFactor(1, 5);

    mainLayout->addWidget(splitter);

    // Context menu
    // ============

    m_markerMenu = new TitleMenu(this);

    m_moveUpAction = m_markerMenu->addAction(tr("Nach oben bewegen"));
    connect(m_moveUpAction, &QAction::triggered,
            this, std::bind(&MarkersWidget::moveMarker, this, -1));

    m_moveDownAction = m_markerMenu->addAction(tr("Nach unten bewegen"));
    connect(m_moveDownAction, &QAction::triggered,
            this, std::bind(&MarkersWidget::moveMarker, this, 1));

    m_markerMenu->addSeparator();

    m_editMarkerAction = m_markerMenu->addAction(tr("Bearbeiten"));
    connect(m_editMarkerAction, &QAction::triggered, this, &MarkersWidget::editMarker);

    m_deleteMarkerAction = m_markerMenu->addAction(tr("Löschen"));
    connect(m_deleteMarkerAction, &QAction::triggered, this, &MarkersWidget::deleteMarker);

    m_markerMenu->addSeparator();

    m_addMarkerAction = m_markerMenu->addAction(tr("Neue Markierung"));
    connect(m_addMarkerAction, &QAction::triggered, this, &MarkersWidget::addMarker);

    m_addMarkerMenu = new QMenu(this);
    m_addMarkerMenu->addAction(m_addMarkerAction);

    // Hide/Show context menu

    m_hideShowMarkedMenu = new QMenu(this);

    m_showOnlyAction = m_hideShowMarkedMenu->addAction(QString());

    m_showAllAction = m_hideShowMarkedMenu->addAction(tr("Alle anzeigen"));
    connect(m_showAllAction, &QAction::triggered, this, &MarkersWidget::showAll);

    // Icons
    m_unused = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("unused.png")));
    m_visible = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("visible.png")));
    m_hidden = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("hidden.png")));
}

int MarkersWidget::markersComboBoxSelection(QComboBox *comboBox) const
{
    const QVariant data = comboBox->currentData();
    return data.isValid() ? data.toInt() : -1;
}

void MarkersWidget::setMarkersComboBoxSelection(QComboBox *comboBox, int markerId)
{
    const int index = comboBox->findData(markerId);
    comboBox->setCurrentIndex(index != -1 ? index : comboBox->findData(0));
}

void MarkersWidget::updateTournamentMode()
{
    if (m_tournamentSettings->isSinglePlayerMode()) {
        m_markSingles->setChecked(false);
        m_singlesManagementBox->setVisible(false);
    } else {
        m_singlesManagementBox->setVisible(true);
    }

    m_markerSettingsWidget->setMinimumViewport(m_markerSettingsWidget->sizeHint());
    m_settingsScroll->setMaximumHeight(m_markerSettingsWidget->sizeHint().height());
}

void MarkersWidget::reload()
{
    const auto [ markers, success ] = m_db->markers();
    if (! success) {
        return;
    }

    QHash<Markers::Type, int> lastMarker;
    for (const auto comboId : ALL_MARKER_TYPES) {
        const auto &comboBox = m_markersComboBoxes.value(comboId);
        lastMarker.insert(comboId, markersComboBoxSelection(comboBox));
        comboBox->blockSignals(true);
        comboBox->clear();
    }

    m_markers->setRowCount(markers.count());

    QVector<int> ids;
    ids.reserve(markers.count());
    int row = 0;
    for (const Markers::Marker &marker : markers) {
        ids.append(marker.id);
        m_markers->setItem(row, 0, new QTableWidgetItem);

        QTableWidgetItem *item = new QTableWidgetItem(marker.name);
        if (marker.name.isEmpty()) {
            item->setText(tr("(unmarkiert)"));
        }
        item->setForeground(marker.color);
        item->setData(MarkerData::Id, QVariant(marker.id));
        item->setData(MarkerData::Sorting, QVariant(marker.sorting));
        m_markers->setItem(row, 1, item);

        const auto text = item->text();
        for (const auto comboId : ALL_MARKER_TYPES) {
            m_markersComboBoxes.value(comboId)->addItem(text, marker.id);
        }

        switch (marker.sorting) {
        case Markers::MarkerSorting::NewBlock:
            item = new QTableWidgetItem(tr("Neuer Block"));
            break;
        case Markers::MarkerSorting::Merged:
            item = new QTableWidgetItem(tr("Fortlaufend"));
            break;
        }
        m_markers->setItem(row, 2, item);

        item = new QTableWidgetItem;
        m_markers->setItem(row, 3, item);

        row++;
    }

    // Set new markers visible
    for (const int id : std::as_const(ids)) {
        if (! m_isVisible.contains(id)) {
            m_isVisible[id] = true;
        }
    }

    // Remove visibility entries for deleted markers
    const auto isVisibleKeys = m_isVisible.keys();
    for (const int id : isVisibleKeys) {
        if (! ids.contains(id)) {
            m_isVisible.remove(id);
        }
    }

    m_markers->horizontalHeaderItem(0)->setWhatsThis(tr(
        "In dieser Spalte wird die Anzahl der mit der entsprechenden Markierung gekennzeichneten "
        "%1 angezeigt.").arg(m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_markers->horizontalHeaderItem(3)->setWhatsThis(tr(
        "Durch Klicken auf das Augensymbol können alle mit der entsprechenden Markierung "
        "gekennzeichneten %1 ein- bzw. ausgeblendet werden.").arg(
        m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_markers->resizeColumnsToContents();

    for (const auto comboId : ALL_MARKER_TYPES) {
        const auto &comboBox = m_markersComboBoxes.value(comboId);
        setMarkersComboBoxSelection(comboBox, lastMarker.value(comboId));
        comboBox->blockSignals(false);
    }

    emit markersChanged();
}

void MarkersWidget::showContextMenu(const QPoint &position)
{
    int column = -1;
    QString name;
    int id;
    int row;
    const QPoint target = m_markers->viewport()->mapToGlobal(position);

    // Collect data
    QTableWidgetItem *markerItem = m_markers->itemAt(position);
    if (markerItem != nullptr) {
        column = markerItem->column();
        row = markerItem->row();
        if (column != 1) {
            markerItem = m_markers->item(row, 1);
        }
        name = m_markers->item(row, 1)->text();
        id = m_markers->item(row, 1)->data(MarkerData::Id).toInt();
    }

    // Show no context menu on the counter column
    if (column == 0) {
        return;
    }

    // Show the "Hide/Show marked entries" context menu on the "Hide/Show" column
    if (column == 3) {
        m_showAllAction->setEnabled(false);
        const auto isVisibleValues = m_isVisible.values();
        for (const bool visible : isVisibleValues) {
            if (! visible) {
                m_showAllAction->setEnabled(true);
                break;
            }
        }

        // Set the context menu to handle the selected marker
        m_showOnlyAction->disconnect();
        m_showOnlyAction->setText(tr("Nur „%1“ anzeigen").arg(name));
        connect(m_showOnlyAction, &QAction::triggered,
                this, std::bind(&MarkersWidget::showOnly, this, id));

        m_hideShowMarkedMenu->exec(target);
        return;
    }

    // All other actions require write access, so return here if the database if ro
    if (! m_db->isChangeable()) {
        return;
    }

    // Show the "Add marker" context menu if no item was clicked
    if (column == -1) {
        m_addMarkerMenu->exec(target);
        return;
    }

    // Otherwise, show the marker context menu

    m_markerMenu->setTitle(tr("Markierung „%1“").arg(name));
    m_selectedMarker.id = id;
    m_selectedMarker.name = name;
    m_selectedMarker.color = markerItem->foreground().color();
    m_selectedMarker.sorting = static_cast<Markers::MarkerSorting>(
                                   markerItem->data(MarkerData::Sorting).toInt());
    m_selectedMarkerSequence = markerItem->row();
    m_moveUpAction->setEnabled(m_selectedMarkerSequence > 0);
    m_moveDownAction->setEnabled(m_selectedMarkerSequence < m_markers->rowCount() - 1);
    m_editMarkerAction->setEnabled(m_selectedMarker.id != 0);
    m_deleteMarkerAction->setEnabled(m_selectedMarker.id != 0);

    m_markerMenu->exec(target);
}

void MarkersWidget::addMarker()
{
    MarkerDialog *dialog = new MarkerDialog(this, m_markers->palette().color(QPalette::Base),
                                            m_markers);
    connect(dialog, &MarkerDialog::markerSaved, this, &MarkersWidget::saveNewMarker);

    QString name = tr("Neue Markierung");
    int counter = 1;
    QString newName = name;
    while (m_markers->findItems(newName, Qt::MatchFixedString).size() > 0) {
        counter++;
        newName = QStringLiteral("%1 %2").arg(name, QString::number(counter));
    }

    dialog->setName(newName);
    dialog->setTitle(tr("Neue Markierung"));
    dialog->setButtonText(QDialogButtonBox::Ok, tr("Hinzufügen"));

    dialog->exec();
}

void MarkersWidget::saveNewMarker(const QString &name, const QColor &color,
                                  Markers::MarkerSorting sorting)
{
    if (m_isClient) {
        m_client->requestAddMarker(name, color, sorting);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    emit statusUpdate(tr("Lege neue Markierung an …"));

    if (! m_db->addMarker(name, color, sorting)) {
        return;
    }

    reload();

    if (m_isServer) {
        m_server->broadcastAddMarker(name, color, sorting);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ hinzugefügt").arg(name));
}

void MarkersWidget::deleteMarker()
{
    if (QMessageBox::question(this,
            tr("Markierung löschen"),
            tr("Soll die Markierung „%1“ wirklich gelöscht werden?").arg(m_selectedMarker.name),
            QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel) == QMessageBox::Cancel) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteMarker(m_selectedMarker.name, m_selectedMarker.id,
                                      m_selectedMarkerSequence);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    emit statusUpdate(tr("Lösche Markierung „%1“ …").arg(m_selectedMarker.name));

    if (! m_db->deleteMarker(m_selectedMarker.id, m_selectedMarkerSequence)) {
        return;
    }

    reload();

    if (m_isServer) {
        m_server->broadcastDeleteMarker(m_selectedMarker.name, m_selectedMarker.id,
                                        m_selectedMarkerSequence);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ gelöscht").arg(m_selectedMarker.name));
}

void MarkersWidget::editMarker()
{
    MarkerDialog *dialog = new MarkerDialog(this, m_markers->palette().color(QPalette::Base),
                                            m_markers);
    connect(dialog, &MarkerDialog::markerSaved, this, &MarkersWidget::saveEdit);

    dialog->setTitle(tr("Markierung bearbeiten"));
    dialog->setButtonText(QDialogButtonBox::Ok, tr("Speichern"));
    dialog->setName(m_selectedMarker.name);
    dialog->setEditedName(m_selectedMarker.name);
    dialog->setColor(m_selectedMarker.color);
    dialog->setSorting(m_selectedMarker.sorting);
    if (m_selectedMarkerSequence == 0) {
        dialog->disableSortingSelection();
    }
    dialog->setIsEditing();

    dialog->exec();
}

void MarkersWidget::saveEdit(const QString &name, const QColor &color,
                             Markers::MarkerSorting sorting)
{
    if (name == m_selectedMarker.name
        && color == m_selectedMarker.color
        && sorting == m_selectedMarker.sorting) {

        emit statusUpdate(tr("Markierung „%1“ wurde nicht verändert").arg(m_selectedMarker.name));
        return;
    }

    if (m_isClient) {
        m_client->requestEditMarker(m_selectedMarker.name, m_selectedMarker.id,
                                    name, color, sorting);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Speichere Änderungen an Markierung „%1“ …").arg(m_selectedMarker.name));

    if (! m_db->editMarker(m_selectedMarker.id, name, color, sorting)) {
        return;
    }

    reload();

    if (m_isServer) {
        m_server->broadcastEditMarker(m_selectedMarker.name, m_selectedMarker.id,
                                      name, color, sorting);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Änderungen an Markierung „%1“ gespeichert").arg(m_selectedMarker.name));
}

void MarkersWidget::moveMarker(int direction)
{
    if (m_isClient) {
        m_client->requestMoveMarker(m_selectedMarker.name, m_selectedMarker.id,
                                    m_selectedMarkerSequence, direction);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Verschiebe Markierung „%1“ %2 …").arg(
                         m_selectedMarker.name, direction == 1 ? tr("nach unten")
                                                               : tr("nach oben")));

    if (! m_db->moveMarker(m_selectedMarker.id, m_selectedMarkerSequence, direction)) {
        return;
    }

    reload();

    if (m_isServer) {
        m_server->broadcastMoveMarker(m_selectedMarker.name, m_selectedMarker.id,
                                      m_selectedMarkerSequence, direction);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ %2 verschoben").arg(
                         m_selectedMarker.name, direction == 1 ? tr("nach unten")
                                                               : tr("nach oben")));
}

QVector<Markers::Marker> MarkersWidget::getMarkers() const
{
    QVector<Markers::Marker> markers;
    markers.reserve(m_markers->rowCount());
    for (int row = 0; row < m_markers->rowCount(); row++) {
        Markers::Marker marker;
        QTableWidgetItem *item = m_markers->item(row, 1);
        marker.id = item->data(MarkerData::Id).toInt();
        marker.name = item->text();
        marker.color = item->foreground().color();
        marker.sorting = static_cast<Markers::MarkerSorting>(
                             item->data(MarkerData::Sorting).toInt());
        markers.append(marker);
    }
    return markers;
}

void MarkersWidget::checkShowContextMenu(int, int column)
{
    if (column == 1 || column == 2) {
        showContextMenu(m_markers->viewport()->mapFromGlobal(QCursor::pos()));
    }
}

void MarkersWidget::updateVisibilityIcon(int id, int row)
{
    if (! m_isVisible[id]) {
        m_markers->item(row, 3)->setIcon(m_hidden);
    } else {
        if (m_markers->item(row, 0)->text().toInt() == 0) {
            m_markers->item(row, 3)->setIcon(m_unused);
        } else {
            m_markers->item(row, 3)->setIcon(m_visible);
        }
    }
}

void MarkersWidget::changeVisibility(int row, int column)
{
    if (column != 3) {
        // Not clicked on eye symbol
        return;
    }

    const int id = m_markers->item(row, 1)->data(MarkerData::Id).toInt();
    m_isVisible[id] = ! m_isVisible[id];
    updateVisibilityIcon(id, row);

    emit visibilityChanged();
}

void MarkersWidget::updateVisibilityIcons()
{
    for (int row = 0; row < m_markers->rowCount(); row++) {
        updateVisibilityIcon(m_markers->item(row, 1)->data(MarkerData::Id).toInt(), row);
    }
    emit visibilityChanged();
}

void MarkersWidget::reset()
{
    const auto isVisibleKeys = m_isVisible.keys();
    for (const int id : isVisibleKeys) {
        m_isVisible[id] = true;
    }

    setCheckableMarker(Markers::Absent, m_tournamentSettings->marker(Markers::Absent));
    setCheckableMarker(Markers::Drawn, m_tournamentSettings->marker(Markers::Drawn));

    setSinglesManagement(m_tournamentSettings->singlesManagement());
    setMarkersComboBoxSelection(m_markersComboBoxes.value(Markers::Default),
                                m_tournamentSettings->marker(Markers::Default));

    m_absentMarkerBox->setEnabled(m_db->isChangeable());
    m_singlesManagementBox->setEnabled(m_db->isChangeable());
    m_automaticMarkersBox->setEnabled(m_db->isChangeable());
}

bool MarkersWidget::isMarkerVisible(int id) const
{
    return m_isVisible.value(id);
}

void MarkersWidget::updateCounters(const QHash<int, int> &counter)
{
    m_allMarkersCount = 0;
    int markerCount;
    int id;

    for (int row = 0; row < m_markers->rowCount(); row++) {
        id = m_markers->item(row, 1)->data(MarkerData::Id).toInt();
        markerCount = counter[id];
        m_markers->item(row, 0)->setText(QString::number(markerCount));
        m_allMarkersCount += markerCount;
    }

    updateVisibilityIcons();
    m_markers->resizeColumnsToContents();
}

void MarkersWidget::showAll()
{
    const auto isVisibleKeys = m_isVisible.keys();
    for (const int id : isVisibleKeys) {
        m_isVisible[id] = true;
    }
    updateVisibilityIcons();
}

void MarkersWidget::showOnly(int idToShow)
{
    const auto isVisibleKeys = m_isVisible.keys();
    for (const int id : isVisibleKeys) {
        m_isVisible[id] = (id == idToShow);
    }
    updateVisibilityIcons();
}

int MarkersWidget::newRegistrationMarker(const QString &name) const
{
    if (m_markSingles->isEnabled() && m_markSingles->isChecked()) {
        // The first m_singleCondition entry is "doesn't contain" (index 0), the second is
        // "contains" (index 1), so we can simply use the selected index here, because it will be
        // (per official C++ behavior) silently casted to a bool value in the following comparison

        if (name.contains(m_tournamentSettings->singlesManagement().conditionString)
            == m_singleCondition->currentIndex()) {

            return m_markersComboBoxes.value(Markers::Single)->currentData().toInt();
        }
    }

    return m_markersComboBoxes.value(Markers::Default)->currentData().toInt();
}

void MarkersWidget::toggleDrawnMarker(bool state)
{
    m_markersComboBoxes.value(Markers::Drawn)->setEnabled(state);
    drawnMarkerChanged();
}

void MarkersWidget::drawnMarkerChanged()
{
    m_tournamentSettings->setMarker(Markers::Drawn, drawnMarker());
}

int MarkersWidget::drawnMarker() const
{
    const auto &combo = m_markersComboBoxes.value(Markers::Drawn);
    return combo->isEnabled() ? combo->currentData().toInt() : -1;
}

std::pair<int, QString> MarkersWidget::markerAfterEdit(const QString &oldName,
                                                       const QString &name) const
{
    if (! m_markSingles->isEnabled() || ! m_markSingles->isChecked()) {
        return { -1, QString() };
    }

    const QString &condition = m_tournamentSettings->singlesManagement().conditionString;
    const bool oldContains = oldName.contains(condition);
    const bool contains = name.contains(condition);

    if (oldContains == contains) {
        // No change in the defined singles state --> leave the marker untouched
        return { -1, QString() };
    }

    // See above about the logic used here
    if (contains == m_singleCondition->currentIndex()) {
        const auto &singleMarker = m_markersComboBoxes.value(Markers::Single);
        return { singleMarker->currentData().toInt(), singleMarker->currentText() };
    } else {
        const auto &assignedMarker = m_markersComboBoxes.value(Markers::Assigned);
        return { assignedMarker->currentData().toInt(), assignedMarker->currentText() };
    }
}

void MarkersWidget::addChecksumData(ChecksumHelper &checksumHelper) const
{
    // Add the markers data
    for (int i = 0; i < m_markers->rowCount(); i++) {
        QTableWidgetItem *item = m_markers->item(i, 1);

        const int id = item->data(MarkerData::Id).toInt();
        const QString name = id != 0 ? item->text() : QString();

        checksumHelper.addData(id);
        checksumHelper.addData(name);
        checksumHelper.addData(item->foreground().color().rgb());
        checksumHelper.addData(item->data(MarkerData::Sorting).toInt());
    }

    // Add the absent marker
    checksumHelper.addData(
        m_markersComboBoxes.value(Markers::Absent)->currentData().toInt());

    if (m_tournamentSettings->isPairMode()) {
        // Add the singles management data
        checksumHelper.addData(m_singlesManagementBox->isChecked());
        checksumHelper.addData(m_markersComboBoxes.value(
            Markers::Single)->currentData().toInt());
        checksumHelper.addData(m_markSingles->isChecked());
        checksumHelper.addData(m_singleConditionString->text());
        checksumHelper.addData(m_singleCondition->currentData().toInt());
        checksumHelper.addData(m_markersComboBoxes.value(
            Markers::Assigned)->currentData().toInt());
    }
}

void MarkersWidget::setServer(Server *server)
{
    m_server = server;
    m_isServer = true;
}

void MarkersWidget::setClient(Client *client)
{
    m_client = client;
    m_isClient = true;
}

void MarkersWidget::serverStopped()
{
    m_isServer = false;
}

void MarkersWidget::clientDisconnected()
{
    m_isClient = false;
}

void MarkersWidget::processNetworkChange()
{
    if (m_markerMenu->isVisible()) {
        m_markerMenu->hide();
    }

    reload();

    MarkerDialog *dialog = findChild<MarkerDialog *>();
    if (dialog != nullptr && dialog->isEditing()) {
        QTableWidgetItem *item = m_markers->item(m_selectedMarkerSequence, 1);
        if (item == nullptr ||
            (item->text() != m_selectedMarker.name
             || item->data(MarkerData::Id).toInt() != m_selectedMarker.id
             || item->data(MarkerData::Sorting) != m_selectedMarker.sorting
             || item->foreground().color() != m_selectedMarker.color)) {

            dialog->reject();
            dialog->deleteLater();

            // This is invoked via a QTimer::singleShot call so that the function returns
            // immediately and the network processing is not blocked until the dialog is closed
            QTimer::singleShot(0, this, [this] {
                QMessageBox::warning(this,
                    tr("Kollidierende Netzwerkänderung"),
                    tr("Die Markierung „%1“ wurde zwischenzeitlich von einem anderen "
                       "Netzwerkteilnehmer entweder gelöscht, umbenannt oder geändert. Das Ändern "
                       "ist nicht mehr möglich und wird abgebrochen.").arg(
                       m_selectedMarker.name)); } );
            return;
        }
    }
}

void MarkersWidget::setProcessingRequest(bool processing)
{
    setEnabled(! processing);
    if (processing) {
        QApplication::setOverrideCursor(Qt::WaitCursor);
    }
}

void MarkersWidget::updateTournamentState()
{
    m_defaultMarkerLabel->setEnabled(! m_db->tournamentStarted());
    m_markersComboBoxes.value(Markers::Default)->setEnabled(! m_db->tournamentStarted());
    m_absentMarkerBox->setEnabled(! m_db->tournamentStarted());
    m_singlesManagementBox->setEnabled(m_tournamentSettings->isPairMode()
                                       && ! m_db->tournamentStarted());
}

void MarkersWidget::singlesManagementChanged()
{
    if (! m_tournamentSettings->singlesManagement().enabled
        && m_singlesManagementBox->isChecked()) {

        warnAboutSpecialMarkers();
    }

    // First, we cache the currently entered data
    const Markers::SinglesManagement singlesManagement {
        m_singlesManagementBox->isChecked(),                                // enabled
        m_markersComboBoxes.value(Markers::Single)->currentData().toInt(),  // singleMarker
        m_markSingles->isChecked(),                                         // markSingles
        m_singleConditionString->text().simplified(),                       // conditionString
        m_singleCondition->currentIndex(),                                  // condition
        m_markersComboBoxes.value(Markers::Assigned)->currentData().toInt() // assignedMarker
    };

    if (m_isClient) {
        // Reset the GUI to the last state in case the request fails
        setSinglesManagement(m_tournamentSettings->singlesManagement());
        // Request the change
        m_client->requestSinglesManagementChange(singlesManagement);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    m_tournamentSettings->setSinglesManagement(singlesManagement);
    emit updatePlayersList();

    if (m_isServer) {
        m_server->broadcastSinglesManagementChange(singlesManagement);
    }

    QApplication::restoreOverrideCursor();
}

int MarkersWidget::singleMarker() const
{
    return m_singlesManagementBox->isChecked()
               ? m_markersComboBoxes.value(Markers::Single)->currentData().toInt()
               : -1;
}

int MarkersWidget::assignedMarker() const
{
    const auto &assignedMarker = m_markersComboBoxes.value(Markers::Assigned);
    return assignedMarker->isEnabled() ? assignedMarker->currentData().toInt() : -1;
}

void MarkersWidget::setSinglesManagement(const Markers::SinglesManagement &data)
{
    const auto &assignedMarker = m_markersComboBoxes.value(Markers::Assigned);
    const auto &singleMarker = m_markersComboBoxes.value(Markers::Single);

    m_singlesManagementBox->blockSignals(true);
    singleMarker->blockSignals(true);
    m_markSingles->blockSignals(true);
    m_singleConditionString->blockSignals(true);
    m_singleCondition->blockSignals(true);
    assignedMarker->blockSignals(true);

    m_singlesManagementBox->setChecked(data.enabled);
    setMarkersComboBoxSelection(singleMarker, data.singleMarker);
    m_markSingles->setChecked(data.markSingles);
    m_singleConditionString->setText(data.conditionString);
    m_singleCondition->setCurrentIndex(data.condition == 0 ? 0 : 1);
    setMarkersComboBoxSelection(assignedMarker, data.assignedMarker);

    m_singlesManagementBox->blockSignals(false);
    singleMarker->blockSignals(false);
    m_markSingles->blockSignals(false);
    m_singleConditionString->blockSignals(false);
    m_singleCondition->blockSignals(false);
    assignedMarker->blockSignals(false);
}

void MarkersWidget::resetSingleConditionString()
{
    m_singleConditionString->setText(m_settings->namesSeparator().simplified());
    singlesManagementChanged();
}

QString MarkersWidget::singleConditionString() const
{
    return m_singleConditionString->text();
}

void MarkersWidget::setCheckableMarker(Markers::Type type, int markerId)
{
    const auto &check = m_enableMarkerCheckBoxes.value(type);
    const auto &combo = m_markersComboBoxes.value(type);

    check->blockSignals(true);
    combo->blockSignals(true);

    check->setChecked(markerId != -1);
    combo->setEnabled(markerId != -1);
    if (markerId != -1) {
        setMarkersComboBoxSelection(combo, markerId);
    }

    check->blockSignals(false);
    combo->blockSignals(false);
}

void MarkersWidget::toggleAbsentMarker(bool state)
{
    m_markersComboBoxes.value(Markers::Absent)->setEnabled(state);
    absentMarkerChanged();
}

void MarkersWidget::absentMarkerChanged()
{
    if (m_tournamentSettings->marker(Markers::Absent) == -1) {
        warnAboutSpecialMarkers();
    }

    QString absentMarkerName;
    int absentMarkerId = -1;
    const auto &absentMarker = m_markersComboBoxes.value(Markers::Absent);
    if (absentMarker->isEnabled()) {
        absentMarkerName = absentMarker->currentText();
        absentMarkerId = absentMarker->currentData().toInt();
    }

    if (m_isClient) {
        // First reset the GUI to the last setting in case the request fails
        setCheckableMarker(Markers::Absent, m_tournamentSettings->marker(Markers::Absent));
        // Then request the change
        m_client->requestSetAbsentMarker(absentMarkerName, absentMarkerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    m_tournamentSettings->setMarker(Markers::Absent, absentMarkerId);

    if (m_isServer) {
        m_server->broadcastSetAbsentMarker(absentMarkerName, absentMarkerId);
    }

    emit updatePlayersList();

    QApplication::restoreOverrideCursor();
}

void MarkersWidget::warnAboutSpecialMarkers()
{
    if (m_specialMarkersWarnigAlreadyShown) {
        return;
    }

    QTimer::singleShot(0, this, [this] {
        QMessageBox::information(this, tr("Beenden der Anmeldung"),
            tr("<p><b>Bitte beachten:</b></p>"
               "<p>So lange es ignorierte und/oder als allein gekommene Anmeldungen gibt, kann "
               "das Turnier nicht gestartet werden.<p>"
               "<p>Bitte zum Starten des Turniers alle entsprechenden Markierungen entfernen, "
               "die so markierten Anmeldungen löschen oder die entsprechende Option deaktivieren."
               "</p>"));
    });

    m_specialMarkersWarnigAlreadyShown = true;
}
