/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MARKERSWIDGET_H
#define MARKERSWIDGET_H

// Local includes
#include "shared/Markers.h"

// Qt includes
#include <QWidget>
#include <QHash>
#include <QIcon>
#include <QVector>

// C++ includes
#include <utility>

// Local classes
class SharedObjects;
class Database;
class Settings;
class TournamentSettings;
class TitleMenu;
class Server;
class Client;
class ChecksumHelper;
class MinimumViewportWidget;
class MinimumViewportScroll;
// Qt classes
class QTableWidget;
class QAction;
class QMenu;
class QComboBox;
class QLineEdit;
class QGroupBox;
class QCheckBox;
class QLabel;

class MarkersWidget : public QWidget
{
    Q_OBJECT

public:
    enum MarkerData {
        Id = Qt::UserRole,
        Sorting
    };

    explicit MarkersWidget(QWidget *parent, SharedObjects *sharedObjects);
    void updateTournamentMode();
    void reload();
    QVector<Markers::Marker> getMarkers() const;
    void reset();
    bool isMarkerVisible(int id) const;
    void updateCounters(const QHash<int, int> &counter);
    int newRegistrationMarker(const QString &name) const;
    int drawnMarker() const;
    std::pair<int, QString> markerAfterEdit(const QString &oldName, const QString &name) const;
    void addChecksumData(ChecksumHelper &checksumHelper) const;
    void setServer(Server *server);
    void setClient(Client *client);
    void serverStopped();
    void clientDisconnected();
    void processNetworkChange();
    void setProcessingRequest(bool processing);
    void updateTournamentState();
    int singleMarker() const;
    int assignedMarker() const;
    void resetSingleConditionString();
    QString singleConditionString() const;
    void setCheckableMarker(Markers::Type type, int markerId);
    void setSinglesManagement(const Markers::SinglesManagement &data);
    void warnAboutSpecialMarkers();

signals:
    void markersChanged();
    void statusUpdate(const QString &text);
    void visibilityChanged();
    void updatePlayersList();

private slots:
    void showContextMenu(const QPoint &position);
    void addMarker();
    void saveNewMarker(const QString &name, const QColor &color,
                       Markers::MarkerSorting sorting);
    void moveMarker(int direction);
    void editMarker();
    void saveEdit(const QString &name, const QColor &color, Markers::MarkerSorting sorting);
    void deleteMarker();
    void checkShowContextMenu(int, int column);
    void changeVisibility(int row, int column);
    void showAll();
    void showOnly(int idToShow);
    void toggleAbsentMarker(bool state);
    void absentMarkerChanged();
    void singlesManagementChanged();
    void toggleDrawnMarker(bool state);
    void drawnMarkerChanged();

private: // Functions
    void updateVisibilityIcons();
    void updateVisibilityIcon(int id, int row);
    bool checkSelectedMarkerChanged(const QString &name, const QString &action);
    int markersComboBoxSelection(QComboBox *comboBox) const;
    void setMarkersComboBoxSelection(QComboBox *comboBox, const int markerId);

private: // Variables
    Database *m_db;
    Settings *m_settings;
    TournamentSettings *m_tournamentSettings;

    QTableWidget *m_markers;
    TitleMenu *m_markerMenu;
    QAction *m_editMarkerAction;
    QAction *m_moveUpAction;
    QAction *m_moveDownAction;
    QAction *m_deleteMarkerAction;
    QAction *m_addMarkerAction;
    QMenu *m_addMarkerMenu;
    QMenu *m_hideShowMarkedMenu;
    QAction *m_showOnlyAction;
    QAction *m_showAllAction;
    Markers::Marker m_selectedMarker;
    int m_selectedMarkerSequence;
    QIcon m_unused;
    QIcon m_visible;
    QIcon m_hidden;
    QHash<int, bool> m_isVisible;
    int m_allMarkersCount;

    QHash<Markers::Type, QComboBox *> m_markersComboBoxes;
    QHash<Markers::Type, QCheckBox *> m_enableMarkerCheckBoxes;
    QGroupBox *m_automaticMarkersBox;
    QLabel *m_defaultMarkerLabel;
    QGroupBox *m_absentMarkerBox;
    QGroupBox *m_singlesManagementBox;
    QCheckBox *m_markSingles;
    QComboBox *m_singleCondition;
    QLineEdit *m_singleConditionString;
    MinimumViewportWidget *m_markerSettingsWidget;
    MinimumViewportScroll *m_settingsScroll;

    bool m_specialMarkersWarnigAlreadyShown = false;

    Server *m_server;
    Client *m_client;
    bool m_isServer = false;
    bool m_isClient = false;
};

#endif // MARKERSWIDGET_H
