/* Copyright (c) 2010-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes

#include "RegistrationPage.h"
#include "RegistrationDelegate.h"
#include "MarkersWidget.h"
#include "MarkersPopup.h"
#include "DrawPopup.h"
#include "DrawOverviewPage.h"
#include "DrawSettingsDialog.h"

#include "ServerPage/Server.h"
#include "ClientPage/Client.h"

#include "shared/SharedObjects.h"
#include "shared/StringTools.h"
#include "shared/Database.h"
#include "shared/SearchWidget.h"
#include "shared/SearchEngine.h"
#include "shared/TitleMenu.h"
#include "shared/BackupEngine.h"
#include "shared/ResourceFinder.h"
#include "shared/SharedStyles.h"
#include "shared/Settings.h"
#include "shared/DefaultValues.h"
#include "shared/TournamentSettings.h"
#include "shared/Tournament.h"
#include "shared/Markers.h"
#include "shared/Draw.h"
#include "shared/PhoneticSearchButton.h"
#include "shared/DefeatableMessageBox.h"
#include "shared/DefeatableMessageBoxes.h"
#include "shared/ChecksumHelper.h"
#include "shared/IconSizeEngine.h"
#include "shared/Urls.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>
#include <QScrollBar>
#include <QAction>
#include <QMessageBox>
#include <QApplication>
#include <QSplitter>
#include <QCheckBox>
#include <QDir>
#include <QTimer>
#include <QCursor>

// C++ includes
#include <functional>
#include <utility>

RegistrationPage::RegistrationPage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_stringTools(sharedObjects->stringTools()),
      m_searchEngine(sharedObjects->searchEngine()),
      m_defeatableMessageBox(sharedObjects->defeatableMessageBox())
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    // Register new players/pairs

    auto *registerWidget = new QWidget;
    auto *registerLayout = new QVBoxLayout(registerWidget);
    registerLayout->setContentsMargins(0, 0, 0, 0);

    m_newPlayersBox = new QGroupBox;
    m_newPlayersBox->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    QHBoxLayout *newPlayersBoxLayout = new QHBoxLayout(m_newPlayersBox);
    registerLayout->addWidget(m_newPlayersBox);

    m_nameLabel = new QLabel;
    newPlayersBoxLayout->addWidget(m_nameLabel);

    m_playersName = new QLineEdit;
    newPlayersBoxLayout->addWidget(m_playersName);
    connect(m_playersName, &QLineEdit::textChanged,
            [this](const QString &text)
            {
                m_registerButton->setEnabled(! text.simplified().isEmpty());
            });
    connect(m_playersName, &QLineEdit::returnPressed, this, &RegistrationPage::registerPlayers);

    m_registerButton = new QPushButton(tr("Anmelden"));
    m_registerButton->setEnabled(false);
    newPlayersBoxLayout->addWidget(m_registerButton);
    connect(m_registerButton, &QPushButton::clicked, this, &RegistrationPage::registerPlayers);

    m_registerPhoneticSearch = new PhoneticSearchButton(
        Settings::registrationPhoneticId,
        sharedObjects,
        tr("<p>Phonetische Suche für die Suche nach Duplikaten bei der Anmeldung "
           "(de)aktivieren</p>"));
    newPlayersBoxLayout->addWidget(m_registerPhoneticSearch);

    // Registered players/pairs

    m_playersBox = new QGroupBox;
    QVBoxLayout *playersBoxLayout = new QVBoxLayout(m_playersBox);
    registerLayout->addWidget(m_playersBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    playersBoxLayout->addWidget(m_isSearchLabel);

    m_playersList = new QTableWidget(this);
    m_playersList->setColumnCount(4);
    m_playersList->setColumnHidden(COLUMN_INDEX.value(Column::Mark), true);
    m_playersList->setColumnHidden(COLUMN_INDEX.value(Column::Draw), true);
    m_playersList->setColumnHidden(COLUMN_INDEX.value(Column::Marker), true);
    m_delegate = new RegistrationDelegate(m_playersList);
    m_playersList->setItemDelegate(m_delegate);
    connect(m_delegate, &RegistrationDelegate::startedEdit,
            [this](int row)
            {
                m_editedRow = row;
            });
    connect(m_delegate, &RegistrationDelegate::editEntered, this, &RegistrationPage::editEntered);
    m_playersList->setHorizontalHeaderLabels( { tr("Name(n)"),
                                                QString(),
                                                tr("Auslosung"),
                                                tr("Markierung") } );

    auto *header = m_playersList->horizontalHeader();
    header->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(header, &QHeaderView::customContextMenuRequested,
            this, &RegistrationPage::showHeaderContextMenu);
    m_playersList->verticalHeader()->hide();
    m_playersList->setIconSize(sharedObjects->iconSizeEngine()->relativeSize(1.35));
    m_playersList->setShowGrid(false);
    m_playersList->setWordWrap(false);
    m_playersList->setSelectionMode(QTableWidget::NoSelection);
    m_playersList->setFocusPolicy(Qt::NoFocus);
    m_playersList->setHorizontalScrollMode(QAbstractItemView::ScrollPerPixel);
    connect(m_playersList->verticalScrollBar(), &QScrollBar::valueChanged,
            m_playersList, &QTableWidget::resizeColumnsToContents);
    m_playersList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_playersList, &QTableWidget::customContextMenuRequested,
            this, &RegistrationPage::showContextMenu);
    connect(m_playersList, &QTableWidget::cellClicked, this, &RegistrationPage::checkShowPopup);

    m_markerIcon = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("marker.png")));
    m_drawIcon = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("draw.png")));

    playersBoxLayout->addWidget(m_playersList);

    // Search widget

    auto *searchLayout = new QHBoxLayout;
    playersBoxLayout->addLayout(searchLayout);

    m_searchWidget = new SearchWidget(QStringLiteral("registration_page_search"), sharedObjects);
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &RegistrationPage::updateSearch);
    searchLayout->addWidget(m_searchWidget);

    m_adoptSearchText = new QToolButton;
    m_adoptSearchText->setArrowType(Qt::UpArrow);
    m_adoptSearchText->setToolTip(tr("Suchbegriff für die Anmeldung übernehmen"));
    m_adoptSearchText->hide();
    connect(m_adoptSearchText, &QToolButton::clicked, this, &RegistrationPage::adoptSearchText);
    searchLayout->addWidget(m_adoptSearchText);

    // Markers

    m_markersWidget = new MarkersWidget(this, sharedObjects);
    connect(m_markersWidget, &MarkersWidget::markersChanged,
            this, &RegistrationPage::markersChanged);
    connect(m_markersWidget, &MarkersWidget::statusUpdate, this, &RegistrationPage::statusUpdate);
    connect(m_markersWidget, &MarkersWidget::visibilityChanged,
            this, &RegistrationPage::markersVisibilityChanged);
    connect(m_markersWidget, &MarkersWidget::updatePlayersList,
            this, &RegistrationPage::updatePlayersList);

    auto *markersPartWidget = new QWidget;
    auto *markersPartWidgetLayout = new QVBoxLayout(markersPartWidget);
    markersPartWidgetLayout->setContentsMargins(0, 0, 0, 0);
    markersPartWidgetLayout->addWidget(m_markersWidget);

    // Options menu

    m_menuButton = new QPushButton(tr("Einstellungen"));
    auto *settingsMenu = new QMenu(this);

    m_columnsMenu = new TitleMenu(tr("Angezeigte Spalten"), this);
    m_columnsMenu->setTitleHidden(true);

    for (const Column column : ALL_COLUMNS) {
        m_showColumn.insert(column, m_columnsMenu->addAction(tr(CHECKBOX_TEXT.value(column))));
        m_showColumn[column]->setCheckable(true);
        connect(m_showColumn.value(column), &QAction::changed,
                this, std::bind(&RegistrationPage::setColumnVisible, this, column));
    }

    m_columnsMenu->addSeparator();
    m_hideHeaderAction = m_columnsMenu->addAction(tr("Kopfzeile ausblenden"));
    m_hideHeaderAction->setCheckable(true);
    connect(m_hideHeaderAction, &QAction::toggled, this, &RegistrationPage::setHideHeader);

    settingsMenu->addMenu(m_columnsMenu);

    settingsMenu->addSeparator();

    auto *showDrawSettingsAction = settingsMenu->addAction(tr("Einstellungen für die Auslosung"));
    connect(showDrawSettingsAction, &QAction::triggered, this, &RegistrationPage::showDrawSettings);

    m_menuButton->setMenu(settingsMenu);
    QHBoxLayout *settingsLayout = new QHBoxLayout;
    markersPartWidgetLayout->addLayout(settingsLayout);
    settingsLayout->addStretch();
    settingsLayout->addWidget(m_menuButton);

    // Arrange the registering part and the markers part in a splitter

    QSplitter *splitter = new QSplitter(this);
    splitter->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Expanding);
    splitter->setHandleWidth(2);
    splitter->setStyleSheet(SharedStyles::SPLITTER_VISIBLE_HORIZONTAL);

    splitter->addWidget(registerWidget);
    splitter->setCollapsible(0, false);
    splitter->setStretchFactor(0, 100);

    splitter->addWidget(markersPartWidget);
    splitter->setCollapsible(1, false);
    splitter->setStretchFactor(1, 1);

    layout->addWidget(splitter);

    // Status label
    m_statusLabel = new QLabel;
    m_statusLabel->setWordWrap(true);
    m_statusLabel->setTextInteractionFlags(Qt::TextBrowserInteraction);
    m_statusLabel->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    connect(m_statusLabel, &QLabel::linkActivated, this, &RegistrationPage::processStatusLinkClick);
    layout->addWidget(m_statusLabel);

    // Markers popup
    m_markersPopup = new MarkersPopup(this);
    connect(m_markersPopup, &QDialog::rejected, this, &RegistrationPage::resetHighlight);
    connect(m_markersPopup, &QDialog::accepted,
            [this]
            {
                updateMarker(m_markersPopup->markerName(), m_markersPopup->markerId());
            });

    // Draw popup
    m_drawPopup = new DrawPopup(this, &m_tournamentSettings->drawSettings());
    connect(m_drawPopup, &QDialog::rejected, this, &RegistrationPage::resetHighlight);
    connect(m_drawPopup, &QDialog::accepted,
            [this]
            {
                updateDraw(m_drawPopup->seat());
            });

    // Flash scrollto target reset timer
    m_resetHighlightTimer = new QTimer(this);
    m_resetHighlightTimer->setInterval(320);
    m_resetHighlightTimer->setSingleShot(true);
    connect(m_resetHighlightTimer, &QTimer::timeout, this, &RegistrationPage::resetHighlight);

    // Context menu

    m_contextMenu = new TitleMenu(this);
    connect(m_contextMenu, &QMenu::aboutToHide, this, &RegistrationPage::resetHighlight);

    m_assignMenu = m_contextMenu->addMenu(tr("Einzelspieler zuordnen"));
    m_assignMenu->menuAction()->setVisible(false);

    m_editAction = m_contextMenu->addAction(QString());
    connect(m_editAction, &QAction::triggered,
            [this]
            {
                m_playersList->editItem(m_playersList->currentItem());
            });

    m_deleteAction = m_contextMenu->addAction(QString());
    connect(m_deleteAction, &QAction::triggered, this, &RegistrationPage::deletePlayers);

    m_contextMenu->addSeparator();

    m_setOnlyMarkerAction = m_contextMenu->addAction(QString());
    m_setOnlyMarkerAction->setVisible(false);

    m_setMarkerMenu = m_contextMenu->addMenu(tr("Markieren als"));

    m_removeMarkerAction = m_contextMenu->addAction(tr("Markierung entfernen"));
    connect(m_removeMarkerAction, &QAction::triggered,
            this, std::bind(&RegistrationPage::updateMarker, this, QString(), 0));

    m_contextMenu->addSeparator();

    m_setDrawAction = m_contextMenu->addAction(tr("Auslosung eingeben"));
    connect(m_setDrawAction, &QAction::triggered,
            [this]
            {
                checkShowPopup(m_playersList->currentItem()->row(),
                               COLUMN_INDEX.value(Column::Draw));
            });

    m_deleteDrawAction = m_contextMenu->addAction(tr("Auslosung löschen"));
    connect(m_deleteDrawAction, &QAction::triggered,
            this, std::bind(&RegistrationPage::updateDraw, this, Draw::Seat { 0, 0 }));

    // Whole list menu

    m_listMenu = new TitleMenu(tr("Ganze Liste"), this);

    m_contextMenu->addSeparator();
    m_contextMenu->addMenu(m_listMenu);

    m_listSetOnlyMarkerAction = m_listMenu->addAction(QString());

    m_listSetMarkerMenu = m_listMenu->addMenu(tr("Markieren als"));

    m_listRemoveOnlyMarkerAction = m_listMenu->addAction(QString());

    m_listRemoveMarkerMenu = m_listMenu->addMenu(tr("Markierungen entfernen"));

    m_listRemoveAllMarkersAction = m_listMenu->addAction(tr("Alle Markierungen entfernen"));
    connect(m_listRemoveAllMarkersAction, &QAction::triggered,
            this, std::bind(&RegistrationPage::listUpdateMarker, this, QString(), 0));

    m_listMenu->addSeparator();

    m_deleteOnlyMarkedAction = m_listMenu->addAction(QString());

    m_deleteMarkedMenu = m_listMenu->addMenu(QString());

    m_deleteUnmarkedAction = m_listMenu->addAction(QString());
    connect(m_deleteUnmarkedAction, &QAction::triggered, this, &RegistrationPage::deleteUnmarked);

    m_deleteAllAction = m_listMenu->addAction(QString());
    connect(m_deleteAllAction, &QAction::triggered, this, &RegistrationPage::deleteAll);

    m_listMenu->addSeparator();

    m_drawAllSeatsAction = m_listMenu->addAction(tr("Alle Plätze auslosen"));
    connect(m_drawAllSeatsAction, &QAction::triggered, this, &RegistrationPage::drawAllSeats);
    m_drawAllSeatsAction->setEnabled(false);

    m_deleteAllDrawsAction = m_listMenu->addAction(tr("Alle Auslosungen löschen"));
    connect(m_deleteAllDrawsAction, &QAction::triggered, this, &RegistrationPage::deleteAllDraws);
}

void RegistrationPage::reload(bool updateMarkersWidget)
{
    m_newPlayersBox->show();

    switch (m_tournamentSettings->tournamentMode()) {
    case Tournament::Mode::FixedPairs:
        m_newPlayersBox->setTitle(tr("Neue Paare anmelden"));
        m_playersBox->setTitle(tr("Angemeldete Paare"));
        m_nameLabel->setText(tr("Name(n):"));
        m_editAction->setText(tr("Name(n) ändern"));
        m_deleteAction->setText(tr("Paar löschen"));
        m_deleteMarkedMenu->menuAction()->setText(tr("Alle Paare mit Markierung löschen"));
        m_deleteUnmarkedAction->setText(tr("Alle unmarkierten Paare löschen"));
        m_deleteAllAction->setText(tr("Alle Paare löschen"));
        break;
    case Tournament::Mode::SinglePlayers:
        m_newPlayersBox->setTitle(tr("Neue Spieler anmelden"));
        m_playersBox->setTitle(tr("Angemeldete Spieler"));
        m_nameLabel->setText(tr("Name:"));
        m_editAction->setText(tr("Name ändern"));
        m_deleteAction->setText(tr("Spieler löschen"));
        m_deleteMarkedMenu->menuAction()->setText(tr("Alle Spieler mit Markierung löschen"));
        m_deleteUnmarkedAction->setText(tr("Alle unmarkierten Spieler löschen"));
        m_deleteAllAction->setText(tr("Alle Spieler löschen"));
        break;
    }

    m_newPlayersBox->setEnabled(m_db->isChangeable());
    m_playersList->setEditTriggers(m_db->isChangeable() ? QAbstractItemView::DoubleClicked
                                                        : QAbstractItemView::NoEditTriggers);
    m_drawPopup->setPairMode(m_tournamentSettings->isPairMode());
    m_markersWidget->updateTournamentMode();

    m_hideHeaderAction->setChecked(m_tournamentSettings->registrationListHeaderHidden());

    // This will also trigger updatePlayersList() each
    if (updateMarkersWidget) {
        m_markersWidget->reload();
    } else {
        markersChanged();
    }

    checkUpdateDrawOverviewPage(true);
}

void RegistrationPage::markersChanged()
{
    const QVector<Markers::Marker> markers(m_markersWidget->getMarkers());
    const int count = markers.count();

    // No need to process count < 2, because the actions will be hidden in this case

    if (count == 2) {
        // We have exactly one marker

        m_setOnlyMarkerAction->disconnect();
        m_listSetOnlyMarkerAction->disconnect();
        m_listRemoveOnlyMarkerAction->disconnect();
        m_deleteOnlyMarkedAction->disconnect();

        for (const Markers::Marker &marker : markers) {
            if (marker.id != 0) {
                m_setOnlyMarkerAction->setText(tr("Markieren als „%1“").arg(marker.name));
                connect(m_setOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&RegistrationPage::updateMarker, this,
                                        marker.name, marker.id));

                m_listSetOnlyMarkerAction->setText(tr("Markieren als „%1“").arg(marker.name));
                connect(m_listSetOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&RegistrationPage::listUpdateMarker, this,
                                        marker.name, marker.id));

                m_listRemoveOnlyMarkerAction->setText(tr("Markierung „%1“ entfernen").arg(
                                                          marker.name));
                connect(m_listRemoveOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&RegistrationPage::listRemoveMarker,
                                        this, marker.name, marker.id));

                m_deleteOnlyMarkedAction->setText((m_tournamentSettings->isPairMode()
                    ? tr("Alle als „%1“ markierten Paare löschen")
                    : tr("Alle als „%1“ markierten Spieler löschen")).arg(marker.name));
                m_deleteOnlyMarkedAction->setData(marker.id);
                connect(m_deleteOnlyMarkedAction, &QAction::triggered,
                        this, std::bind(&RegistrationPage::deleteMarked, this,
                                        marker.name, marker.id));

                break;
            }
        }
    }

    if (count > 2) {
        // We have multiple markers
        m_setMarkerMenu->clear();
        m_listSetMarkerMenu->clear();
        m_listRemoveMarkerMenu->clear();
        m_deleteMarkedMenu->clear();

        for (const Markers::Marker &marker : markers) {
            if (marker.id == 0) {
                continue;
            }

            QAction *setMarkerAction = m_setMarkerMenu->addAction(marker.name);
            setMarkerAction->setData(marker.id);
            connect(setMarkerAction, &QAction::triggered,
                    this, std::bind(&RegistrationPage::updateMarker, this,
                                    marker.name, marker.id));

            QAction *listSetMarkerAction = m_listSetMarkerMenu->addAction(marker.name);
            listSetMarkerAction->setData(marker.id);
            connect(listSetMarkerAction, &QAction::triggered,
                    this, std::bind(&RegistrationPage::listUpdateMarker, this,
                                    marker.name, marker.id));

            QAction *listRemoveMarkerAction = m_listRemoveMarkerMenu->addAction(marker.name);
            listRemoveMarkerAction->setData(marker.id);
            connect(listRemoveMarkerAction, &QAction::triggered,
                    this, std::bind(&RegistrationPage::listRemoveMarker, this,
                                    marker.name, marker.id));

            QAction *deleteMarkedAction = m_deleteMarkedMenu->addAction(marker.name);
            deleteMarkedAction->setData(marker.id);
            connect(deleteMarkedAction, &QAction::triggered,
                    this, std::bind(&RegistrationPage::deleteMarked, this,
                                    marker.name, marker.id));
        }
    }

    m_setOnlyMarkerAction->setVisible(count == 2);
    m_setMarkerMenu->menuAction()->setVisible(count > 2);
    m_removeMarkerAction->setVisible(count > 1);

    m_listSetOnlyMarkerAction->setVisible(count == 2);
    m_listSetMarkerMenu->menuAction()->setVisible(count > 2);
    m_listRemoveOnlyMarkerAction->setVisible(count == 2);
    m_listRemoveMarkerMenu->menuAction()->setVisible(count > 2);
    m_listRemoveAllMarkersAction->setVisible(count > 2);

    m_deleteOnlyMarkedAction->setVisible(count == 2);
    m_deleteMarkedMenu->menuAction()->setVisible(count > 2);
    m_deleteUnmarkedAction->setVisible(count > 2);

    listModified();

    // Update the markers popup
    m_markersPopup->updateMarkers(markers);
}

void RegistrationPage::updatePlayersList()
{
    // Be sure not to produce a segfault via the resetHighlight timer call
    m_highlightedItem = nullptr;
    if (m_resetHighlightTimer->isActive()) {
        m_resetHighlightTimer->stop();
    }

    // This is only for a possibly existing draw adjustment
    // dialog if we just processed a network change.
    emit listWasModified();

    // Create a list of all blocks we need
    QHash<int, int> targetBlock;
    QVector<QVector<QString>> block;
    QHash <int, QColor> markerColor;
    QHash <int, QString> markerName;
    const auto markers = m_markersWidget->getMarkers();
    for (const Markers::Marker &marker : markers) {
        switch (marker.sorting) {
        case Markers::MarkerSorting::NewBlock:
            block.append(QVector<QString>());
            targetBlock[marker.id] = block.count() - 1;
            break;
        case Markers::MarkerSorting::Merged:
            targetBlock[marker.id] = block.count() - 1;
            break;
        }
        if (marker.id != 0) {
            markerColor[marker.id] = marker.color;
            markerName[marker.id] = marker.name;
        }
    }

    // In case we have invalid markers set in the players list, the program would crash. So we
    // have to catch this case. This shouldn't happen however unless the database is manipulated
    // by hand or reading the markers failed.

    // If targetBlock doesn't contain a value for 0 ("unmarked"), we have to add it
    if (! targetBlock.contains(0)) {
        block.append(QVector<QString>());
        targetBlock.insert(0, block.count() - 1);
    }

    // Fetch the players data
    auto [ playersData, success ] = m_db->playersData();
    if (! success) {
        return;
    }
    QStringList registrations = playersData.keys();
    m_stringTools->sort(registrations);

    // Search for invalid markers and map them to "unmarked"
    for (const QString &name : std::as_const(registrations)) {
        Database::PlayersData &data = playersData[name];
        if (! targetBlock.contains(data.marker)) {
            data.marker = 0;
        }
    }

    const int oldRegistrationsCount = m_playersList->rowCount();
    const int registrationsCount = registrations.count();
    m_playersList->blockSignals(true);
    m_playersList->setRowCount(registrationsCount);
    m_searchEngine->clearSearchCache();

    // Iterate through all players, add them to their respective blocks and count the markers

    QHash<int, int> markersCount;

    for (const QString &name : std::as_const(registrations)) {
        m_searchEngine->updateSearchCache(name);
        block[targetBlock.value(playersData.value(name).marker)].append(name);
        markersCount[playersData.value(name).marker]++;
    }

    // Display all blocks

    int blockRow = 0;
    int drawsCount = 0;
    QVector<int> assignedPair1Tables;
    QVector<int> assignedPair2Tables;

    m_allRegistrations.clear();
    m_allSetMarkers.clear();
    m_allDraws.clear();

    for (const QVector<QString> &list : std::as_const(block)) {
        for (const QString &name : list) {
            const Database::PlayersData &data = playersData.value(name);

            m_allRegistrations.append(name);
            m_allSetMarkers.append(data.marker);
            m_allDraws.append({ data.assignedTable, data.assignedPairNumber });

            QTableWidgetItem *nameCell = new QTableWidgetItem(name);
            if (data.assignedTable != 0) {
                if (data.assignedPairNumber == 1) {
                    assignedPair1Tables.append(data.assignedTable);
                } else {
                    assignedPair2Tables.append(data.assignedTable);
                }
            }

            QTableWidgetItem *setMarkerCell = new QTableWidgetItem;
            setMarkerCell->setFlags(setMarkerCell->flags() & (~Qt::ItemIsEditable));
            setMarkerCell->setIcon(m_markerIcon);

            QTableWidgetItem *drawCell = new QTableWidgetItem;
            drawCell->setFlags(drawCell->flags() & (~Qt::ItemIsEditable));
            if (data.assignedTable != 0) {
                drawsCount++;
                drawCell->setText(tr("Tisch %1 Paar %2").arg(
                                        QString::number(data.assignedTable),
                                        QString::number(data.assignedPairNumber)));
            } else {
                drawCell->setIcon(m_drawIcon);
            }

            QTableWidgetItem *markerCell = new QTableWidgetItem;
            markerCell->setFlags(markerCell->flags() & (~Qt::ItemIsEditable));
            if (playersData.value(name).marker != 0) {
                nameCell->setForeground(markerColor.value(playersData.value(name).marker));
                markerCell->setText(markerName.value(playersData.value(name).marker));
            }

            m_playersList->setItem(blockRow, COLUMN_INDEX.value(Column::Name), nameCell);
            m_playersList->setItem(blockRow, COLUMN_INDEX.value(Column::Mark), setMarkerCell);
            m_playersList->setItem(blockRow, COLUMN_INDEX.value(Column::Draw), drawCell);
            m_playersList->setItem(blockRow, COLUMN_INDEX.value(Column::Marker), markerCell);

            blockRow++;
        }
    }

    std::sort(assignedPair1Tables.begin(), assignedPair1Tables.end());
    std::sort(assignedPair2Tables.begin(), assignedPair2Tables.end());
    const bool drawsSet = assignedPair1Tables.count() > 0 || assignedPair2Tables.count() > 0;
    m_deleteAllDrawsAction->setEnabled(drawsSet);
    m_drawPopup->setAssignedTables(assignedPair1Tables, assignedPair2Tables);
    updateVisibility();

    // Update the context menus

    if (m_listSetMarkerMenu->menuAction()->isVisible()) {
        const auto listSetMarkerMenuActions = m_listSetMarkerMenu->actions();
        for (QAction *action : listSetMarkerMenuActions) {
            action->setEnabled(markersCount.value(action->data().toInt()) < registrationsCount);
        }
        const auto listRemoveMarkerMenuActions = m_listRemoveMarkerMenu->actions();
        for (QAction *action : listRemoveMarkerMenuActions) {
            action->setEnabled(markersCount.value(action->data().toInt()) > 0);
        }
        const auto deleteMarkedMenuActions = m_deleteMarkedMenu->actions();
        for (QAction *action : deleteMarkedMenuActions) {
            action->setEnabled(markersCount.value(action->data().toInt()) > 0);
        }
    }

    if (markersCount.value(0) == registrationsCount) {
        m_listRemoveOnlyMarkerAction->setEnabled(false);
        m_listRemoveMarkerMenu->setEnabled(false);
        m_listRemoveAllMarkersAction->setEnabled(false);
    } else {
        m_listRemoveOnlyMarkerAction->setEnabled(true);
        m_listRemoveMarkerMenu->setEnabled(true);
        m_listRemoveAllMarkersAction->setEnabled(true);
    }

    if (m_db->tournamentStarted() || markersCount.value(0) == registrationsCount) {
        m_deleteOnlyMarkedAction->setEnabled(false);
        m_deleteMarkedMenu->setEnabled(false);
        m_deleteUnmarkedAction->setEnabled(false);
    } else {
        m_deleteOnlyMarkedAction->setEnabled(
            markersCount.value(m_deleteOnlyMarkedAction->data().toInt()) < registrationsCount);
        m_deleteMarkedMenu->setEnabled(true);
        m_deleteUnmarkedAction->setEnabled(markersCount.value(0) > 0);
    }

    // Update the GUI

    m_markersWidget->updateCounters(markersCount);

    m_playersList->resizeColumnsToContents();
    m_playersList->blockSignals(false);

    m_searchWidget->setVisible(registrationsCount > 0);
    emit registrationsAvailable(registrationsCount > 0);

    const int singleMarker = m_db->tournamentStarted() ? -1 : m_markersWidget->singleMarker();
    const bool oldSinglesPresent = m_singlesPresent;
    m_singlesPresent = markersCount.value(singleMarker) > 0;
    const bool singlesPresentChanged = oldSinglesPresent != m_singlesPresent;

    const int absentMarker = m_db->tournamentStarted()
        ? -1 : m_tournamentSettings->marker(Markers::Absent);
    const bool oldAbsentsPresent = m_absentsPresent;
    m_absentsPresent = markersCount.value(absentMarker) > 0;
    const bool absentsPresentChanged = oldAbsentsPresent != m_absentsPresent;

    int presentPairsWithoutSingles = registrationsCount;
    if (singleMarker != -1) {
        presentPairsWithoutSingles -= markersCount.value(singleMarker);
    }
    if (absentMarker != -1) {
        presentPairsWithoutSingles -= markersCount.value(absentMarker);
    }

    // Update the registration status

    m_drawAdjustmentNeeded = false;
    QString registrationStatus;

    if (m_absentsPresent) {
        registrationStatus.append(tr("<span style=\"%1\">%n Anmeldung(en) ignoriert</span> – ", "",
                                     markersCount.value(absentMarker)).arg(
                                     SharedStyles::RED_TEXT));
    }

    registrationStatus.append(m_tournamentSettings->isPairMode()
                                  ? tr("%n Paar(e)", "", presentPairsWithoutSingles)
                                  : tr("%n Spieler", "", presentPairsWithoutSingles));

    if (! m_singlesPresent) {
        // All registrations are valid and can be counted as-is
        registrationStatus.append(tr(" angemeldet"));

        if (presentPairsWithoutSingles > 0) {
            if (m_db->playersDividable(presentPairsWithoutSingles)) {
                registrationStatus.append(takenTablesString(
                    presentPairsWithoutSingles / m_db->necessaryDivisor(), true));
            } else {
                registrationStatus.append(m_tournamentSettings->isPairMode()
                    ? tr(" – <span style=\"%1\">ein Paar fehlt noch</span>").arg(
                         SharedStyles::RED_TEXT)
                    : tr(" – <span style=\"%1\">%n Spieler fehlt/fehlen noch</span>", "",
                         m_db->necessaryDivisor()
                         - presentPairsWithoutSingles % m_db->necessaryDivisor()).arg(
                         SharedStyles::RED_TEXT));
                registrationStatus.append(takenTablesString(
                    presentPairsWithoutSingles / m_db->necessaryDivisor() + 1, false));
            }
        }

    } else {
        // We're in pair mode and have to consider single players
        const int remainder = (presentPairsWithoutSingles * 2
                               + markersCount.value(singleMarker)) % 4;
        registrationStatus.append(tr(" und <span style=\"%1\"> %n allein gekommene(r) Spieler"
                                     "</span> angemeldet<br/>", "",
                                     markersCount.value(singleMarker)).arg(
                                     SharedStyles::RED_TEXT));
        if (remainder == 0) {
            registrationStatus.append(tr("<span style=\"%1\">Wenn alle zugeordnet sind, kann es "
                                         "losgehen</span>").arg(SharedStyles::GREEN_TEXT));
        } else if (remainder == 1) {
            registrationStatus.append(tr("<span style=\"%1\">Es fehlt noch ein Paar und ein "
                                         "einzelner Spieler</span>").arg(
                                         SharedStyles::RED_TEXT));
        } else if (remainder == 2) {
            registrationStatus.append(tr("<span style=\"%1\">Es fehlt noch ein Paar</span>").arg(
                                         SharedStyles::RED_TEXT));
        } else if (remainder == 3) {
            registrationStatus.append(tr("<span style=\"%1\">Es fehlt noch ein einzelner Spieler"
                                         "</span>").arg(SharedStyles::RED_TEXT));
        }
        registrationStatus.append(takenTablesString(
            (presentPairsWithoutSingles * 2 + markersCount.value(singleMarker)) / 4
            + (remainder != 0), false));
    }

    // Add the "no draw yet" line if we have a draw

    if (m_showColumn.value(Column::Draw)->isChecked() && ! drawsSet) {
        registrationStatus.append(tr("<br/>Es wurde noch keine Auslosung eingegeben"));

    } else if (drawsSet) {
        if (m_drawOverviewPage == nullptr) {
            registrationStatus.append(tr("<br/>Auslosung (<a href=\"#drawOverview\">Übersicht"
                                         "</a>): "));
        } else {
            registrationStatus.append(tr("<br/>Auslosung: "));
        }

        const bool unfinished = drawsCount != registrationsCount;

        // Generate a sorted list of all drawn tables
        QSet<int> assignedTablesSet;
        for (const int table : std::as_const(assignedPair1Tables)) {
            assignedTablesSet.insert(table);
        }
        for (const int table : std::as_const(assignedPair2Tables)) {
            assignedTablesSet.insert(table);
        }
#if QT_VERSION >= QT_VERSION_CHECK(5, 14, 0)
        QVector<int> allAssignedTables(assignedTablesSet.begin(), assignedTablesSet.end());
#else
        QList<int> allAssignedTables = QList<int>::fromSet(assignedTablesSet);
#endif
        std::sort(allAssignedTables.begin(), allAssignedTables.end());

        // Check for tables with incomplete (missing pair 1 or 2 or incomplete pairs) draw
        m_incompleteTables.clear();
        const int neededCount = m_tournamentSettings->isPairMode() ? 1 : 2;
        for (const int table : std::as_const(allAssignedTables)) {
            if (   assignedPair1Tables.count(table) != neededCount
                || assignedPair2Tables.count(table) != neededCount) {

                m_incompleteTables.append(table);
            }
        }
        const bool incompleteTables = m_incompleteTables.count() > 0;

        // Search for gaps in the draw
        m_tableGaps.clear();
        int consecutiveNumber = 1;
        for (int i = 0; i < allAssignedTables.count(); i++) {
            if (consecutiveNumber < allAssignedTables.at(i)) {
                for (int j = consecutiveNumber; j < allAssignedTables.at(i); j++) {
                    m_tableGaps.append(j);
                }
                consecutiveNumber = allAssignedTables.at(i) + 1;
            } else {
                consecutiveNumber++;
            }
        }
        const bool tableGaps = m_tableGaps.count() > 0;

        // Append the draw status to the status line

        registrationStatus.append(unfinished
            ? tr("<a href=\"#unfinished\">Unvollständig</a>; ")
            : tr("Vollständig; "));

        if (incompleteTables) {
            registrationStatus.append(tr("<a href=\"#incompleteTables\">unvollständige Tische"
                                         "</a>"));
        }

        if (tableGaps) {
            registrationStatus.append(incompleteTables
                ? tr("; <a href=\"#tableGaps\">nicht fortlaufend</a>")
                : tr("<a href=\"#tableGaps\">nicht fortlaufend</a>"));
        }

        if (! incompleteTables && ! tableGaps) {
            registrationStatus.append(tr("alle Tische sind fortlaufend und vollständig vergeben"));
        }

        m_drawAdjustmentNeeded = ! unfinished && ! m_singlesPresent && ! m_absentsPresent
                                 && m_db->playersDividable(presentPairsWithoutSingles)
                                 && (incompleteTables || tableGaps);
        if (m_drawAdjustmentNeeded) {
            registrationStatus.append(tr(" – <a href=\"#adjustDraw\">Auslosung bereinigen</a>"));
        }
    }

    emit drawAdjustmentNeeded(m_drawAdjustmentNeeded);

    // Finish the update

    m_statusLabel->setText(registrationStatus);

    if (registrationsCount != oldRegistrationsCount
        || singlesPresentChanged || absentsPresentChanged) {

        m_db->setPairsOrPlayersCount(registrationsCount);
        // Emitting the registrationsCountChanged signal will make the MainWindow check for single
        // players being present and hide/show the score dock accordingly, also if the number of
        // registered players wasn't changed by the previous action (e. g. marking an entry as
        // single)
        emit playersCountChanged(registrationsCount);

        m_drawAllSeatsAction->setEnabled(m_db->playersDividable(registrationsCount)
                                         && ! m_singlesPresent
                                         && ! m_absentsPresent);
    }

    if (m_db->tournamentStarted()) {
        m_newPlayersBox->hide();
    }

    scoreChanged();

    if (m_isServer || m_isClient) {
        updateChecksum();
    }
}

QString RegistrationPage::takenTablesString(int tables, bool definitive) const
{
    const int tablesLimit = m_tournamentSettings->drawSettings().tablesLimit;

    if (tablesLimit == -1 || tables <= tablesLimit) {
        if (definitive) {
            return tr(" (%n Tisch(e) besetzt)", "", tables);
        } else {
            return tr(" (dann wäre(n) %n Tisch(e) besetzt)", "", tables);
        }
    } else {
        if (definitive) {
            return tr(" (%n Tisch(e) besetzt – <span style=\"%1\">Limit laut "
                      "Auslosungseinstellungen: %2</span>)", "", tables).arg(
                      SharedStyles::RED_TEXT, QString::number(tablesLimit));
        } else {
            return tr(" (dann wäre(n) %n Tisch(e) besetzt – <span style=\"%1\">Limit laut "
                      "Auslosungseinstellungen: %2</span>)", "", tables).arg(
                      SharedStyles::RED_TEXT, QString::number(tablesLimit));
        }
    }
}

void RegistrationPage::scoreChanged()
{
    m_deleteAction->setEnabled(! m_db->tournamentStarted());
    m_deleteAllAction->setEnabled(! m_db->tournamentStarted());
    m_markersWidget->updateTournamentState();
}

void RegistrationPage::reset()
{
    m_playersList->clearContents();
    m_newPlayersBox->show();
}

void RegistrationPage::registerPlayers()
{
    if (! m_registerButton->isEnabled()) {
        return;
    }

    const QString name = m_playersName->text().simplified();

    // Check for an exact match. In this case, we can't register the pair/player
    const QString lowerName = m_stringTools->toLower(name);
    for (const auto &registered : std::as_const(m_allRegistrations)) {
        if (lowerName == m_stringTools->toLower(registered)) {
            QMessageBox::warning(this, m_newPlayersBox->title(),
                tr("<p><b>%1 „%2“ kann nicht angemeldet werden</b></p>"
                   "<p>Der Name ist bereits vergeben.</p>").arg(
                   m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"), name));
            m_playersName->selectAll();
            m_playersName->setFocus();
            return;
        }
    }

    // Search for a fuzzy match

    m_searchEngine->setSearchTerm(name);

    for (const auto &registered : std::as_const(m_allRegistrations)) {
        if (m_searchEngine->checkMatch(registered, SearchEngine::SearchType::Default,
                                       m_registerPhoneticSearch->isChecked())) {

            if (QMessageBox::warning(this,
                m_newPlayersBox->title(),
                tr("<p>Es ist bereits ein %1 mit einem ähnlichen Namen angemeldet:</p>"
                   "<p><table>"
                   "<tr><td>Eingegeben:&nbsp;</td><td><i>%2</i></td></tr>"
                   "<tr><td>Vorhanden:</td><td><i>%3</i></td></tr>"
                   "</table></p>"
                   "<p>Soll %4 trotzdem angemeldet werden?</p>").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                   name, registered,
                   m_tournamentSettings->isPairMode() ? tr("das Paar") : tr("der Spieler")),
                QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No) {

                m_playersName->selectAll();
                m_playersName->setFocus();
                return;
            }
            break;
        }
    }

    if (m_isClient || m_isServer) {
        if (m_playersList->findItems(name, Qt::MatchExactly).count() != 0) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("Es wurde mittlerweile von einem anderen Netzwerkteilnehmer ein %1 mit dem "
                   "Namen „%2“ angemeldet. Die Anmeldung ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name));
            m_playersName->clear();
            return;
        }
    }

    if (m_isClient) {
        m_requestingRegistration = true;
        m_playersList->setCurrentItem(nullptr);
        m_client->requestRegisterPlayers(name, m_markersWidget->newRegistrationMarker(name));
        m_focusPlayersName = true;
        return;
    }

    emit statusUpdate(m_tournamentSettings->isPairMode() ? tr("Melde Paar an …")
                                                         : tr("Melde Spieler an …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (! m_db->registerPlayers(name, m_markersWidget->newRegistrationMarker(name))) {
        m_playersName->clear();
        return;
    }
    m_playersName->clear();
    m_playersName->setFocus();
    updatePlayersList();
    m_searchWidget->clear();

    emit statusUpdate((m_tournamentSettings->isPairMode()
                           ? tr("Paar „%1“ angemeldet") : tr("Spieler „%1“ angemeldet")).arg(name));

    scrollTo(name);

    if (m_isServer) {
        m_server->broadcastRegisterPlayers(name, m_markersWidget->newRegistrationMarker(name));
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::scrollTo(const QString &text)
{
    const auto matches = m_playersList->findItems(text, Qt::MatchExactly);
    if (matches.isEmpty()) {
        itemError(Q_FUNC_INFO, text);
        return;
    }
    auto *target = matches.at(0);

    if (target->row() == m_playersList->rowCount() - 1) {
        m_playersList->scrollToBottom();
    } else {
        m_playersList->scrollToItem(target, QAbstractItemView::PositionAtCenter);
    }

    highlightItem(target);
    m_resetHighlightTimer->start();
}

bool RegistrationPage::checkForExactMatch(const QString &name)
{
    const QString lowerName = m_stringTools->toLower(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (row == m_editedRow) {
            continue;
        }
        if (lowerName == m_stringTools->toLower(m_playersList->item(
                row, COLUMN_INDEX.value(Column::Name))->text())) {

            return true;
        }
    }
    return false;
}

void RegistrationPage::editEntered(const QString &originalName, const QString &name)
{
    // Check for an exact match. In this case, we can't rename the pair/player
    if (checkForExactMatch(name)) {
        QMessageBox::critical(this,
            m_newPlayersBox->title(),
            tr("<p><b>%1 „%2“ kann nicht umbenannt werden:</b></p>"
               "<p>Der Name „%3“ ist bereits vergeben.</p>").arg(
               m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
               originalName, name));
        return;
    }

    // Seach for a fuzzy match

    bool fuzzyMatchFound = false;
    QString fuzzyMatchText;

    m_searchEngine->setSearchTerm(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (row == m_editedRow) {
            continue;
        }

        fuzzyMatchText = m_playersList->item(row, COLUMN_INDEX.value(Column::Name))->text();
        if (m_searchEngine->checkMatch(fuzzyMatchText, SearchEngine::SearchType::Default, true)) {
            fuzzyMatchFound = true;
            break;
        }
    }

    if (fuzzyMatchFound) {
        auto *warning = new QMessageBox(
            QMessageBox::Warning,
            m_newPlayersBox->title(),
            tr("<p>Es ist bereits ein %1 mit einem ähnlichen Namen angemeldet:</p>"
               "<p><table>"
               "<tr><td>Eingegeben:&nbsp;</td><td><i>%2</i></td></tr>"
               "<tr><td>Vorhanden:</td><td><i>%3</i></td></tr>"
               "</table></p>"
               "<p>Das könnte zu Verwechslungen führen. Soll %4 trotzdem umbenannt "
               "werden?</p>").arg(
               m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
               name, fuzzyMatchText,
               m_tournamentSettings->isPairMode() ? tr("das Paar") : tr("Der Spieler")),
            QMessageBox::Yes | QMessageBox::No,
            this);
        warning->setDefaultButton(QMessageBox::No);
        m_fuzzyMatchWarning = warning;

        bool abort = ! warning->exec()
                     || warning->clickedButton() != warning->button(QMessageBox::Yes);

        warning->deleteLater();
        m_fuzzyMatchWarning = nullptr;

        if (abort) {
            return;
        }
    }

    if (m_isClient || m_isServer) {
        if (m_playersList->findItems(originalName, Qt::MatchExactly).count() != 1) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("%1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer entweder "
                   "umbenannt oder gelöscht. Das Umbenennen ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
                   originalName));
            return;
        }
        if (m_playersList->findItems(name, Qt::MatchExactly).count() != 0) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("Es wurde mittlerweile von einem anderen Netzwerkteilnehmer %1 mit dem "
                   "Namen „%2“ angemeldet oder %1 entsprechend umbenannt. Das Umbenennen ist "
                   "nicht mehr möglich und wird abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("ein Paar") : tr("ein Spieler"), name));
            return;
        }
    }

    const auto [ markerAfterEditId, markerAfterEditName ]
        = m_markersWidget->markerAfterEdit(originalName, name);

    if (m_isClient) {
        m_client->requestRenamePlayers(originalName, name, markerAfterEditId, markerAfterEditName);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    QString changeMarkerStartMessage;
    QString changeMarkerFinishedMessage;
    if (markerAfterEditId != -1) {
        if (markerAfterEditId == 0) {
            changeMarkerStartMessage = tr("und entferne die Markierung ");
            changeMarkerFinishedMessage = tr("und die Markierung entfernt");
        } else {
            changeMarkerStartMessage = tr(" und setze die Markierung auf „%1“ ").arg(
                                          markerAfterEditName);
            changeMarkerFinishedMessage = tr(" und die Markierung auf „%1“ gesetzt").arg(
                                             markerAfterEditName);
        }
    }

    emit statusUpdate((m_tournamentSettings->isPairMode()
        ? tr("Benenne Paar um %1…")
        : tr("Benenne Spieler um %1…")).arg(
             changeMarkerStartMessage));

    if (! m_db->renamePlayers(originalName, name, markerAfterEditId)) {
        return;
    }

    emit statusUpdate((m_tournamentSettings->isPairMode()
        ? tr("Paar „%1“ umbenannt zu „%2“%3")
        : tr("Spieler „%1“ umbenannt zu „%2“%3")).arg(
             originalName, name, changeMarkerFinishedMessage));

    listModified();
    checkUpdateDrawOverviewPage(true, name);

    const auto matches = m_playersList->findItems(name, Qt::MatchExactly);
    if(matches.isEmpty()) {
        itemError(Q_FUNC_INFO, name);
        return;
    }

    if (matches.at(0)->row() != m_editedRow) {
        scrollTo(name);
    }

    if (m_isServer) {
        m_server->broadcastRenamePlayers(originalName, name,
                                         markerAfterEditId, markerAfterEditName);
    }

    emit nameEdited();
    QApplication::restoreOverrideCursor();
}

void RegistrationPage::showContextMenu(QPoint point)
{
    if (m_playersList->rowCount() == 0 || ! m_db->isChangeable()) {
        return;
    }

    QTableWidgetItem *selectedItem = m_playersList->itemAt(point);

    if (m_playersList->horizontalHeader()->isVisible()) {
        point.setY(point.y() + m_playersList->horizontalHeader()->height());
    }
    if (selectedItem == nullptr) {
        m_listMenu->exec(m_playersList->mapToGlobal(point));
        return;
    }

    if (selectedItem->column() != COLUMN_INDEX.value(Column::Name)) {
        selectedItem = m_playersList->item(selectedItem->row(), COLUMN_INDEX.value(Column::Name));
    }

    highlightItem(selectedItem);

    m_selectedName = selectedItem->text();
    const int selectedMark = m_allSetMarkers.at(selectedItem->row());

    const int singleMarker = m_markersWidget->singleMarker();
    const bool isSinglePlayer = selectedMark == singleMarker;
    m_assignMenu->menuAction()->setVisible(isSinglePlayer);
    m_contextMenu->setTitle((m_tournamentSettings->isPairMode() && ! isSinglePlayer
                                ? tr("Paar „%1“") : tr("Spieler „%1“")).arg(m_selectedName));
    if (isSinglePlayer) {
        m_assignMenu->clear();
        for (int row = 0; row < m_playersList->rowCount(); row++) {
            const QString name = m_playersList->item(row, COLUMN_INDEX.value(Column::Name))->text();
            if (m_allSetMarkers.at(row) != singleMarker || name == m_selectedName) {
                continue;
            }
            QAction *action = m_assignMenu->addAction(name);
            connect(action, &QAction::triggered,
                    this, std::bind(&RegistrationPage::assignPlayer, this, name));
        }
        if (m_assignMenu->actions().count() == 0) {
            QAction *action = m_assignMenu->addAction(tr("keiner sonst angemeldet"));
            action->setEnabled(false);
        }

        m_editAction->setText(tr("Name ändern"));
        m_deleteAction->setText(tr("Spieler löschen"));


    } else if (m_tournamentSettings->isPairMode()) {
        m_editAction->setText(tr("Name(n) ändern"));
        m_deleteAction->setText(tr("Paar löschen"));
    }

    m_removeMarkerAction->setEnabled(selectedMark != 0);
    const auto setMarkerMenuActions = m_setMarkerMenu->actions();
    for (QAction *action : setMarkerMenuActions) {
        action->setEnabled(action->data().toInt() != selectedMark);
    }

    const int assignedTable = m_allDraws.at(selectedItem->row()).table;
    m_setDrawAction->setVisible(assignedTable == 0);
    m_deleteDrawAction->setVisible(assignedTable != 0);

    m_contextMenu->exec(m_playersList->mapToGlobal(point));
}

void RegistrationPage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_playersList->rowCount(); row++) {
            m_playersList->setRowHidden(row, ! isRowVisible(row));
        }
        m_playersList->resizeColumnsToContents();
        m_isSearchLabel->hide();
        m_adoptSearchText->hide();
        return;
    }

    int matches = 0;
    int hiddenMatches = 0;
    m_searchEngine->setSearchTerm(searchText);

    const bool phoneticSearch = m_searchWidget->phoneticSearch();
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        bool rowMatches = m_searchEngine->checkMatch(m_allRegistrations.at(row),
                                                     SearchEngine::SearchType::Default,
                                                     phoneticSearch);
        if (rowMatches && ! isRowVisible(row)) {
            rowMatches = false;
            hiddenMatches++;
        }

        m_playersList->setRowHidden(row, ! rowMatches);
        matches += rowMatches;
    }

    m_searchWidget->updateMatches(matches, hiddenMatches);
    m_playersList->resizeColumnsToContents();
    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText.simplified()));
    m_isSearchLabel->show();
    m_adoptSearchText->setVisible(matches == 0 && hiddenMatches == 0);
}

void RegistrationPage::listModified()
{
    m_searchWidget->saveSearch();
    updatePlayersList();
    m_searchWidget->restoreSearch();
}

void RegistrationPage::resetSearch()
{
    m_searchWidget->clear();
}

void RegistrationPage::displayMarkedEntry()
{
    const int oldRow = m_playersList->currentItem()->row();
    listModified();

    const auto matches = m_playersList->findItems(m_selectedName, Qt::MatchExactly);
    if (matches.isEmpty()) {
        itemError(Q_FUNC_INFO, m_selectedName);
        return;
    }

    const int newRow = matches.at(0)->row();
    if (newRow != oldRow && isRowVisible(newRow)) {
        scrollTo(m_selectedName);
    }
}

void RegistrationPage::updateMarker(const QString &markerName, int markerId)
{
    if (m_isClient) {
        m_client->requestUpdateMarker(m_selectedName, markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (markerId != 0) {
        emit statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Markiere Paar „%1“ als „%2“ …")
                             : tr("Markiere Spieler „%1“ als „%2“ …")).arg(
                          m_selectedName, markerName));
    } else {
        emit statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Entferne Markierung von Paar „%2“ …")
                             : tr("Entferne Markierung von Spieler „%2“ …")).arg(
                          m_selectedName));
    }

    if (! m_db->updateMarker(m_selectedName, markerId)) {
        return;
    }

    if (markerId != 0) {
        emit statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Paar „%1“ als „%2“ markiert")
                             : tr("Spieler „%1“ als „%2“ markiert")).arg(
                          m_selectedName, markerName));
    } else {
        emit statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Markierung von Paar „%2“ entfernt")
                             : tr("Markierung von Spieler „%2“ entfernt")).arg(
                          m_selectedName));
    }
    displayMarkedEntry();

    if (m_isServer) {
        m_server->broadcastUpdateMarker(m_selectedName, markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::listUpdateMarker(const QString &markerName, int markerId)
{
    if (m_isClient) {
        m_client->requestListUpdateMarker(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    if (markerId != 0) {
        emit statusUpdate(tr("Markiere alle %1 als „%2“").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                             markerName));
    } else {
        emit statusUpdate(tr("Entferne alle Markierungen …"));
    }

    if (! m_db->listUpdateMarker(markerId)) {
        return;
    }

    if (markerId != 0) {
        emit statusUpdate(tr("Alle %1 als „%2“ markiert").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                             markerName));
    } else {
        emit statusUpdate(tr("Alle Markierungen entfernt"));
    }
    listModified();

    if (m_isServer) {
        m_server->broadcastListUpdateMarker(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::listRemoveMarker(const QString &markerName, int markerId)
{
    if (m_isClient) {
        m_client->requestListRemoveMarker(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Entferne alle „%1“-Markierungen …").arg(markerName));

    if (! m_db->listRemoveMarker(markerId)) {
        return;
    }

    emit statusUpdate(tr("Alle „%1“-Markierungen entfernt").arg(markerName));
    listModified();

    if (m_isServer) {
        m_server->broadcastListRemoveMarker(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::deletePlayers()
{
    if (QMessageBox::question(this,
            m_tournamentSettings->isPairMode() ? tr("Paar löschen") : tr("Spieler löschen"),
            (m_tournamentSettings->isPairMode()
                ? tr("Soll Paar „%1“ wirklich gelöscht werden?")
                : tr("Soll Spieler „%1“ wirklich gelöscht werden?")).arg(
            m_selectedName),
            QMessageBox::Yes | QMessageBox::Cancel,
            QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    if (m_isClient || m_isServer) {
        if (m_playersList->findItems(m_selectedName, Qt::MatchExactly).count() != 1) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("%1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer entweder "
                   "gelöscht oder umbenannt. Das Löschen ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
                   m_selectedName));
            return;
        }
    }

    if (m_isClient) {
        m_client->requestDeletePlayers(m_selectedName);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_tournamentSettings->isPairMode()
                          ? tr("Lösche Paar „%1“ …") : tr("Lösche Spieler „%1“ …")).arg(
                       m_selectedName));

    if (! m_db->deletePlayers(m_selectedName)) {
        return;
    }

    emit statusUpdate((m_tournamentSettings->isPairMode()
                          ? tr("Paar „%1“ gelöscht") : tr("Spieler „%1“ gelöscht")).arg(
                      m_selectedName));

    listModified();
    checkUpdateDrawOverviewPage(true, m_selectedName, true);

    if (m_isServer) {
        m_server->broadcastDeletePlayers(m_selectedName);
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::cacheChecksum()
{
    if (m_isServer || m_isClient) {
        m_cachedChecksum = m_checksum;
    }
}

bool RegistrationPage::checkCachedChecksum(const QString &title)
{
    if (! m_isClient && ! m_isServer) {
        return true;
    }

    if (m_checksum != m_cachedChecksum) {
        QMessageBox::warning(this, title,
            tr("<p><b>Kollidierende Netzwerkänderung</b></p>"
               "<p>Die Anmeldungsliste wurde zwischenzeitlich von einem anderen Netzwerkteilnehmer "
               "verändert. Die Aktion kann nicht mehr durchgeführt werden und wird abgebrochen."
               "</p>"));
        return false;
    }

    return true;
}

bool RegistrationPage::confirmListDeleteAction(const QString &title, const QString &text)
{
    m_backupFile.clear();

    QMessageBox messageBox(QMessageBox::Warning, title, text,
                           QMessageBox::Yes | QMessageBox::Cancel, this);
    messageBox.setDefaultButton(QMessageBox::Cancel);
    QCheckBox *createBackup = new QCheckBox(tr("Vorher ein Backup der Datenbank anlegen"));
    createBackup->setChecked(true);
    messageBox.setCheckBox(createBackup);

    if (messageBox.exec() == QMessageBox::Cancel) {
        return false;
    }

    if (createBackup->isChecked()) {
        const QString openedDb = m_db->dbFile();
        BackupEngine backupEngine(openedDb);
        if (! m_db->saveSettings()) {
            return false;
        }
        m_db->closeDatabaseConnection();
        bool success = backupEngine.createBackup();
        m_db->openDatabase(openedDb);

        if (! success) {
            QMessageBox::critical(this, title,
                tr("<p><b>Das Anlegen eines Backups ist fehlgeschlagen!</b></p>"
                   "<p>Die Datei konnte nicht angelegt werden. Bitte die Zugriffsrechte von „%1“ "
                   "bzw. dem Verzeichnis „%2“) überprüfen!</p>").arg(
                   backupEngine.backupFile(), QDir::toNativeSeparators(backupEngine.path())));
            return false;
        }

        m_backupFile = backupEngine.backupFile();
    }

    return true;
}

void RegistrationPage::checkDisplayBackup(const QString &title)
{
    if (! m_backupFile.isEmpty()) {
        QMessageBox::information(this, title,
            tr("Es wurde erfolgreich ein Backup der Datenbank mit dem Dateinamen „%1“ "
               "erstellt.").arg(m_backupFile));
    }
}

void RegistrationPage::deleteMarked(const QString &markerName, int markerId)
{
    cacheChecksum();
    const QString title = tr("Markierte %1 löschen").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
            tr("Sollen wirklich alle als „%1“ markierten %2 gelöscht werden?").arg(markerName,
               m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")))
        || ! checkCachedChecksum(title)) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteMarkedPlayers(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle als „%1“ markierten %2 …").arg(
                         markerName, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                        : tr("Spieler")));
    if (! m_db->deleteMarkedPlayers(markerId)) {
        return;
    }

    emit statusUpdate(tr("Alle als „%1“ markierten %2 gelöscht").arg(
                         markerName, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                        : tr("Spieler")));

    listModified();
    checkUpdateDrawOverviewPage(true);

    if (m_isServer) {
        m_server->broadcastDeleteMarkedPlayers(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void RegistrationPage::deleteUnmarked()
{
    cacheChecksum();
    const QString title = tr("Unmarkierte %1 löschen").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
            tr("Sollen wirklich alle unmarkierten %1 gelöscht werden?").arg(
               m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")))
        || ! checkCachedChecksum(title)) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteUnmarkedPlayers();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle unmarkierten %1 …").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteUnmarkedPlayers()) {
        return;
    }

    emit statusUpdate(tr("Alle unmarkierten %1 gelöscht").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    listModified();
    checkUpdateDrawOverviewPage(true);

    if (m_isServer) {
        m_server->broadcastDeleteUnmarkedPlayers();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void RegistrationPage::deleteAll()
{
    const QString title = tr("Alle %1 löschen").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
              tr("Sollen wirklich alle %1 gelöscht werden?").arg(
                 m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")))) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteAllPlayers();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle %1 …").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteAllPlayers()) {
        return;
    }

    emit statusUpdate(tr("Alle %1 gelöscht").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    listModified();
    checkUpdateDrawOverviewPage(true);

    if (m_isServer) {
        m_server->broadcastDeleteAllPlayers();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void RegistrationPage::markersVisibilityChanged()
{
    m_searchWidget->saveSearch();
    updateVisibility();
    m_searchWidget->restoreSearch();
}

void RegistrationPage::updateVisibility()
{
    int hidden = 0;
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        bool visible = isRowVisible(row);
        m_playersList->setRowHidden(row, ! visible);
        if (! visible) {
            hidden++;
        }
    }
    if (hidden > 0) {
        m_playersBox->setTitle(m_tournamentSettings->isPairMode()
                                   ? tr("Angemeldete Paare (%n ausgeblendet)", "", hidden)
                                   : tr("Angemeldete Spieler (%n ausgeblendet)", "", hidden));
    } else {
        m_playersBox->setTitle(m_tournamentSettings->isPairMode() ? tr("Angemeldete Paare")
                                                                  : tr("Angemeldete Spieler"));
    }
    m_playersList->resizeColumnsToContents();
}

bool RegistrationPage::isRowVisible(int row) const
{
    return m_markersWidget->isMarkerVisible(m_allSetMarkers.at(row));
}

void RegistrationPage::setColumnVisible(Column column)
{
    const bool state = m_showColumn.value(column)->isChecked();
    m_playersList->setColumnHidden(COLUMN_INDEX.value(column), state == Qt::Unchecked);
    m_playersList->resizeColumnsToContents();

    if (column == Column::Draw && m_db->isOpen()) {
        // We need to call this to update the status
        updatePlayersList();
    }

    // Inform the database
    QVector<Column> columns;
    for (const Column column : ALL_COLUMNS) {
        if (m_showColumn.value(column)->isChecked()) {
            columns.append(column);
        }
    }
    m_tournamentSettings->setRegistrationColumns(columns);

    // If the draw column is shown, ask if the draw overview page should also be shown
    if (column == Column::Draw && state
        && ! m_defeatableMessageBox->dontAsk(ASK_SHOW_DRAW_OVERVIEW)
        && ! m_tournamentSettings->loadingDb()
        && m_drawOverviewPage == nullptr
        && m_defeatableMessageBox->show(ASK_SHOW_DRAW_OVERVIEW, this,
               tr("Spalte „Auslosung“ anzeigen"),
               tr("Soll die Auslosungs-Übersichts-Seite ebenfalls angezeigt werden?"))
           == QMessageBox::Yes) {

        emit showDrawOverviewPage();
        QTimer::singleShot(0, qobject_cast<QWidget *>(parent()), &QWidget::raise);
    }
}

void RegistrationPage::updateChecksum()
{
    ChecksumHelper checksumHelper;

    // Add the players data

    for (int row = 0; row < m_playersList->rowCount(); row++) {
        // Name
        checksumHelper.addData(m_allRegistrations.at(row));

        // Marker
        checksumHelper.addData(m_allSetMarkers.at(row));

        // Draw
        checksumHelper.addData(m_allDraws.at(row));
    }

    // Add the marker data
    m_markersWidget->addChecksumData(checksumHelper);

    // Add the tournament state
    checksumHelper.addData(m_db->networkTournamentStarted());

    m_checksum = checksumHelper.checksum();

    emit checksumChanged(m_checksum);
}

void RegistrationPage::setServer(Server *server)
{
    updateChecksum();
    m_server = server;
    m_isServer = true;
    m_markersWidget->setServer(server);
    connect(server, &Server::nameEdited, this, &RegistrationPage::nameEdited);
}

void RegistrationPage::serverStopped()
{
    m_isServer = false;
    m_markersWidget->serverStopped();
    markersChanged();
    m_checksum.clear();
}

void RegistrationPage::hideContextMenus()
{
    const auto contextMenus = findChildren<QMenu *>();
    for (QMenu *menu : contextMenus) {
        menu->hide();
    }
}

int RegistrationPage::visibleRow(const QTableWidgetItem *item) const
{
    const auto *firstItem = m_playersList->itemAt(0, 0);
    if (firstItem == nullptr) {
        return -1;
    }

    const int targetRow = item->row();
    const int firstRow = firstItem->row();

    int lastRow = m_playersList->rowCount();
    const int viewportHeight = m_playersList->viewport()->height();
    const auto *lastItem = m_playersList->itemAt(0, viewportHeight - 1);
    if (lastItem != nullptr) {
        const bool completelyVisible = m_playersList->visualItemRect(lastItem).bottomLeft().y()
                                       <= viewportHeight;
        lastRow = lastItem->row() + completelyVisible;
    }

    int visibleRow = 0;
    for (int row = firstRow; row < lastRow; row++) {
        if (m_playersList->item(row, 0)->row() == targetRow) {
            return visibleRow;
        }
        visibleRow += isRowVisible(row);
    }

    return -1;
}

void RegistrationPage::requestRejected()
{
    if (m_requestingRegistration) {
        m_requestingRegistration = false;
    }
}

void RegistrationPage::processNetworkChange(int clientId)
{
    // Clear the name(s) lineedit if we requested a registration and it did not fail
    if (m_isClient && m_requestingRegistration && clientId == m_client->id()) {
        m_playersName->clear();
        m_requestingRegistration = false;
    }

    const bool contextMenuVisible = m_contextMenu->isVisible();
    const bool listMenuVisible = m_listMenu->isVisible();
    const bool editorOpen = m_delegate->isEditing();
    const bool markersPopupVisible = m_markersPopup->isVisible();
    const bool drawPopupVisible = m_drawPopup->isVisible();

    int originalVisibleRow = -1;
    int verticalOffset = 0;
    int originalY = 0;

    const auto *currentItem = m_playersList->currentItem();
    if (currentItem != nullptr) {
        originalVisibleRow = visibleRow(currentItem);
        originalY = m_playersList->mapToGlobal(
                        m_playersList->visualItemRect(currentItem).topLeft()).y();
    }

    // Check if we're doing anything. If not, we can simply display the change (possibly our own)

    if (! contextMenuVisible && ! listMenuVisible && ! editorOpen && ! markersPopupVisible
        && ! drawPopupVisible) {

        m_searchWidget->saveSearch();
        updatePlayersList();

        if (m_isClient && clientId == m_client->id() && ! m_scrollToTarget.isEmpty()) {
            const auto matches = m_playersList->findItems(m_scrollToTarget, Qt::MatchExactly);
            if (matches.isEmpty()) {
                itemError(Q_FUNC_INFO, m_scrollToTarget);
                return;
            }

            const auto *processedItem =  matches.at(0);
            const int processedVisibleRow = visibleRow(processedItem);

            if (processedItem != nullptr
                && (processedVisibleRow == -1 || processedVisibleRow != originalVisibleRow)) {

                // The last selected row doesn't match the row of the scrollTarget, so we want to
                // be sure that the item is shown and scroll to it
                scrollTo(m_scrollToTarget);
            }

            // Otherwise, we can assume that the item is inside the viewport anyway.
        }

        m_searchWidget->restoreSearch();

        if (m_focusPlayersName) {
            m_focusPlayersName = false;
            m_playersName->setFocus();
        }

        return;
    }

    // Some action is currently executed, so we want to restore it

    // Cache what we have
    // ------------------

    const auto processedText = editorOpen ? m_delegate->editedText() : m_selectedName;

    QPoint listMenuPos;
    if (listMenuVisible) {
        // Process this first so that the list menu will be restored if it's visible,
        // even if it was shown by using the context menu's list menu action
        listMenuPos = m_listMenu->pos();
    }

    QPoint contextMenuPos;
    if (contextMenuVisible) {
        contextMenuPos = m_playersList->mapFromGlobal(m_contextMenu->pos());
    }

    QPoint popupPos;

    if (markersPopupVisible) {
        popupPos = m_markersPopup->pos();
    }

    if (drawPopupVisible) {
        popupPos = m_drawPopup->pos();
    }

    // Abort all actions
    // -----------------

    if (contextMenuVisible || listMenuVisible) {
        hideContextMenus();
    }
    if (editorOpen) {
        m_delegate->saveEditorState();
    }

    // Display the change
    // ------------------

    listModified();

    // Restore what we had
    // -------------------

    // Restore the "whole list" menu (we don't need a specific item to still exist for this one)
    if (listMenuVisible) {
        if (m_playersList->rowCount() > 0) {
            QTimer::singleShot(0,
                [this, listMenuPos]
                {
                    m_listMenu->exec(listMenuPos);
                });
        } else {
            QTimer::singleShot(0,
                [this]
                {
                    QMessageBox::warning(this,
                        tr("Kollidierende Netzwerkänderung"),
                        tr("Es wurden von einem anderen Netzwerkteilnehmer alle Anmeldungen "
                           "gelöscht. Ein Bearbeiten der ganzen Liste ist nicht mehr möglich."));
                });
        }
        return;
    }

    // From here on, we need the processed item

    const auto matches = m_playersList->findItems(processedText, Qt::MatchExactly);
    if (matches.isEmpty()) {
        itemError(Q_FUNC_INFO, processedText);
        return;
    }

    auto *processedItem = matches.at(0);

    if (! isRowVisible(processedItem->row())) {
        if (m_fuzzyMatchWarning != nullptr) {
            m_fuzzyMatchWarning->reject();
        }

        QTimer::singleShot(0, this, [this, processedText]
        {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("<p>%1 „%2“ wurde durch eine Änderung eines anderen Netzwerkteilnehmers "
                   "entweder umbenannt, gelöscht, oder der Eintrag ist jetzt ausgeblendet.</p>"
                   "<p>Das Bearbeiten wird abgebrochen.</p>").arg(
                   m_tournamentSettings->isPairMode() ? tr("Der gerade bearbeitete Spieler")
                                                      : tr("Das gerade bearbeitete Paar"),
                   processedText));
        });
        return;
    }

    // Re-select the item
    m_playersList->setCurrentItem(processedItem);

    const int processedVisibleRow = visibleRow(processedItem);

    // Try to scroll so that it's at the same place on screen
    if (processedVisibleRow != originalVisibleRow) {
        m_playersList->verticalScrollBar()->setValue(
            m_playersList->verticalScrollBar()->value() - originalVisibleRow + processedVisibleRow);
        verticalOffset = m_playersList->mapToGlobal(
                             m_playersList->visualItemRect(processedItem).topLeft()).y()
                         - originalY;

        // Also move the mouse cursor
        auto cursorPos = QCursor::pos();
        cursorPos.setY(cursorPos.y() + verticalOffset);
        QCursor::setPos(cursorPos);
    }

    // Restore the editor
    if (editorOpen) {
        QTimer::singleShot(0, this, [this, processedItem]
        {
            m_playersList->editItem(processedItem);
            m_delegate->restoreEditorState();
        });
        return;
    }

    // Restore the context menu

    if (contextMenuVisible) {
        contextMenuPos.setY(contextMenuPos.y() + verticalOffset);

        const auto *iteamAtContextMenuPos = m_playersList->itemAt(contextMenuPos);
        if (iteamAtContextMenuPos == nullptr
            || m_playersList->item(iteamAtContextMenuPos->row(), COLUMN_INDEX.value(Column::Name))
               != processedItem) {

            // This can happen if the context menu has been displayed from bottom to top and now,
            // it can be displayed from top to bottom or vice versa. In this case, display the
            // context menu at the vertical middle of the processed item
            contextMenuPos.setY(m_playersList->visualItemRect(processedItem).topLeft().y()
                                + m_playersList->rowHeight(processedVisibleRow) / 2);
        }

        QTimer::singleShot(0, this, [this, contextMenuPos]
        {
            showContextMenu(contextMenuPos);
        });
        return;
    }

    // Restore the markers popup
    if (markersPopupVisible) {
        popupPos.setY(popupPos.y() + verticalOffset);
        QTimer::singleShot(0, this, [this, processedItem, popupPos, processedVisibleRow]
        {
            highlightItem(processedItem);
            m_markersPopup->show(m_allSetMarkers.at(processedVisibleRow), popupPos);
        });
        return;
    }

    // Restore the draw popup

    if (drawPopupVisible) {
        const auto currentSeat = m_allDraws.at(processedVisibleRow);
        popupPos.setY(popupPos.y() + verticalOffset);
        QTimer::singleShot(0, this,
            [this, processedItem, currentSeat, popupPos, processedVisibleRow]
            {
                highlightItem(processedItem);
                const int marker = m_allSetMarkers.at(processedVisibleRow);
                const bool empty = currentSeat.isEmpty();
                m_drawPopup->show(empty ? m_drawPopup->seat() : currentSeat, empty,
                                  m_db->allRoundsFinished(),
                                  marker == m_markersWidget->singleMarker(), popupPos);
            });
        return;
    }
}

void RegistrationPage::setClient(Client *client)
{
    m_isClient = true;
    m_client = client;
    m_markersWidget->setClient(client);
    connect(client, &Client::nameEdited, this, &RegistrationPage::nameEdited);
}

void RegistrationPage::clientDisconnected()
{
    m_isClient = false;
    m_markersWidget->clientDisconnected();
    markersChanged();
    m_checksum.clear();
    m_requestingRegistration = false;
}

void RegistrationPage::setProcessingRequest(bool processing, const QString &scrollToTarget)
{
    m_newPlayersBox->setEnabled(! processing);
    m_playersBox->setEnabled(! processing);
    if (processing) {
        m_scrollToTarget = scrollToTarget;
        QApplication::setOverrideCursor(Qt::WaitCursor);
    }
}

MarkersWidget *RegistrationPage::markersWidget() const
{
    return m_markersWidget;
}

void RegistrationPage::finishRegistration()
{
    if (QMessageBox::warning(this,
        tr("Anmeldung beenden"),
        tr("<p>Wenn die Anmeldung beendet ist, kann (mindestens während das Netzwerk läuft) kein "
           "Netzwerkteilnehmer mehr neue %1 anmelden oder löschen!</p>"
           "<p>Soll die Anmeldung wirklich beendet werden?</p>").arg(
           m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")),
        QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel) == QMessageBox::Cancel) {

        return;
    }

    if (m_isClient) {
        m_client->requestFinishRegistration();
        return;
    }

    // Otherwise, we're the server
    networkFinishRegistration();
    m_server->broadcastFinishRegistration();
}

void RegistrationPage::networkFinishRegistration(int clientId)
{
    QString networkChange;
    if ((m_isClient && clientId != m_client->id()) || (m_isServer && clientId != 0)) {
        networkChange = tr(" von einem anderen Netzwerkteilnehmer");
        if (m_contextMenu->isVisible()) {
            hideContextMenus();
        }
        if (m_delegate->isEditing()) {
            m_delegate->abortEdit();
        }
    }

    m_db->setNetworkTournamentStarted(true);
    reload();
    emit networkTournamentStarted(true);

    QTimer::singleShot(0, this, [this, networkChange]
    {
        QMessageBox::information(this,
            tr("Anmeldung beendet"),
            tr("<p>Die Anmeldung wurde%1 beendet. Es können ab jetzt keine neuen %2 mehr "
               "angemeldet oder bestehende Anmeldungen gelöscht werden.</p>"
               "<p>Sollte das trotzdem noch nötig sein, dann bitte den Server stoppen, neu "
               "starten und alle Clients neu verbinden. Sofern noch keine Ergebnisse eingegeben "
               "wurden, ist danach das komplette Bearbeiten der Anmeldungsliste wieder möglich."
               "</p>").arg(networkChange, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                             : tr("Spieler")));
    });
}

void RegistrationPage::checkShowPopup(int row, int column)
{
    if ((column != COLUMN_INDEX.value(Column::Mark) && column != COLUMN_INDEX.value(Column::Draw))
        || ! m_db->isChangeable()) {

        return;
    }

    QPoint position(-1, -1);
    if (column == COLUMN_INDEX.value(Column::Draw)
        && m_playersList->isColumnHidden(COLUMN_INDEX.value(Column::Draw))) {

        position = QCursor::pos();
        askShowDrawColumn();
    }

    QTableWidgetItem *item = m_playersList->item(row, COLUMN_INDEX.value(Column::Name));
    m_selectedName = item->text();
    highlightItem(item);

    if (column == COLUMN_INDEX.value(Column::Mark)) {
        m_markersPopup->show(m_allSetMarkers.at(item->row()));

    } else if (column == COLUMN_INDEX.value(Column::Draw)) {
        const auto &seat = m_allDraws.at(row);
        const int marker = m_allSetMarkers.at(row);

        if (marker == m_tournamentSettings->marker(Markers::Absent)
            && m_allDraws.at(row).isEmpty()) {

            QMessageBox::warning(this,
                tr("Auslosung eingeben"),
                tr("<p><b>Auslosung eingeben</b></p>")
                + (m_tournamentSettings->isPairMode()
                       ? tr("<p>Das Paar „%1“ ist momentan als abwesend markiert. Für ein "
                            "abwesendes Paar kann keine Auslosung eingegeben werden."
                            "</p>").arg(m_selectedName)
                       : tr("<p>Der Spieler „%1“ ist momentan als abwesend markiert. Für einen "
                            "abwesenden Spieler kann keine Auslosung eingegeben "
                            "werden.").arg(m_selectedName)));
            resetHighlight();
            return;
        }

        m_drawPopup->show(seat, seat.isEmpty(), m_db->allRoundsFinished(),
                                marker == m_markersWidget->singleMarker(), position);
    }
}

void RegistrationPage::askShowDrawColumn()
{
    if (m_defeatableMessageBox->dontAsk(ASK_SHOW_DRAW_COLUMN)
        || m_showColumn.value(Column::Draw)->isChecked()) {

        return;
    }

    if (m_defeatableMessageBox->show(ASK_SHOW_DRAW_COLUMN, this,
            tr("Auslosung speichern"),
            tr("Momentan ist die Spalte „Auslosung“ auf der Anmeldeseite ausgeblendet. Soll sie "
               "angezeigt werden?")) == QMessageBox::Yes) {

        m_showColumn[Column::Draw]->setChecked(true);
    }
}

void RegistrationPage::highlightItem(QTableWidgetItem *item)
{
    m_lastForeground = item->foreground();
    m_lastBackground = item->background();
    item->setForeground(m_playersList->palette().highlightedText());
    item->setBackground(m_playersList->palette().highlight());
    m_highlightedItem = item;
}

void RegistrationPage::resetHighlight()
{
    if (m_highlightedItem == nullptr) {
        return;
    }

    m_highlightedItem->setForeground(m_lastForeground);
    m_highlightedItem->setBackground(m_lastBackground);
    m_highlightedItem = nullptr;
}

void RegistrationPage::updateDraw(const Draw::Seat &seat)
{
    if (seat.table >= DefaultValues::DRAW_TABLE_WARNING
        && QMessageBox::question(this,
               tr("Auslosung speichern"),
               tr("Soll wirklich eine Auslosung für Tisch Nr. %1 eingegeben werden (vielleicht "
                  "war es ein Tippfehler)?").arg(QString::number(seat.table)),
               QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel)
        == QMessageBox::Cancel) {

        return;
    }

    const int marker = seat.isEmpty() ? -1 : m_markersWidget->drawnMarker();

    if (m_isClient) {
        m_client->requestUpdateDraw(m_selectedName, seat, marker);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(seat.isEmpty() ? tr("Lösche Auslosung …") : tr("Speichere Auslosung …"));

    if (! m_db->updateDraw(m_selectedName, seat, marker)) {
        return;
    }

    emit drawChanged();
    if (marker != -1) {
        displayMarkedEntry();
    } else {
        listModified();
    }

    checkUpdateDrawOverviewPage(true, m_selectedName, seat.isEmpty());

    if (m_isServer) {
        m_server->broadcastUpdateDraw(m_selectedName, seat, marker);
    }

    if (seat.table == 0) {
        emit statusUpdate(tr("Auslosung für %1 „%2“ gelöscht").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                             m_selectedName));
    } else {
        emit statusUpdate(tr("Auslosung für %1 „%2“ gespeichert (Tisch %3 Paar %4)").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                             m_selectedName, QString::number(seat.table),
                             QString::number(seat.pair)));
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::assignPlayer(const QString &name)
{
    const QString pairName = m_selectedName + m_settings->namesSeparator() + name;
    if (checkForExactMatch(pairName)) {
        QMessageBox::warning(this, tr("Allein gekommene Spieler zuordnen"),
            tr("<p><b>Zuordnung nicht möglich</b></p>"
               "<p>Der Spieler „%1“ kann dem Spieler „%2“ nicht zugeordnet werden, da der Paarname "
               "„%3“ bereits vergeben ist.</p>"
               "<p>Bitte mindestestens einen Namen ändern, damit ein eindeutiger Paarname erstellt "
               "werden kann!</p>").arg(
               name, m_selectedName, pairName));
        return;
    }

    if (m_isClient) {
        m_client->requestAssemblePair(m_selectedName, name, m_settings->namesSeparator(),
                                      m_markersWidget->assignedMarker());
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Erstelle Paar aus Einzelspielern …"));

    if (! m_db->assemblePair(m_selectedName, name, m_settings->namesSeparator(),
                             m_markersWidget->assignedMarker())) {
        return;
    }

    listModified();
    checkUpdateDrawOverviewPage(true, m_selectedName + m_settings->namesSeparator() + name);

    const auto matches = m_playersList->findItems(pairName, Qt::MatchExactly);
    if (matches.isEmpty()) {
        itemError(Q_FUNC_INFO, pairName);
        return;
    }

    auto *item = matches.at(0);
    if (item->row() != m_editedRow) {
        scrollTo(pairName);
    }

    if (m_isServer) {
        m_server->broadcastAssemblePair(m_selectedName, name, m_settings->namesSeparator(),
                                        m_markersWidget->assignedMarker());
    }

    emit statusUpdate(tr("Paar „%1“ erstellt").arg(pairName));
    QApplication::restoreOverrideCursor();
}

bool RegistrationPage::absentsPresent() const
{
    return m_absentsPresent;
}

bool RegistrationPage::singlesPresent() const
{
    return m_singlesPresent;
}

void RegistrationPage::deleteAllDraws()
{
    const QString title = tr("Alle Auslosungen löschen");
    if (! confirmListDeleteAction(title, tr("Sollen wirklich alle bisher eingegebenen Auslosungen "
                                            "gelöscht werden?"))) {
        return;
    }

    if (m_isClient) {
        m_client->requestDeleteAllDraws();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle Auslosungen …"));

    if (! m_db->deleteAllDraws()) {
        return;
    }

    emit drawChanged();
    emit statusUpdate(tr("Alle Auslosungen gelöscht"));
    listModified();
    checkUpdateDrawOverviewPage(true);

    if (m_isServer) {
        m_server->broadcastDeleteAllDraws();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void RegistrationPage::processStatusLinkClick(const QString &anchor)
{
    if (anchor == QLatin1String("#drawOverview")) {
        emit showDrawOverviewPage();
        return;
    }

    if (anchor == QLatin1String("#adjustDraw")) {
        emit showAdjustDrawDialog();
        return;
    }

    QString text;

    if (anchor == QLatin1String("#unfinished")) {
        text = tr("<p><b>Unvollständige Auslosung</b></p>");

        QString list;
        int missingDraws = 0;
        for (int row = 0; row < m_playersList->rowCount(); row++) {
            if (m_playersList->item(row, COLUMN_INDEX.value(Column::Draw))->text().isEmpty()) {
                missingDraws++;
                if (missingDraws <= 10) {
                    list.append(QStringLiteral("<li>%1</li>").arg(
                        m_playersList->item(row, COLUMN_INDEX.value(Column::Name))->text()));
                }
            }
        }

        text.append(m_tournamentSettings->isPairMode()
            ?  tr("<p>Das/die folgende(n) Paar(e) hat/haben noch keine Auslosung:</p><ul>%1</ul>",
                  "", missingDraws).arg(list)
            :  tr("<p>Der/die folgende(n) Spieler hat/haben noch keine Auslosung:</p><ul>%1</ul>",
                  "", missingDraws).arg(list));

        if (missingDraws > 10) {
            text.append(m_tournamentSettings->isPairMode()
                ? tr("<p>und %n weitere(s)</p>", "für Paare", missingDraws - 10)
                : tr("<p>und %n weitere(r)</p>", "für Spieler", missingDraws - 10));
        }

    } else if (anchor == QLatin1String("#incompleteTables")) {
        text = tr("<p><b>Tische mit unvollständiger Auslosung</b></p>");
        const int incompleteCount = m_incompleteTables.count();
        if (incompleteCount == 1) {
            text.append(tr("<p>Zu Tisch <b>%1</b> wurde bisher nur ein Paar 1 oder ein Paar 2 "
                           "zugelost%2.</p>").arg(
                           QString::number(m_incompleteTables.at(0)),
                           m_tournamentSettings->isPairMode()
                               ? QString() : tr(" (bzw. die Paare sind unvollständig)")));
        } else if (incompleteCount == 2) {
            text.append(tr("<p>Zu den Tischen <b>%1</b> und <b>%2</b> wurden bisher nur ein Paar 1 "
                           "oder ein Paar 2 zugelost%3.</p>").arg(
                           QString::number(m_incompleteTables.at(0)),
                           QString::number(m_incompleteTables.at(1)),
                           m_tournamentSettings->isPairMode()
                               ? QString() : tr(" (bzw. die Paare sind unvollständig)")));
        } else {
            text.append(tr("<p>Zu folgenden Tischen wurden bisher nur ein Paar 1 oder ein Paar 2 "
                           "zugelost%2:</p><p>").arg(
                           m_tournamentSettings->isPairMode()
                               ? QString() : tr(" (bzw. die Paare sind unvollständig)")));
            for (int i = 0; i < incompleteCount - 1; i++) {
                text.append(tr("<b>%1</b>, ").arg(QString::number(m_incompleteTables.at(i))));
            }
            text = text.left(text.size() - 2);
            text.append(tr(" und <b>%1</b>.</p>").arg(m_incompleteTables.at(incompleteCount - 1)));
        }

    } else if (anchor == QLatin1String("#tableGaps")) {
        text = tr("<p><b>Tische, die in der Auslosung fehlen</b></p>");
        const int gapsCount = m_tableGaps.count();
        if (gapsCount == 1) {
            text.append(tr("<p>Tisch <b>%1</b> ist bisher nicht ausgelost, müsste es aber sein (da "
                           "es sonst Lücken gibt)</p>").arg(
                           QString::number(m_tableGaps.at(0))));
        } else if (gapsCount == 2) {
            text.append(tr("<p>Die Tische <b>%1</b> und <b>%2</b> sind bisher nicht zugewiesen, "
                           "müssten es aber sein (da es sonst Lücken gibt).</p>").arg(
                           QString::number(m_tableGaps.at(0)),
                           QString::number(m_tableGaps.at(1))));
        } else {
            text.append(tr("<p>Folgende Tische sind bisher nicht ausgelost, müssten es aber "
                           "sein (da es sonst Lücken gibt):</p><p>"));
            for (int i = 0; i < gapsCount - 1; i++) {
                text.append(tr("<b>%1</b>, ").arg(QString::number(m_tableGaps.at(i))));
            }
            text = text.left(text.size() - 2);
            text.append(tr(" und <b>%1</b>.</p>").arg(m_tableGaps.at(gapsCount - 1)));
        }
    }

    QMessageBox::information(this, tr("Auslosung"), text);
}

void RegistrationPage::setDrawOverviewPage(DrawOverviewPage *drawOverviewPage)
{
    m_drawOverviewPage = drawOverviewPage;
    updatePlayersList();
}

void RegistrationPage::checkUpdateDrawOverviewPage(bool ownChange, const QString &target,
                                                   bool deleted) const
{
    if (m_drawOverviewPage != nullptr) {
        m_drawOverviewPage->updateDraw(ownChange, target, deleted);
    }
}

void RegistrationPage::setColumnsVisible(const QVector<Column> &columns)
{
    // We have to call this via a QTimer call if the database is opened via command line argument
    // to be sure that everything is constructed and set up before we do this
    QTimer::singleShot(0, this, [this, columns]
    {
        for (const Column column : ALL_COLUMNS) {
            m_showColumn[column]->setChecked(columns.contains(column));
        }
    });
}

void RegistrationPage::showDrawSettings()
{
    auto dialog = DrawSettingsDialog(this);
    dialog.updateSettings(m_tournamentSettings->drawSettings());
    if (dialog.exec()) {
        m_tournamentSettings->setDrawSettings(dialog.settings());
        listModified();
        if (m_playersList->isColumnHidden(COLUMN_INDEX.value(Column::Draw))) {
            askShowDrawColumn();
        }
    }
}

void RegistrationPage::drawAllSeats()
{
    cacheChecksum();

    if (m_playersList->isColumnHidden(COLUMN_INDEX.value(Column::Draw))) {
        askShowDrawColumn();
    }

    const int tablesLimit = m_tournamentSettings->drawSettings().tablesLimit;
    const int neededTables = m_allRegistrations.count() / m_db->necessaryDivisor();
    if (tablesLimit != -1 && neededTables > tablesLimit) {
        // The number of registrations would need more tables than the maximum number of tables
        // that has been defined in the draw settings
        const QString title = tr("Alle Plätze auslosen");
        if (QMessageBox::warning(this, title,
                tr("<p><b>Alle Plätze auslosen</b></p>"
                   "<p>Die Anzahl der Tische, die momentan für eine komplette Auslosung benötigt "
                   "werden (%1), ist höher, als das Tisch-Limit, das in den Einstellungen für die "
                   "Auslosung festgelegt wurde (%2)!</p>"
                   "<p>Sollen wirklich alle Plätze ausgelost werden?</p>").arg(
                   QString::number(neededTables), QString::number(tablesLimit)),
                QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel)
                == QMessageBox::Cancel
            || ! checkCachedChecksum(title)) {

            return;
        }
    }

    if (m_deleteAllDrawsAction->isEnabled()) {
        // We have existing entered draws
        const QString title = tr("Alle Plätze auslosen");
        if (QMessageBox::warning(this, title,
                tr("<p><b>Alle Plätze auslosen</b></p>"
                   "<p>Es wurden bereits Auslosungen eingegeben. Wenn alle Plätze ausgelost "
                   "werden, dann werden alle bestehenden Auslosungen überschrieben und neu "
                   "vergeben.</p>"
                   "<p>Sollen wirklich alle Plätze ausgelost werden?</p>"),
                QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel)
                == QMessageBox::Cancel
            || ! checkCachedChecksum(title)) {

            return;
        }
    }

    if (m_db->tournamentStarted() && ! m_db->allRoundsFinished()) {
        const QString title = tr("Alle Plätze auslosen");
        if (QMessageBox::warning(this, title,
                tr("<p><b>Alle Plätze auslosen</b></p>"
                   "<p>Es wurden in mindestens einer Runde noch nicht alle Ergebnisse eingegeben. "
                   "Wenn jetzt neu ausgelost wird, dann wird das evtl. dazu führen, dass manche "
                   "Ergebnisse nicht entsprechend der Auslosung eingegeben werden können, weil "
                   "bereits ein Ergebnis für einen anders ausgelosten Tisch eingegeben wurde.</p>"
                   "<p>Soll die Auslosung wirklich neu erstellt werden?</p>"),
                QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No
            || ! checkCachedChecksum(title)) {

            return;
        }
    }

    const auto draw = m_drawPopup->completeDraw(m_db->pairsOrPlayersCount());
    setListDraw(draw);
}

void RegistrationPage::setListDraw(const QVector<Draw::Seat> &draw)
{
    if (m_isClient) {
        m_client->requestSetListDraw(m_allRegistrations, draw);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Setze alle Auslosungen …"));

    if (! m_db->setListDraw(m_allRegistrations, draw)) {
        return;
    }

    emit drawChanged();
    emit statusUpdate(tr("Alle Auslosungen gesetzt"));
    listModified();
    checkUpdateDrawOverviewPage(true);

    if (m_isServer) {
        m_server->broadcastSetListDraw(m_allRegistrations, draw);
    }

    QApplication::restoreOverrideCursor();
}

bool RegistrationPage::registerPhoneticSearchEnabled() const
{
    return m_registerPhoneticSearch->isChecked();
}

const QVector<QString> &RegistrationPage::allRegistrations() const
{
    return m_allRegistrations;
}

const QVector<Draw::Seat> &RegistrationPage::allDraws() const
{
    return m_allDraws;
}

const QVector<int> &RegistrationPage::incompleteTables() const
{
    return m_incompleteTables;
}

const QVector<int> &RegistrationPage::tableGaps() const
{
    return m_tableGaps;
}

void RegistrationPage::adoptDrawAdjustment(const QVector<QString> &names,
                                           const QVector<Draw::Seat> &seats)
{
    auto newDraw = m_allDraws;
    for (int i = 0; i < m_allRegistrations.count(); i++) {
        const auto &name = m_allRegistrations.at(i);
        if (names.contains(name)) {
            newDraw[i] = seats.at(names.indexOf(name));
        }
    }
    setListDraw(newDraw);
}

void RegistrationPage::emitDrawAdjustmentNeeded() const
{
    emit drawAdjustmentNeeded(m_drawAdjustmentNeeded);
}

bool RegistrationPage::drawAdjustmentNeededState() const
{
    return m_drawAdjustmentNeeded;
}

void RegistrationPage::adoptSearchText()
{
    const QString text = m_searchWidget->searchText().simplified();
    const auto words = text.split(QStringLiteral(" "));
    QString newText;
    for (const auto &word : words) {
        newText += QStringLiteral("%1%2 ").arg(m_stringTools->toUpper(word.at(0)),
                                               word.mid(1, word.length()));
    }

    m_playersName->setText(newText.simplified());
    m_searchWidget->clear();
    m_playersName->setFocus();
}

void RegistrationPage::showHeaderContextMenu(const QPoint &point)
{
    m_menuButton->clearFocus();
    m_columnsMenu->setTitleHidden(false);
    m_columnsMenu->exec(m_playersList->mapToGlobal(point));
    m_columnsMenu->setTitleHidden(true);
}

void RegistrationPage::setHideHeader(bool state)
{
    if (state && ! m_menuButton->hasFocus() && ! m_tournamentSettings->loadingDb()) {
        QMessageBox::information(this, tr("Kopfzeile ausblenden"),
            tr("<p><b>Kopfzeile ausblenden</b></p>"
               "<p>Die Kopfzeile der Anmeldungsliste kann über den „Einstellungen“-Knopf wieder "
               "eingeblendet werden.</p>"));
    }
    m_playersList->horizontalHeader()->setVisible(! state);
    m_tournamentSettings->setRegistrationListHeaderHidden(state);
}

void RegistrationPage::itemError(const char *funcInfoText, const QString &text)
{
    qCritical() << "Error:" << funcInfoText << "\n   Item not found:" << text;
    QTimer::singleShot(0, this,
        [this, funcInfoText, text]
        {
            QMessageBox::critical(this, tr("Unerwarteter Fehler"),
                tr("<p><b>Unerwarteter Fehler in<br/>"
                   "%1</b></p>"
                   "<p>Der Eintrag „%2“ konnte nicht gefunden werden. Bestenfalls funktioniert "
                   "alles weiter, aber womöglich müssen die Netzwerkverbindung erneut gestartet "
                   "und die Daten neu synchronisiert werden.</p>"
                   "<p><b>Das sollte nicht passieren!</b></p>"
                   "<p>Bitte einen Bugreport auf "
                   "<a href=\"%3\">%3</a> schreiben!</p>").arg(
                   QString::fromUtf8(funcInfoText), text, Urls::BUGTRACKER));
        });
}
