/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes

#include "RestoreBackupDialog.h"

#include "shared/Database.h"
#include "shared/BackupEngine.h"

// Qt includes
#include <QDebug>
#include <QRegularExpression>
#include <QMessageBox>
#include <QDir>
#include <QVBoxLayout>
#include <QListWidget>
#include <QLabel>
#include <QApplication>
#include <QTimer>

// C++ includes
#include <functional>

RestoreBackupDialog::RestoreBackupDialog(QWidget *parent, Database *db, BackupEngine &backupEngine,
                                         const QString &dbToOverwrite, bool autoAccept)
    : TitleDialog(parent),
      m_db(db), m_dbToOverwrite(dbToOverwrite), m_autoAccept(autoAccept)
{
    setTitle(tr("Backup wiederherstellen"));

    QLabel *header = new QLabel(tr("Für die Turnierdatenbank<br/><b>%1.mtdb</b><br/>"
                                   "wurden Backups zu folgenden Zeitpunkten gefunden:").arg(
                                   backupEngine.baseName()));
    header->setWordWrap(true);
    layout()->addWidget(header);

    m_backups = new QListWidget;
    layout()->addWidget(m_backups);
    connect(m_backups, &QListWidget::itemSelectionChanged,
            this, std::bind(&TitleDialog::setOkButtonEnabled, this, true));

    addButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
    setButtonText(QDialogButtonBox::Ok, tr("Wiederherstellen"));
    setOkButtonEnabled(false);

    // Search for Backups and fill in the data

    QDir dir(backupEngine.path());
    dir.setNameFilters( { backupEngine.baseName() + QLatin1String("_*.mtdb") } );
    QRegularExpression re(QStringLiteral("^")
                          + QRegularExpression::escape(backupEngine.baseName())
                          + QLatin1String("_(\\d{4}-\\d{2}-\\d{2}_"
                                          "\\d{2}-\\d{2}-\\d{2})_?(\\d+)?\\.mtdb$"));

    const auto entryList = dir.entryList(QDir::Files, QDir::Name);
    for (const QString &file : entryList) {
        QRegularExpressionMatch match = re.match(file);
        if (match.hasMatch()) {
            QString date = formatDateString(match.captured(1));
            if (! match.captured(2).isEmpty()) {
                date += QStringLiteral(" (%1)").arg(match.captured(2));
            }
            QListWidgetItem *item = new QListWidgetItem(date);
            item->setData(Qt::UserRole, QVariant(backupEngine.path() + QLatin1String("/") + file));
            m_backups->addItem(item);
        }
    }

    // Select the opened backup in case we have one and searched for the base db
    for (int i = 0; i < m_backups->count(); i++) {
        if (m_backups->item(i)->data(Qt::UserRole).toString() == m_db->dbFile()) {
            m_backups->item(i)->setText(tr("%1 (jetzt geöffnet)").arg(
                                           m_backups->item(i)->text()));
            m_backups->setCurrentItem(m_backups->item(i));
            break;
        }
    }

    if (m_backups->count() == 0) {
        m_backups->addItem(tr("Keine Backups gefunden"));
        m_backups->setEnabled(false);
    }

    if (m_autoAccept) {
        QTimer::singleShot(0, this, &RestoreBackupDialog::runAutoAccept);
    }
}

void RestoreBackupDialog::runAutoAccept()
{
    accept();
    TitleDialog::accept();
}

QString RestoreBackupDialog::formatDateString(const QString &backupDate) const
{
    const QDateTime backupDateTime = QDateTime::fromString(backupDate,
                                                           QStringLiteral("yyyy-MM-dd_hh-mm-ss"));

    if (QDateTime::currentDateTime().toString(QStringLiteral("dd.MM.yyyy"))
        != backupDateTime.toString(QStringLiteral("dd.MM.yyyy"))) {

        return backupDateTime.toString(QStringLiteral("dd.MM.yyyy hh:mm:ss"));
    } else {
        return backupDateTime.toString(QStringLiteral("hh:mm:ss"));
    }
}

void RestoreBackupDialog::accept()
{
    QFileInfo backupInfo(m_backups->currentItem()->data(Qt::UserRole).toString());
    QFileInfo dbToOverwriteInfo(m_dbToOverwrite);

    if (QMessageBox::question(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Soll das Backup wirklich wiederhergestellt werden?</b></p>"
               "<p>Die Turnierdatenbank „%1“ wird dabei mit dem Backup „%2“ überschrieben und "
               "alle Änderungen gehen verloren!</p>").arg(
               dbToOverwriteInfo.fileName(), backupInfo.fileName()),
            QMessageBox::Ok | QMessageBox::Cancel) == QMessageBox::Cancel) {

        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    const QString openedDb = m_db->dbFile();
    emit closeTournament();

    if (m_db->checkDatabase(backupInfo.filePath()) != Database::Error::NoError) {
        QApplication::restoreOverrideCursor();
        emit statusUpdate(tr("Wiederherstellung fehlgeschlagen!"));
        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Das Backup kann nicht wiederhergestellt werden:</b></p>")
            + m_db->errorMessage()
            + tr("<p>Die zuletzt geöffnete Datenbank wird jetzt neu geladen.</p>"));
        emit openTournament(openedDb);
        return;
    } else {
        m_db->closeDatabaseConnection();
    }

    emit statusUpdate(tr("Stelle Backup wieder her …"));

    const bool dbToOverwriteExists = QFileInfo::exists(m_dbToOverwrite);

    // To be sure, create a backup of the database to be overwritten
    BackupEngine oldDbEngine(m_dbToOverwrite);
    if (dbToOverwriteExists && ! oldDbEngine.createBackup(QStringLiteral("~"))) {
        QApplication::restoreOverrideCursor();
        emit statusUpdate(tr("Wiederherstellung fehlgeschlagen!"));
        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
               "<p>Es konnte keine Sicherungskopie der Datenbank „%1“ erstellt werden.</p>"
               "<p>Die zuletzt geöffnete Datenbank wird jetzt neu geladen.</p>").arg(
               dbToOverwriteInfo.fileName()));
        emit openTournament(openedDb);
        return;
    }

    // Copy the backup to restore

    if (dbToOverwriteExists && ! QFile::remove(m_dbToOverwrite)) {
        // This shouldn't happen
        QApplication::restoreOverrideCursor();
        emit statusUpdate(tr("Wiederherstellung fehlgeschlagen!"));
        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
               "<p>Das Backup „%1“ konnte nicht wiederhergestellt werden, weil die zu "
               "überschreibende Datenbank „%2“ nicht gelöscht werden konnte.</p>"
               "<p>Die zuletzt geöffnete Datenbank wird jetzt neu geladen.</p>").arg(
               backupInfo.fileName(), dbToOverwriteInfo.fileName()));
        emit openTournament(openedDb);
        return;
    }

    QString dbToOverwriteBackup;
    if (dbToOverwriteExists) {
        dbToOverwriteBackup = oldDbEngine.path() + QLatin1String("/") + oldDbEngine.backupFile();
    }

    if (! QFile::copy(backupInfo.filePath(), m_dbToOverwrite)) {
        // This shouldn't happen either
        QApplication::restoreOverrideCursor();
        emit statusUpdate(tr("Wiederherstellung fehlgeschlagen!"));
        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
               "<p>Das Backup „%1“ konnte nicht nach „%2“ kopiert werden.</p>"
               "<p>Die Sicherungskopie „%3“ der zu überschreibenden Datenbank wird jetzt wieder "
               "hergestellt und die zuletzt geöffnete Datenbank neu geladen.</p>").arg(
               backupInfo.fileName(), dbToOverwriteInfo.fileName(), oldDbEngine.backupFile()));

        if (dbToOverwriteExists && ! QFile::rename(dbToOverwriteBackup, m_dbToOverwrite)) {
            // This really should not happen ...
            QMessageBox::critical(this,
                tr("Backup wiederherstellen"),
                tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
                   "<p>Die Sicherungskopie „%1“ der zu überschreibenden Datenbank konnte nicht "
                   "nach „%2“ umbenannt werden.</p>").arg(
                oldDbEngine.backupFile(), dbToOverwriteInfo.fileName()));
            return;
        }

        emit openTournament(openedDb);
        return;
    }

    // Everything is fine :-)

    emit openTournament(m_dbToOverwrite);
    QApplication::restoreOverrideCursor();

    QMessageBox::information(this,
        tr("Backup wiederherstellen"),
        tr("<p><b>Backup erfolgreich wiederhergestellt</b></p>"
           "<p>Das Backup „%1“ der Turnierdatenbank „%2“ wurde erfolgreich wiederhergestellt!"
           "</p>").arg(backupInfo.fileName(), dbToOverwriteInfo.fileName()));

    if (dbToOverwriteExists && ! QFile::remove(dbToOverwriteBackup)) {
        // Also this should not happen
        QMessageBox::warning(this,
            tr("Backup wiederherstellen"),
            tr("Die temporäre Sicherungskopie „%1“, die vor dem Überschreiben der Datenbank "
               "angelegt wurde, konnte nicht gelöscht werden. Bitte manuell entfernen.").arg(
               QDir::toNativeSeparators(dbToOverwriteBackup)));
    }

    TitleDialog::accept();
}
