/* Copyright (c) 2010-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes

#include "ScorePage.h"
#include "BoogerWidget.h"
#include "NumbersComboBox.h"
#include "PlayersComboBox.h"
#include "BoogerSpinBox.h"
#include "ScorePopup.h"
#include "TableComboBox.h"
#include "PlayersWidget.h"
#include "ScoreWidget.h"

#include "shared/SharedStyles.h"
#include "shared/SharedObjects.h"
#include "shared/StringTools.h"
#include "shared/SearchWidget.h"
#include "shared/TableDelegate.h"
#include "shared/SearchEngine.h"
#include "shared/TitleMenu.h"
#include "shared/Settings.h"
#include "shared/TournamentSettings.h"
#include "shared/PhoneticSearchButton.h"
#include "shared/IconButton.h"
#include "shared/ChecksumHelper.h"
#include "shared/Scores.h"
#include "shared/MinimumViewportScroll.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QGridLayout>
#include <QCheckBox>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>
#include <QAction>
#include <QMessageBox>
#include <QApplication>
#include <QScrollBar>
#include <QLineEdit>
#include <QSplitter>
#include <QActionGroup>
#include <QTime>
#include <QTimer>

// C++ includes
#include <functional>
#include <utility>

ScorePage::ScorePage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_stringTools(sharedObjects->stringTools()),
      m_searchEngine(sharedObjects->searchEngine())
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    m_networkBlocked = new QLabel(tr(
        "<p>Im Netzwerkbetrieb muss die Anmeldung explizit abgeschlossen werden,<br/>"
        "bevor Ergebnisse eingegeben werden können. <a href=\"#finishRegistration\">"
        "Anmeldung beenden</a></p>"));
    m_networkBlocked->setTextInteractionFlags(Qt::TextBrowserInteraction);
    connect(m_networkBlocked, &QLabel::linkActivated, this, &ScorePage::processLinkActivated);

    m_networkBlocked->setAlignment(Qt::AlignCenter);
    m_networkBlocked->setWordWrap(true);
    layout->addWidget(m_networkBlocked);
    m_networkBlocked->hide();

    m_notUsable = new QLabel(tr("<i>Es können noch keine Ergebnisse eingetragen werden</i>"));
    m_notUsable->setAlignment(Qt::AlignCenter);
    m_notUsable->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Expanding);
    layout->addWidget(m_notUsable);

    // Round selection part
    // ====================

    m_roundBox = new QGroupBox;
    m_roundBox->setProperty("notitle", true);
    QHBoxLayout *roundBoxLayout = new QHBoxLayout(m_roundBox);
    layout->addWidget(m_roundBox);

    QLabel *roundLabel = new QLabel(tr("Runde:"));
    roundLabel->setStyleSheet(SharedStyles::BOLD_TEXT);
    roundBoxLayout->addWidget(roundLabel);

    m_roundSelect = new NumbersComboBox;
    roundBoxLayout->addWidget(m_roundSelect);
    connect(m_roundSelect, &NumbersComboBox::currentNumberChanged, this, &ScorePage::openRound);

    roundBoxLayout->addStretch();

    QPushButton *optionsButton = new QPushButton(tr("Einstellungen"));

    QMenu *optionsMenu = new QMenu(this);

    auto *autoSelectMenu = optionsMenu->addMenu(tr("Automatische Auswahl"));

    m_autoSelectPairs = new QAction(tr("Auswahl laut Auslosung"));
    m_autoSelectPairs->setCheckable(true);
    connect(m_autoSelectPairs, &QAction::changed, this, &ScorePage::autoSelectPairsChanged);
    autoSelectMenu->addAction(m_autoSelectPairs);

    m_selectByLastRound = new QAction(tr("Paar 2 rückt pro Runde weiter"));
    m_selectByLastRound->setCheckable(true);
    m_selectByLastRound->setEnabled(false);
    connect(m_selectByLastRound, &QAction::changed, this, &ScorePage::selectByLastRoundChanged);
    autoSelectMenu->addAction(m_selectByLastRound);

    QActionGroup *scoreTypeGroup = new QActionGroup(this);

    m_horizontalScore = new QAction(tr("horizontal"), scoreTypeGroup);
    m_horizontalScore->setCheckable(true);
    connect(m_horizontalScore, &QAction::changed, this,
            std::bind(&ScorePage::scoreTypeChanged, this, Tournament::ScoreType::Horizontal));

    m_verticalScore = new QAction(tr("vertikal"), scoreTypeGroup);
    m_verticalScore->setCheckable(true);
    connect(m_verticalScore, &QAction::changed, this,
            std::bind(&ScorePage::scoreTypeChanged, this, Tournament::ScoreType::Vertical));

    m_scoreTypeMenu = optionsMenu->addMenu(tr("Spielstandzettel"));
    m_scoreTypeMenu->addAction(m_horizontalScore);
    m_scoreTypeMenu->addAction(m_verticalScore);

    optionsMenu->addSeparator();

    m_noTableNumbers = new QAction(tr("Eingabe ohne Tischnummern"), this);
    m_noTableNumbers->setCheckable(true);
    connect(m_noTableNumbers, &QAction::changed, this, &ScorePage::noTableNumbersChanged);
    optionsMenu->addAction(m_noTableNumbers);

    optionsButton->setMenu(optionsMenu);
    roundBoxLayout->addWidget(optionsButton);

    // Entering results part
    // =====================

    m_resultBox = new QGroupBox(tr("Ergebnisse eintragen"));
    QVBoxLayout *resultBoxLayout = new QVBoxLayout(m_resultBox);

    // Table and pairs/players

    m_playersLayout = new QHBoxLayout;
    resultBoxLayout->addLayout(m_playersLayout);

    m_tableLabel = new QLabel(tr("Tisch:"));
    m_tableLabel->setStyleSheet(SharedStyles::BOLD_TEXT);

    m_tableSelect = new TableComboBox;
    connect(m_tableSelect, &TableComboBox::currentNumberChanged, this, &ScorePage::tableChanged);
    connect(m_tableSelect, &TableComboBox::checkSavePossible, this, &ScorePage::checkSavePossible);

    m_phoneticSearch = new PhoneticSearchButton(QStringLiteral("score_page_players"),
                                                sharedObjects);

    m_playersSelect.reserve(4);
    for (int i = 0; i < 4; i++) {
        PlayersComboBox *players = new PlayersComboBox(m_searchEngine, i);
        connect(players, &PlayersComboBox::selectionChanged,
                this, &ScorePage::playersChanged);
        connect(players, &PlayersComboBox::disableSaveScore,
                this, [this] { m_saveScoreButton->setEnabled(false); });
        connect(players, &PlayersComboBox::checkSavePossible,
                this, &ScorePage::checkSavePossible);
        connect(m_phoneticSearch, &PhoneticSearchButton::toggled,
                players, &PlayersComboBox::setPhoneticSearch);
        m_playersSelect.append(players);
    }

    m_noAutoselect = new IconButton(sharedObjects, QStringLiteral("noautoselect.png"), this);
    m_noAutoselect->setToolTip(tr("<p>Automatische Paar-, Spieler- bzw. Tischauswahl "
                                  "(de)aktivieren</p>"));
    m_noAutoselect->setWhatsThis(m_noAutoselect->toolTip()
                                 + tr("<p>Wenn diese Option aktiviert ist, können Ergebnisse "
                                      "eingegeben werden, die nicht zur Auslosung passen.</p>"));

    auto *line = new QFrame;
    line->setFrameShape(QFrame::HLine);
    resultBoxLayout->addWidget(line);

    // Score input

    m_scoreBoxScroll = new MinimumViewportScroll(Qt::Vertical);
    m_scoreBoxScroll->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Expanding);
    m_scoreBoxScroll->setFrameShape(QFrame::NoFrame);
    resultBoxLayout->addWidget(m_scoreBoxScroll);

    m_saveScoreButton = new QPushButton(tr("Ergebnis\neintragen"));
    connect(m_saveScoreButton, &QPushButton::clicked, this, &ScorePage::saveScore);

    // Round summary
    // =============

    m_roundSummaryBox = new QGroupBox;
    m_roundSummaryBox->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Ignored);
    auto *roundSummaryBoxLayout = new QVBoxLayout(m_roundSummaryBox);

    // Search Label
    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    roundSummaryBoxLayout->addWidget(m_isSearchLabel);

    // Scores

    auto *scoresLayout = new QHBoxLayout;
    roundSummaryBoxLayout->addLayout(scoresLayout);

    m_scoreList = new QTableWidget;
    m_scoreList->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_tableDelegate = new TableDelegate(this, TableDelegate::HighlightMode::NRows);
    m_scoreList->setItemDelegate(m_tableDelegate);
    m_scoreList->verticalHeader()->hide();
    m_scoreList->setShowGrid(false);
    m_scoreList->setWordWrap(false);
    m_scoreList->setSelectionMode(QTableWidget::NoSelection);
    m_scoreList->setFocusPolicy(Qt::NoFocus);
    m_scoreList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_scoreList, &QTableWidget::cellDoubleClicked,
            [this] { showContextMenu(m_scoreList->viewport()->mapFromGlobal(QCursor::pos())); } );
    connect(m_scoreList, &QTableWidget::customContextMenuRequested,
            this, &ScorePage::showContextMenu);
    scoresLayout->addWidget(m_scoreList);

    // Client server sync display

    auto *clientLayout = new QVBoxLayout;
    scoresLayout->addLayout(clientLayout);

    m_lastChange = new QLabel;
    m_lastChange->hide();
    clientLayout->addWidget(m_lastChange);

    m_syncStatus = new QLabel;
    m_syncStatus->setTextInteractionFlags(Qt::TextBrowserInteraction);
    connect(m_syncStatus, &QLabel::linkActivated, this, &ScorePage::processLinkActivated);
    m_syncStatus->hide();
    clientLayout->addWidget(m_syncStatus);

    clientLayout->addStretch();

    // Splitter
    // ========

    m_splitter = new QSplitter(Qt::Vertical);
    m_splitter->setHandleWidth(3);

    m_splitter->addWidget(m_resultBox);
    m_splitter->setCollapsible(0, false);
    m_splitter->setStretchFactor(0, 1);

    m_splitter->addWidget(m_roundSummaryBox);
    m_roundSummaryBox->setMinimumHeight(100);
    m_splitter->setCollapsible(1, false);
    m_splitter->setStretchFactor(1, 100);

    layout->addWidget(m_splitter);

    connect(m_splitter, &QSplitter::splitterMoved, this, &ScorePage::toggleSplitterHandle);

    // Search widget
    // =============

    m_searchWidget = new SearchWidget(QStringLiteral("score_page_search"), sharedObjects);
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &ScorePage::updateSearch);
    roundSummaryBoxLayout->addWidget(m_searchWidget);

    // Context menu
    // ============

    m_contextMenu = new TitleMenu(this);

    m_editAction = new QAction(tr("Ändern (Löschen und neu eingeben)"), this);
    m_contextMenu->addAction(m_editAction);
    connect(m_editAction, &QAction::triggered, this, &ScorePage::editScore);

    m_swapPairsAction = new QAction(tr("Paar 1 und Paar 2 vertauschen"), this);
    m_contextMenu->addAction(m_swapPairsAction);
    connect(m_swapPairsAction, &QAction::triggered, this, &ScorePage::swapPairs);

    m_contextMenu->addSeparator();

    m_deleteAction = new QAction(tr("Löschen"), this);
    m_contextMenu->addAction(m_deleteAction);
    connect(m_deleteAction, &QAction::triggered, this, &ScorePage::deleteScore);

    // Score dialog
    m_scorePopup = new ScorePopup(this, &m_tournamentSettings->boogerScore());
    connect(m_scorePopup, &QDialog::accepted, this, &ScorePage::setSelectedScore);
}

void ScorePage::setupGui()
{
    // Setup the players part
    // ----------------------

    if (m_playersWidget != nullptr) {
        m_tableLabel->setParent(this);
        m_tableSelect->setParent(this);
        for (PlayersComboBox *players : std::as_const(m_playersSelect)) {
            players->setParent(this);
        }
        m_phoneticSearch->setParent(this);
        m_playersWidget->deleteLater();
        m_playersLayout->removeWidget(m_noAutoselect);
    }

    m_playersWidget = new PlayersWidget(m_tournamentSettings->scoreType(),
                                        m_tableLabel,
                                        m_tableSelect,
                                        m_playersSelect,
                                        m_phoneticSearch);
    m_playersLayout->addWidget(m_playersWidget);
    m_playersLayout->addWidget(m_noAutoselect);

    // Show the needed input fields for pairs or players
    m_playersSelect.at(1)->setVisible(m_tournamentSettings->isSinglePlayerMode());
    m_playersSelect.at(3)->setVisible(m_tournamentSettings->isSinglePlayerMode());

    // Setup the score part
    // --------------------

    if (m_scoreWidget != nullptr) {
        m_saveScoreButton->setParent(this);
        m_scoreWidget->deleteLater();
    }

    m_boogerWidgets.clear();
    const Tournament::ScoreType scoreType = m_tournamentSettings->scoreType();
    for (int i = 1; i <= m_tournamentSettings->boogersPerRound(); i++) {
        BoogerWidget *boogerWidget = new BoogerWidget(this, scoreType, i,
                                                      m_tournamentSettings->boogersPerRound(),
                                                      m_tournamentSettings->boogerScore());
        connect(boogerWidget, &BoogerWidget::abortedStateChanged,
                this, &ScorePage::boogerAbortedStateChanged);
        connect(boogerWidget, &BoogerWidget::scorePopupRequested,
                this, &ScorePage::showScorePopup);

        m_boogerWidgets.append(boogerWidget);
    }
    connect(m_boogerWidgets.last(), &BoogerWidget::returnPressed,
            this, &ScorePage::checkEnterScore);

    m_scoreSize = m_boogerWidgets.count() * m_db->necessaryDivisor();

    m_scoreWidget = new ScoreWidget(scoreType, m_boogerWidgets, m_saveScoreButton);
    m_saveScoreButton->setEnabled(false);
    m_scoreBoxScroll->setWidget(m_scoreWidget);

    m_horizontalScore->blockSignals(true);
    m_verticalScore->blockSignals(true);
    m_horizontalScore->setChecked(scoreType == Tournament::ScoreType::Horizontal);
    m_verticalScore->setChecked(scoreType == Tournament::ScoreType::Vertical);
    m_horizontalScore->blockSignals(false);
    m_verticalScore->blockSignals(false);

    // Disable data input if we have a read-only database
    m_resultBox->setEnabled(m_db->isChangeable());
    m_noTableNumbers->setEnabled(m_db->isChangeable());
    m_scoreTypeMenu->setEnabled(m_db->isChangeable());
    m_autoSelectPairs->setEnabled(m_db->isChangeable());
    m_autoSelectPairs->setEnabled(m_db->isChangeable());

    // Update the results widget

    m_scoreList->clearSpans();
    m_scoreList->clear();

    switch (m_tournamentSettings->scoreType()) {

    case Tournament::ScoreType::Horizontal:
    {
        m_scoreList->setColumnCount(m_tournamentSettings->boogersPerRound() + 2);
        QStringList headerLabels { tr("Tisch"), tr("Paare") };
        for (int i = 1; i <= m_tournamentSettings->boogersPerRound(); i++) {
            headerLabels.append(tr("%1. Bobbl").arg(i));
        }
        m_scoreList->setHorizontalHeaderLabels(headerLabels);
        m_resultsAdvance = 2;
        m_pair1Column = 1;
        m_pair2Column = 1;
        m_pair2RowOffset = 1;
    }
    break;

    case Tournament::ScoreType::Vertical:
    {
        m_scoreList->setColumnCount(4);
        m_scoreList->setHorizontalHeaderLabels({ tr("Tisch"), tr("Bobbl"),
                                                 tr("Paar 1"), tr("Paar 2") });
        m_resultsAdvance = m_tournamentSettings->boogersPerRound() + 1;
        m_pair1Column = 2;
        m_pair2Column = 3;
        m_pair2RowOffset = 0;
    }
    break;

    }

    m_tableDelegate->setHighlighRows(m_resultsAdvance);
    m_scoreList->setContextMenuPolicy(m_db->isChangeable() ? Qt::CustomContextMenu
                                                           : Qt::NoContextMenu);

    // Update the score dialog
    m_scorePopup->setupGui();

    // Adjust the booger widget's scroll area's height
    toggleSplitterHandle();
}

void ScorePage::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);
    toggleSplitterHandle();
}

void ScorePage::toggleSplitterHandle()
{
    // This has to be called via a zero single shot so that all widgets are
    // rendered correctly, have their sizes and scrollbars visible or hidden.

    QTimer::singleShot(0, this, [this]
    {
        const auto *horizontalScrollBar = m_scoreBoxScroll->horizontalScrollBar();
        const int horizontalScrollBarHeight
            = horizontalScrollBar->isVisible() ? horizontalScrollBar->height() : 0;

        if (m_scoreBoxScroll->widget()->height() + horizontalScrollBarHeight
                == m_scoreBoxScroll->height()
            && ! m_scoreBoxScroll->verticalScrollBar()->isVisible()) {

            m_splitter->setStyleSheet(SharedStyles::SPLITTER_HIDDEN);

        } else {
            m_splitter->setStyleSheet(SharedStyles::SPLITTER_VISIBLE_VERTICAL);
        }

    });
}

void ScorePage::setUsable(bool state)
{
    m_isUsable = state;
    m_networkBlocked->setVisible(state && m_isNetworkBlocked);
    m_notUsable->setVisible(! state);
    m_roundBox->setVisible(state);
    m_splitter->setVisible(state);

    if (state) {
        toggleSplitterHandle();
    }
}

void ScorePage::setNetworkBlocked(bool state)
{
    m_isNetworkBlocked = state;
    m_networkBlocked->setVisible(state && m_isUsable);
    m_roundBox->setEnabled(! state);
    m_splitter->setEnabled(! state);
}

void ScorePage::playersNameEdited()
{
    const QVector<Database::BoogerResult> score = currentScore();
    const int lastOpenedRound = m_roundSelect->currentNumber();
    const int selectedTable = m_tableSelect->currentNumber();

    // Check if we currently search for a name
    int editedIndex = -1;
    QString editedText;

    for (int i = 0; i < 4; i += m_tournamentSettings->isSinglePlayerMode() ? 1 : 2) {
        if (m_playersSelect.at(i)->itemText(m_playersSelect.at(i)->currentIndex())
            != m_playersSelect.at(i)->currentText()) {

            editedIndex = i;
            editedText = m_playersSelect.at(i)->currentText();
            break;
        }
    }

    reload(m_roundSelect->currentNumber());

    fillInScore(lastOpenedRound, selectedTable, score);
    if (editedIndex != -1) {
        m_playersSelect.at(editedIndex)->lineEdit()->setText(editedText);
        m_playersSelect.at(editedIndex)->lineEditTextChanged(editedText);
    }
}

void ScorePage::reload(int lastRound)
{
    if (! isVisible()) {
        return;
    }

    // Reset all booger widgets if the GUI has been already setup
    if (lastRound != -1) {
        for (BoogerWidget *boogerWidget : std::as_const(m_boogerWidgets)) {
            boogerWidget->reset();
        }
        m_boogerWidgets.at(0)->setFocus();
    }

    // Fill the round select

    blockSelectSignals(true);

    int allRounds = 0;
    int firstOpenRound = -1;

    m_roundSelect->clear();
    const auto roundsState = m_db->roundsState();
    const auto roundsStateKeys = roundsState.keys();
    for (const int roundNumber : roundsStateKeys) {
        if (roundsState.value(roundNumber)) {
            m_roundSelect->addItem(tr("%1 (abgeschlossen)").arg(roundNumber),
                                   roundNumber);
        } else {
            m_roundSelect->addItem(tr("%1 (offen)").arg(roundNumber),
                                   roundNumber);
            if (firstOpenRound == -1) {
                firstOpenRound = roundNumber;
            }
        }
        allRounds++;
    }

    if (firstOpenRound == -1) {
        firstOpenRound = 1;
    }
    if (lastRound != -1) {
        firstOpenRound = lastRound;
    }

    // Add one for a new round
    if (m_db->isChangeable()) {
        allRounds++;
        m_roundSelect->addItem(tr("%1 (neue Runde)").arg(allRounds), allRounds);
        // The "false" value for the new round's missing roundsState entry will be provided
        // by setting the default value to "false" in openRound().
    }

    // Display the data
    m_roundSelect->setCurrentNumber(firstOpenRound);
    blockSelectSignals(false);
    openRound(firstOpenRound);
}

void ScorePage::openRound(int round)
{
    if (round == -1) {
        return;
    }

    m_currentRound = round;
    emit roundSelectionChanged();

    blockSelectSignals(true);

    m_searchWidget->saveSearch();

    // We set the "false" default value here because if a new round is selected,
    // the Database's roundsState map won't contain a value for it yet

    if (! m_db->roundsState().value(round, false)) {
        m_resultBox->setVisible(true);

        // Insert all missing tables

        m_tableSelect->clear();

        const auto [ missingTables, success1 ] = m_db->missingTables(round);
        if (! success1) {
            return;
        }

        for (const int table : missingTables) {
            m_tableSelect->addItem(QString::number(table), table);
        }
        m_tableSelect->setCurrentIndex(m_noTableNumbers->isChecked());

        // Insert all missing players/pairs

        for (int i = 0; i < 4; i += m_tournamentSettings->isSinglePlayerMode() ? 1 : 2) {
            m_playersSelect.at(i)->clear();
        }

        const auto [ missingPlayers, success2 ] = m_db->missingPlayers(round);
        if (! success2) {
            return;
        }

        QStringList availableNames = missingPlayers.keys();
        m_stringTools->sort(availableNames);
        for (const QString &name : std::as_const(availableNames)) {
            for (int i = 0; i < 4; i += m_tournamentSettings->isSinglePlayerMode() ? 1 : 2) {
                m_playersSelect.at(i)->addItem(name, missingPlayers.value(name));
            }
        }

    } else {
        m_resultBox->setVisible(false);
    }

    updateScore();
    updateAssignment(round);

    blockSelectSignals(false);
    checkSavePossible();
    m_searchWidget->restoreSearch();
}

void ScorePage::updateAssignment(int round)
{
    if (! m_autoSelectPairs->isChecked()) {
        // If we don't use automatic selection, we can skip the assignment generation
        return;
    }

    if (! m_selectByLastRound->isChecked()) {
        // If we don't consider entered results, we always use the preassignment
        // we get when we request the assignment for round 1
        round = 1;
    } else if (round == -1) {
        // This happens when this function is triggered by an assignment change on the players
        // page. In this case, we have to set the round variable to the opened round
        round = m_roundSelect->currentNumber();
    }

    const auto [ assignment, success ] = m_db->assignment(round);
    if (! success) {
        return;
    }
    m_assignment = assignment;

    // Reset the selection if there's one
    if (m_tableSelect->currentNumber() != -1) {
        blockSelectSignals(true);
        m_tableSelect->setCurrentIndex(0);
        for (int i = 0; i < 4; i++) {
            m_playersSelect.at(i)->setCurrentIndex(0);
        }
        blockSelectSignals(false);
        checkSavePossible();
    }
}

void ScorePage::updateScore()
{
    const int round = m_roundSelect->currentNumber();

    // Get all scores for the current round
    const auto [ score, success ] = m_db->scores(round);
    if (! success) {
        return;
    }

    // If we're on a network: Process checksums

    if (m_isServer || m_isClient) {
        const QString newChecksum = scoreChecksum(score);
        const bool checksumChanged = newChecksum != m_checksums[round];
        m_checksums[round] = newChecksum;
        if (m_isServer && checksumChanged) {
            // We have to announce the checksum change via a single shot call, because otherwise,
            // if this happens via a remote name change, the round checksum is updated and it's
            // change is broadcasted before the name change is broadcasted and thus the checkums
            // won't match on the client side.
            QTimer::singleShot(0, this, [this, round, newChecksum]
            {
                processChangedChecksums(m_checksums);
                emit scoreChecksumChanged(round);
            });
        }

        if (m_isClient) {
            updateSyncStatus();
        }
    }

    const int resultsPerRound = m_db->pairsOrPlayersCount() / m_db->necessaryDivisor();
    const int currentResults = score.tables.count();

    // Update the box header
    if (currentResults == 0) {
        m_roundSummaryBox->setTitle(
            tr("Bisher keine Ergebnisse aus der %1. Runde").arg(round));
    } else if (currentResults < resultsPerRound) {
        m_roundSummaryBox->setTitle(
            tr("Bisher %n Ergebnis(se) aus der %1. Runde", "", currentResults).arg(round)
            + QStringLiteral(" ")
            + tr("(%n fehlt/fehlen noch)", "", resultsPerRound - currentResults));
    } else {
        m_roundSummaryBox->setTitle(
            tr("Ergebnisse der %1. Runde").arg(round));
    }

    // Hide/Show the search widget and update the cache for single player tournaments
    if (currentResults == 0) {
        m_searchWidget->clear();
        m_isSearchLabel->hide();
        m_searchWidget->hide();
    } else {
        m_searchWidget->show();
        if (m_tournamentSettings->isSinglePlayerMode()) {
            m_searchEngine->clearSearchCache(SearchEngine::SearchType::DrawnPairsScore);
            for (const QString &pairName : score.pairs) {
                m_searchEngine->updateSearchCache(pairName,
                                                  SearchEngine::SearchType::DrawnPairsScore);
            }
        }
    }

    // Display all results

    const int boogersPerRound = m_tournamentSettings->boogersPerRound();

    switch (m_tournamentSettings->scoreType()) {

    case Tournament::ScoreType::Horizontal:
        m_scoreList->setRowCount(score.scores.count());

        // Tables
        for (int i = 0; i < score.tables.count(); i++) {
            const int row = i * 2;
            m_scoreList->setItem(row, 0, new QTableWidgetItem(QString::number(score.tables.at(i))));
            m_scoreList->setSpan(row, 0, 2, 1);
        }

        // Pairs
        for (int row = 0; row < score.pairs.count(); row++) {
            m_scoreList->setItem(row, 1, new QTableWidgetItem(score.pairs.at(row)));
        }

        // Results
        for (int row = 0; row < score.scores.count(); row++) {
            const auto &pairResult = score.scores.at(row);
            for (int booger = 1; booger <= boogersPerRound; booger++) {
                m_scoreList->setItem(row, 1 + booger,
                    new QTableWidgetItem(m_stringTools->boogify(pairResult.at(booger - 1))));
            }
        }

        break;

    case Tournament::ScoreType::Vertical:
        const int rowsPerResult = m_tournamentSettings->boogersPerRound() + 1;
        m_scoreList->setRowCount(currentResults * rowsPerResult);

        // Table numbers
        for (int i = 0; i < score.tables.count(); i++) {
            const int row = i * rowsPerResult;
            m_scoreList->setItem(row, 0, new QTableWidgetItem(QString::number(score.tables.at(i))));
            m_scoreList->setSpan(row, 0, rowsPerResult, 1);
        }

        // Pairs
        for (int i = 0; i < score.pairs.count(); i += 2) {
            const int row = (i / 2) * rowsPerResult;
            // This dummy item is needed so that the context menu
            // will show up if it's requested at this area:
            m_scoreList->setItem(row, 1, new QTableWidgetItem);
            m_scoreList->setItem(row, 2, new QTableWidgetItem(score.pairs.at(i)));
            m_scoreList->setItem(row, 3, new QTableWidgetItem(score.pairs.at(i + 1)));
        }

        // Booger labels
        for (int i = 0; i < score.scores.count(); i++) {
            for (int booger = 1; booger <= boogersPerRound; booger++) {
                m_scoreList->setItem(i * rowsPerResult + booger, 1,
                                     new QTableWidgetItem(tr("%1.").arg(booger)));
            }
        }

        // Results
        int column = 0;
        for (int i = 0; i < score.scores.count(); i++) {
            const auto &pairResult = score.scores.at(i);
            for (int booger = 1; booger <= boogersPerRound; booger++) {
                m_scoreList->setItem(i / 2 * rowsPerResult + booger, 2 + column,
                    new QTableWidgetItem(m_stringTools->boogify(pairResult.at(booger - 1))));
            }
            column = (column + 1) % 2;
        }

        break;

    }

    m_scoreList->resizeColumnsToContents();
    updateLastChange(m_lastLocalChange);
}

void ScorePage::tableChanged(int table)
{
    blockSelectSignals(true);

    if (! m_autoSelectPairs->isChecked() || m_noAutoselect->isChecked()) {
        checkSavePossible();
        return;
    }

    bool allSelectable = true;
    const bool isPairMode = m_tournamentSettings->isPairMode();
    QVector<int> missingPlayers;

    if (m_assignment.idsForTable.contains(table)) {
        const auto &ids = m_assignment.idsForTable[table];
        for (int i = 0; i < 4; i++) {
            const int id = ids.at(i);
            if (id != 0) {
                if (! m_playersSelect.at(i)->setCurrentNumber(id)) {
                    allSelectable = false;
                }
            } else {
                if ((isPairMode && Draw::playerForIndex(i) == 1) || ! isPairMode) {
                    missingPlayers.append(i);
                }
            }
        }
    }

    checkSavePossible();

    const int missingPlayersCount = missingPlayers.count();
    if (missingPlayersCount > 0) {
        if (isPairMode) {
            const int pair = Draw::pairForIndex(missingPlayers.at(0));
            QMessageBox::warning(this,
                tr("Automatische Auswahl"),
                tr("Für Tisch %1 wurde Paar %2 bisher nicht zugelost.\n"
                   "Es wurde nur Paar %3 ausgewählt!").arg(
                   QString::number(table), QString::number(pair),
                   QString::number(Draw::otherPair(pair))));
        } else {
            QString list;
            for (int i = 0; i < missingPlayersCount; i++) {
                if (i > 0 && i < missingPlayersCount - 1) {
                    list.append(tr(", "));
                } else if (missingPlayersCount > 1 && i == missingPlayersCount - 1) {
                    list.append(tr(" und "));
                }
                const int index = missingPlayers.at(i);
                list.append(tr("Paar %1 Spieler %2").arg(
                               QString::number(Draw::pairForIndex(index)),
                               QString::number(Draw::playerForIndex(index))));
            }
            QMessageBox::warning(this,
                tr("Automatische Auswahl"),
                tr("<p>Tisch %1 wurde noch nicht komplett ausgelost. Folgende(r) Spieler wurde(n) "
                   "nicht ausgewählt:</p><p>%2</p>", "", missingPlayersCount).arg(
                   QString::number(table), list));
        }
    }

    if (! allSelectable) {
        QMessageBox::warning(this,
            tr("Automatische Auswahl"),
            tr("Es konnten nicht alle %1 entsprechend der Auslosung ausgewählt werden, da entweder "
               "für die entsprechenden %1 bereits manuell ein Ergebnis eingegeben wurde, oder die "
               "Auslosung geändert wurde und noch nicht alle Ergebnisse eingegeben waren.").arg(
               m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
    }

    blockSelectSignals(false);
}

void ScorePage::playersChanged(int index, int pairId)
{
    if (! m_autoSelectPairs->isChecked() || m_noAutoselect->isChecked()
        || ! m_assignment.tableForId.contains(pairId)) {

        return;
    }

    const auto &ids = m_assignment.idsForTable[m_assignment.tableForId.value(pairId)];

    const auto selectedPairNumber = (index > 1) + 1;
    const auto assignedPairNumber = (ids.indexOf(pairId) > 1) + 1;
    if (selectedPairNumber != assignedPairNumber) {
        QString text = tr("<p><b>Die Auswahl passt nicht zur Auslosung</b><p>");
        if (m_tournamentSettings->isPairMode()) {
            text += tr("<p>Als Paar %1 wurde gerade „%2“ ausgewählt. Laut der Auslosung bzw. dem "
                       "Ergebnis der letzten Runde ist dieses Paar aber ein Paar %3!</p>").arg(
                       QString::number(selectedPairNumber),
                       m_playersSelect.at(index)->currentText(),
                       QString::number(assignedPairNumber));
        } else {
            text += tr("<p>Als Spieler von Paar %1 wurde gerade „%2“ ausgewählt. Laut der "
                       "Auslosung bzw. dem Ergebnis der letzten Runde gehört dieser Spieler aber "
                       "zu Paar %3!</p>").arg(
                       QString::number(selectedPairNumber),
                       m_playersSelect.at(index)->currentText(),
                       QString::number(assignedPairNumber));
        }
        QMessageBox::warning(this, tr("Die Auswahl passt nicht zur Auslosung"), text);
        return;
    }

    const int assignedTable = m_assignment.tableForId.value(pairId);
    if (m_tableSelect->currentNumber() == assignedTable) {
        tableChanged(assignedTable);
    } else {
        if (! m_tableSelect->setCurrentNumber(assignedTable)) {
            QMessageBox::warning(this,
                tr("Automatische Auswahl"),
                tr("Der zum ausgewählten %1 zugeloste Tisch Nr. %2 kann nicht ausgewählt werden, "
                   "weil entweder bereits manuell ein Ergebnis für diesen Tisch eingegeben wurde, "
                   "oder die Auslosung geändert wurde und noch nicht alle Ergebnisse eingegeben "
                   "waren.").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                   QString::number(assignedTable)));
        }
    }
}

void ScorePage::checkSavePossible()
{
    switch (m_tournamentSettings->tournamentMode()) {
    case Tournament::Mode::FixedPairs:
        m_saveScoreButton->setEnabled(
            m_tableSelect->inputOkay()
            && m_tableSelect->currentNumber() != -1
            && m_playersSelect.at(0)->currentIndex() != m_playersSelect.at(2)->currentIndex());
        break;
    case Tournament::Mode::SinglePlayers:
        m_saveScoreButton->setEnabled(
            m_tableSelect->inputOkay()
            && m_tableSelect->currentNumber() != -1
            && m_playersSelect.at(0)->currentIndex() != m_playersSelect.at(1)->currentIndex()
            && m_playersSelect.at(0)->currentIndex() != m_playersSelect.at(2)->currentIndex()
            && m_playersSelect.at(0)->currentIndex() != m_playersSelect.at(3)->currentIndex()
            && m_playersSelect.at(1)->currentIndex() != m_playersSelect.at(2)->currentIndex()
            && m_playersSelect.at(1)->currentIndex() != m_playersSelect.at(3)->currentIndex()
            && m_playersSelect.at(2)->currentIndex() != m_playersSelect.at(3)->currentIndex()
        );
        break;
    }
}

QVector<Database::BoogerResult> ScorePage::currentScore() const
{
    QVector<Database::BoogerResult> score;
    score.reserve(m_scoreSize);
    for (int i = 0; i < 4; i += m_tournamentSettings->isSinglePlayerMode() ? 1 : 2) {
        for (const BoogerWidget *boogerWidget : std::as_const(m_boogerWidgets)) {
            score.append(Database::BoogerResult { m_playersSelect.at(i)->currentNumber(),
                                                  boogerWidget->score(i < 2 ? 1 : 2) });
        }
    }
    return score;
}

void ScorePage::saveScore()
{
    emit statusUpdate(tr("Speichere Spielergebnis …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    const bool showRankingPage = ! m_db->localTournamentStarted();

    const int lastOpenedRound = m_roundSelect->currentNumber();
    const int selectedTable = m_tableSelect->currentNumber();
    if (! m_db->saveScore(lastOpenedRound, selectedTable, currentScore())) {
        return;
    }

    QString tableIdentifier;
    if (! m_noTableNumbers->isChecked()) {
        tableIdentifier = tr("Tisch %1").arg(selectedTable);
    } else {
        switch (m_tournamentSettings->tournamentMode()) {
        case Tournament::Mode::FixedPairs:
            tableIdentifier = formatOpponentsString(m_playersSelect.at(0)->currentText(),
                                                    m_playersSelect.at(2)->currentText());
            break;
        case Tournament::Mode::SinglePlayers:
            tableIdentifier = formatOpponentsString(m_playersSelect.at(0)->currentText()
                                                    + m_settings->namesSeparator()
                                                    + m_playersSelect.at(1)->currentText(),
                                                    m_playersSelect.at(2)->currentText()
                                                    + m_settings->namesSeparator()
                                                    + m_playersSelect.at(3)->currentText());
            break;
        }
    }

    if (showRankingPage) {
        emit tournamentStarted();
    }

    emit scoreChanged();
    reload(lastOpenedRound);

    // Make sure the entered score is visible

    int targetRow = 0;

    for (int i = 0; i < m_scoreList->rowCount(); i += m_resultsAdvance) {
        if (m_scoreList->item(i, 0)->text().toInt() == selectedTable) {
            targetRow = i;
            break;
        }
    }

    if (targetRow == m_scoreList->rowCount() - m_resultsAdvance) {
        m_scoreList->scrollToBottom();
    } else {
        m_scoreList->scrollToItem(m_scoreList->item(targetRow, 1),
                                  QAbstractItemView::PositionAtCenter);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Spielergebnis für %1 in Runde %2 gespeichert").arg(
                         tableIdentifier, QString::number(lastOpenedRound)));
}

void ScorePage::showContextMenu(QPoint point)
{
    // This is also triggered by a double click, so we have to check for the ContextMenuPolicy
    if (m_scoreList->contextMenuPolicy() == Qt::NoContextMenu) {
        return;
    }

    QTableWidgetItem *selectedScore = m_scoreList->itemAt(point);
    if (selectedScore == nullptr) {
        return;
    }

    int row = selectedScore->row();
    while (m_scoreList->item(row, 0) == nullptr) {
        row--;
    }
    m_selectedTable = m_scoreList->item(row, 0)->text().toInt();

    if (m_noTableNumbers->isChecked()) {
        m_contextMenu->setTitle(formatOpponentsString(
            m_scoreList->item(row, m_pair1Column)->text(),
            m_scoreList->item(row + m_pair2RowOffset, m_pair2Column)->text()));
    } else {
        m_contextMenu->setTitle(tr("Ergebnis für Tisch %1").arg(m_selectedTable));
    }

    point.setY(point.y() + m_scoreList->horizontalHeader()->height());
    m_contextMenu->exec(m_scoreList->mapToGlobal(point));
}

void ScorePage::deleteScore()
{
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("Spielergebnis löschen"));
    messageBox.setText(tr("Soll das Spielergebnis von %1 wirklich gelöscht werden?").arg(
                          getStatusUpdateIdentifier()));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        return;
    }

    doDeleteScore(m_roundSelect->currentNumber(), m_selectedTable);
}

void ScorePage::doDeleteScore(int round, int table)
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche Spielergebnis …"));

    if (! m_db->deleteScore(round, table)) {
        return;
    }

    emit statusUpdate(tr("Spielergebnis für %1 in Runde %2 gelöscht").arg(
                         getStatusUpdateIdentifier(), QString::number(round)));
    reload(round);
    QApplication::restoreOverrideCursor();

    if (! m_db->localTournamentStarted()) {
        // Re-emitting this will cause the players page to be updated again,
        // as it checks for db.tournamentStarted() itself
        emit tournamentStarted();
    }
    emit scoreChanged();
}

void ScorePage::editScore()
{
    const int round = m_roundSelect->currentNumber();

    const auto [ score, success ] = m_db->singleScore(round, m_selectedTable);
    if (! success) {
        return;
    }

    doDeleteScore(round, m_selectedTable);
    fillInScore(round, m_selectedTable, score);
}

void ScorePage::fillInScore(int round, int table, const QVector<Database::BoogerResult> &score)
{
    m_roundSelect->setCurrentNumber(round);

    blockSelectSignals(true);
    m_tableSelect->setCurrentNumber(table);
    switch (m_tournamentSettings->tournamentMode()) {
    case Tournament::Mode::FixedPairs:
        m_playersSelect.at(0)->setCurrentNumber(
            score.at(0).pairOrPlayer);
        m_playersSelect.at(2)->setCurrentNumber(
            score.at(m_tournamentSettings->boogersPerRound()).pairOrPlayer);
        break;
    case Tournament::Mode::SinglePlayers:
        for (int i = 0; i < 4; i++) {
            m_playersSelect.at(i)->setCurrentNumber(
                score.at(m_tournamentSettings->boogersPerRound() * i).pairOrPlayer);
        }
        break;
    }
    blockSelectSignals(false);

    // Get the scores to set
    QVector<int> scores;
    QVectorIterator<Database::BoogerResult> boogerResult(score);
    while (boogerResult.hasNext()) {
        scores.append(boogerResult.next().score);
        for (int i = 1; i < m_tournamentSettings->boogersPerRound(); i++) {
            scores.append(boogerResult.next().score);
        }
        if (m_tournamentSettings->isSinglePlayerMode()) {
            // Leave out the double scores
            for (int i = 0; i < m_tournamentSettings->boogersPerRound(); i++) {
                boogerResult.next();
            }
        }
    }

    // Check if we have an aborted score
    int abortedBooger = -1;
    for (int i = 0; i < m_boogerWidgets.count(); i++) {
        if (scores.at(i) < m_tournamentSettings->boogerScore()
            && scores.at(i + scores.count() / 2) < m_tournamentSettings->boogerScore()) {

            abortedBooger = i;
            break;
        }
    }

    // First reset all boogers
    for (BoogerWidget *booger : std::as_const(m_boogerWidgets)) {
        booger->reset();
    }

    // Then mark the first aborted booger if we have one
    if (abortedBooger > -1) {
        boogerAbortedStateChanged(abortedBooger, true);
        m_boogerWidgets.at(abortedBooger)->setAborted(true);
        m_boogerWidgets.at(abortedBooger)->setAsFirstAborted();
    }

    // Pre-set all scores
    for (int i = 0; i < m_boogerWidgets.count(); i++) {
        m_boogerWidgets.at(i)->setScore(scores.at(i), scores.at(i + scores.count() / 2));
    }

    checkSavePossible();
}

void ScorePage::blockSelectSignals(bool state)
{
    m_roundSelect->blockSignals(state);
    m_tableSelect->blockSignals(state);
    for (int i = 0; i < 4; i += m_tournamentSettings->isSinglePlayerMode() ? 1 : 2) {
        m_playersSelect.at(i)->blockSignals(state);
    }
}

void ScorePage::checkEnterScore()
{
    if (! m_saveScoreButton->isEnabled()) {
        return;
    }
    saveScore();
}

void ScorePage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_scoreList->rowCount(); row++) {
            m_scoreList->showRow(row);
        }
        m_scoreList->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    m_searchEngine->setSearchTerm(searchText);
    const SearchEngine::SearchType searchType
        = m_tournamentSettings->isPairMode() ? SearchEngine::SearchType::Default
                                             : SearchEngine::SearchType::DrawnPairsScore;
    const bool phoneticSearch = m_searchWidget->phoneticSearch();

    int matches = 0;
    for (int row = 0; row < m_scoreList->rowCount(); row += m_resultsAdvance) {
        const bool rowMatches =
               m_searchEngine->checkMatch(m_scoreList->item(
                   row, m_pair1Column)->text(), searchType, phoneticSearch)
            || m_searchEngine->checkMatch(m_scoreList->item(
                   row + m_pair2RowOffset, m_pair2Column)->text(), searchType, phoneticSearch);

        for (int i = row; i < row + m_resultsAdvance; i++) {
            m_scoreList->setRowHidden(i, ! rowMatches);
        }

        matches += rowMatches;
    }

    m_searchWidget->updateMatches(matches);
    m_scoreList->resizeColumnsToContents();
    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText.simplified()));
    m_isSearchLabel->show();
}

void ScorePage::swapPairs()
{
    const int round = m_roundSelect->currentNumber();

    auto [ score, success ] = m_db->singleScore(round, m_selectedTable);
    if (! success) {
        return;
    }

    if (! m_db->deleteScore(round, m_selectedTable)) {
        return;
    }

    QVector<Database::BoogerResult> newScore = score.mid(score.count() / 2, score.count() / 2);
    newScore.append(score.mid(0, score.count() / 2));
    if (! m_db->saveScore(round, m_selectedTable, newScore)) {
        return;
    }

    emit statusUpdate(tr("Paar 1 und Paar 2 für %1 in Runde %2 vertauscht").arg(
                         getStatusUpdateIdentifier(), QString::number(round)));

    // This will reload the page and restore the currently entered score
    playersNameEdited();
}

void ScorePage::noTableNumbersChanged()
{
    const bool state = m_noTableNumbers->isChecked();

    m_tableLabel->setVisible(! state);
    m_tableSelect->setVisible(! state);
    m_tableSelect->setCurrentIndex(state);

    m_scoreList->setColumnHidden(0, state);
    if (! state) {
        m_scoreList->resizeColumnsToContents();
    }

    if (state) {
        m_lastAutoSelectState = m_autoSelectPairs->isChecked();
        m_autoSelectPairs->setChecked(false);
        m_autoSelectPairs->setEnabled(false);
    } else {
        m_autoSelectPairs->setChecked(m_lastAutoSelectState);
        m_autoSelectPairs->setEnabled(true);
    }

    m_tournamentSettings->setNoTableNumbers(state);
    checkSavePossible();
}

QString ScorePage::getStatusUpdateIdentifier() const
{
    return m_noTableNumbers->isChecked() ? m_contextMenu->title()
                                         : tr("Tisch %1").arg(m_selectedTable);
}

void ScorePage::autoSelectPairsChanged()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    const auto state = m_autoSelectPairs->isChecked();
    m_tournamentSettings->setAutoSelectPairs(state);
    m_selectByLastRound->setEnabled(state);
    m_noAutoselect->setVisible(state);
    if (state) {
        updateAssignment(m_roundSelect->currentNumber());
        tableChanged(m_tableSelect->currentNumber());
    }
    QApplication::restoreOverrideCursor();
}

void ScorePage::selectByLastRoundChanged()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    const auto state = m_selectByLastRound->isChecked();
    m_tournamentSettings->setSelectByLastRound(state);
    updateAssignment(m_roundSelect->currentNumber());
    tableChanged(m_tableSelect->currentNumber());
    QApplication::restoreOverrideCursor();
}

QString ScorePage::formatOpponentsString(const QString &pair1, const QString &pair2) const
{
    if (pair1.size() <= 15 && pair2.size() <= 15) {
        return tr("„%1“ gegen „%2“").arg(pair1, pair2);
    } else if (pair1.size() > 15 && pair2.size() <= 15) {
        return tr("„%1…“ gegen „%2“").arg(pair1.left(15), pair2);
    } else if (pair1.size() <= 15 && pair2.size() > 15) {
        return tr("„%1“ gegen „%2…“").arg(pair1, pair2.left(15));
    } else {
        return tr("„%1…“ gegen „%2…“").arg(pair1.left(15), pair2.left(15));
    }
}

void ScorePage::setNoTableNumbers(bool state)
{
    m_noTableNumbers->setChecked(state);
}

void ScorePage::setAutoSelectState(bool enabled, bool byLastRound)
{
    m_autoSelectPairs->setChecked(enabled);
    m_selectByLastRound->setChecked(byLastRound);
}

void ScorePage::boogerAbortedStateChanged(int booger, bool aborted)
{
    if (aborted) {
        for (int i = booger + 1; i < m_boogerWidgets.count(); i++) {
            m_boogerWidgets.at(i)->setAborted(true);
        }
    } else {
        for (int i = 0; i <= booger; i++) {
            if (m_boogerWidgets.at(i)->isAborted()) {
                m_boogerWidgets.at(i)->reset();
            }
        }
        if (booger + 1 < m_boogerWidgets.count()) {
            m_boogerWidgets.at(booger + 1)->setAsFirstAborted();
        }
    }
}

void ScorePage::resetSearch()
{
    m_searchWidget->clear();
}

void ScorePage::showScorePopup(BoogerSpinBox *spinBox)
{
    m_targetSpinBox = spinBox;
    m_scorePopup->show();
}

void ScorePage::setSelectedScore()
{
    m_targetSpinBox->setValue(m_scorePopup->score());
    m_targetSpinBox->setFocus();
    m_targetSpinBox->selectAll();
}

void ScorePage::scoreTypeChanged(Tournament::ScoreType scoreType)
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    const int round = m_roundSelect->currentNumber();
    m_tournamentSettings->setScoreType(scoreType);
    setupGui();
    openRound(round);
    QApplication::restoreOverrideCursor();
}

QString ScorePage::scoreChecksum(const Scores::Score &score) const
{
    if (score.tables.count() == 0) {
        return QString();
    }

    ChecksumHelper checksumHelper;

    for (const int table : score.tables) {
        checksumHelper.addData(table);
    }

    for (const QString &pair : score.pairs) {
        checksumHelper.addData(pair);
    }

    for (const auto &pairResults : score.scores) {
        for (const int result : pairResults) {
            checksumHelper.addData(result);
        }
    }

    return checksumHelper.checksum();
}

const QMap<int, QString> &ScorePage::checksums() const
{
    return m_checksums;
}

void ScorePage::setIsServer(bool state)
{
    if (state) {
        m_isServer = true;

        QApplication::setOverrideCursor(Qt::BusyCursor);

        for (int round = 1; round <= m_db->roundsState().count(); round++) {
            const auto [ score, success ] = m_db->scores(round);
            if (! success) {
                QApplication::restoreOverrideCursor();
                return;
            }

            m_checksums[round] = scoreChecksum(score);
        }

        processChangedChecksums(m_checksums);

        QApplication::restoreOverrideCursor();

    } else {
        m_isServer = false;
        m_checksums.clear();
    }
}

void ScorePage::setIsClient(bool state)
{
    m_isClient = state;
    m_lastChange->setVisible(state);
    m_syncStatus->setVisible(state);

    if (! state) {
        m_remoteChecksums.clear();
    }
}

void ScorePage::updateLastChange(QString &cache)
{
    cache = QTime::currentTime().toString(QStringLiteral("hh:mm:ss"));
    if (m_lastChange->isVisible()) {
        m_lastChange->setText(tr("<table>"
                                 "<tr><td colspan=\"2\"><b>Letzte Änderung</b></td></tr>"
                                 "<tr><td>Server:</td><td>%1</td></tr>"
                                 "<tr><td>Lokal:</td><td>%2</td></tr>"
                                 "</table>").arg(m_lastServerChange, m_lastLocalChange));
    }
}

void ScorePage::remoteChecksumChanged(int round, const QString &checksum)
{
    m_remoteChecksums[round] = checksum;
    updateLastChange(m_lastServerChange);

    if (m_roundSelect->currentNumber() == round) {
        updateSyncStatus();
    }

    processChangedChecksums(m_remoteChecksums);
}

void ScorePage::processChangedChecksums(const QMap<int, QString> &checksums) const
{
    ChecksumHelper checksumHelper;
    QMapIterator<int, QString> it(checksums);
    while (it.hasNext()) {
        it.next();
        checksumHelper.addData(it.key());
        checksumHelper.addData(it.value());
    }
    emit checksumChanged(checksumHelper.checksum());
}

void ScorePage::updateSyncStatus()
{
    const int round = m_roundSelect->currentNumber();
    m_syncStatus->setText(m_checksums.value(round) == m_remoteChecksums.value(round)
        ? tr("<span style=\"%1\">Ergebnisse sind<br/>identisch</span>").arg(
             SharedStyles::GREEN_TEXT)
        : tr("<p><span style=\"%1\">Ergebnisse sind<br/>nicht identisch</span></p>"
             "<p><a href=\"#showCompareScoreDialog\">Unterschiede<br/>anzeigen</a></p>").arg(
             SharedStyles::RED_TEXT));
}

const int *ScorePage::currentRound() const
{
    return &m_currentRound;
}

void ScorePage::processLinkActivated(const QString &anchor) const
{
    if (anchor == QStringLiteral("#showCompareScoreDialog")) {
        emit showCompareScoreDialog();
    } else if (anchor == QStringLiteral("#finishRegistration")) {
        emit finishRegistration();
    }
}
