/* Copyright (c) 2019-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes
#include "TableComboBox.h"
#include "TableLineEdit.h"

// Qt includes
#include <QDebug>
#include <QIntValidator>

TableComboBox::TableComboBox(QWidget *parent) : NumbersComboBox(parent)
{
    setEditable(true);
    setInsertPolicy(QComboBox::NoInsert);
    connect(this, QOverload<int>::of(&QComboBox::activated), this, &TableComboBox::tableSelected);

    m_lineEdit = new TableLineEdit;
    m_validator = new QIntValidator(this);
    m_lineEdit->setValidator(m_validator);
    setLineEdit(m_lineEdit);
    disconnect(m_lineEdit, &QLineEdit::editingFinished, 0, 0);
    connect(m_lineEdit, &QLineEdit::textEdited, this, &TableComboBox::lineEditTextChanged);
}

void TableComboBox::clear()
{
    QComboBox::clear();

    m_inputOkay = true;
    m_lineEdit->displayInvalidInput(false);

    m_validator->setBottom(0);
    m_validator->setTop(0);

    // Add a "no table selected" dummy entry ...
    addItem(m_lineEdit->noItemText(), -1);

    // ... and make it unselectable, both for the popup (hide it) ...
    qobject_cast<QListView *>(view())->setRowHidden(0, true);

    // ... and the mouse wheel/arrow keys selection (disable it).
    QStandardItemModel *itemModel = qobject_cast<QStandardItemModel *>(model());
    QStandardItem *item = itemModel->item(0);
    item->setFlags(item->flags() & ~Qt::ItemIsEnabled);
}

void TableComboBox::addItem(const QString &text, const QVariant &userData)
{
    QComboBox::addItem(text, userData);
    const int number = userData.toInt();
    const int bottom = m_validator->bottom();

    if (bottom == 0) {
        m_validator->setBottom(number);
        m_validator->setTop(number);
    } else {
        if (bottom > number) {
            m_validator->setBottom(number);
        }
        if (m_validator->top() < number) {
            m_validator->setTop(number);
        }
    }
}

bool TableComboBox::setCurrentNumber(int number)
{
    if (! NumbersComboBox::setCurrentNumber(number)) {
        return false;
    }

    if (! m_inputOkay) {
        m_inputOkay = true;
        m_lineEdit->displayInvalidInput(false);
    }

    return true;
}

void TableComboBox::lineEditTextChanged(const QString &text)
{
    const int index = findData(text.toInt());

    if (index == -1) {
        m_lineEdit->displayInvalidInput(true);
        // This will disable the save button:
        m_inputOkay = false;
        emit checkSavePossible();
        return;
    }

    if (! m_inputOkay) {
        m_inputOkay = true;
        m_lineEdit->displayInvalidInput(false);
    }

    if (currentIndex() != index) {
        setCurrentIndex(index);
    } else {
        // The index didn't change, thus, numberChanged() won't be emitted. But the save button is
        // still disabled, so we have to trigger ScorePage::checkSavePossible() manually here to
        // re-enable it
        emit checkSavePossible();
    }
}

bool TableComboBox::inputOkay() const
{
    return m_inputOkay;
}

void TableComboBox::tableSelected()
{
    m_lineEdit->displayInvalidInput(false);

    if (! m_inputOkay) {
        // If m_inputOkay is false, currentNumberChanged() already triggered
        // ScorePage::checkSavePossible(), but with m_inputOkay still set to false.
        // Thus, we have to emit checkSavePossible() to trigger it again, but with m_inputOkay
        // set to true.
        m_inputOkay = true;
        emit checkSavePossible();
    }
}
