/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes

#include "Server.h"
#include "Connection.h"

#include "network/ServerProtocol.h"

#include "RegistrationPage/RegistrationPage.h"
#include "RegistrationPage/MarkersWidget.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/TournamentSettings.h"
#include "shared/StringTools.h"
#include "shared/JsonHelper.h"
#include "shared/Settings.h"
#include "shared/Scores.h"

// Qt includes
#include <QDebug>
#include <QTimer>
#include <QJsonArray>
#include <QApplication>

// C++ includes
#include <functional>
#include <utility>

using namespace ServerProtocol;

Server::Server(QObject *parent, int connectTimeout, SharedObjects *sharedObjects,
               RegistrationPage *registrationPage, const QMap<int, QString> *scoreChecksums)
    : QTcpServer(parent),
      m_connectTimeout(connectTimeout),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_registrationPage(registrationPage),
      m_markersWidget(registrationPage->markersWidget()),
      m_scoreChecksums(scoreChecksums),
      m_locale(sharedObjects->stringTools()->locale()),
      m_requestHandlers {
          // Registration changes

          { ACTION_PLAYERS_REGISTER,           &Server::processRegisterPlayers },
          { ACTION_PLAYERS_RENAME,             &Server::processRenamePlayers },
          { ACTION_PLAYERS_ASSEMBLE_PAIR,      &Server::processAssemblePair },
          { ACTION_PLAYERS_DELETE,             &Server::processDeletePlayers },
          { ACTION_PLAYERS_UPDATE_MARKER,      &Server::processUpdatePlayersMarker },

          { ACTION_PLAYERS_DELETE_MARKED,      &Server::processDeleteMarkedPlayers },
          { ACTION_PLAYERS_DELETE_UNMARKED,    &Server::processDeleteUnmarkedPlayers },
          { ACTION_PLAYERS_DELETE_ALL,         &Server::processDeleteAllPlayers },
          { ACTION_PLAYERS_LIST_UPDATE_MARKER, &Server::processUpdateListMarker },
          { ACTION_PLAYERS_LIST_REMOVE_MARKER, &Server::processRemoveListMarker },

          // Draw changes
          { ACTION_UPDATE_DRAW,      &Server::processUpdateDraw },
          { ACTION_SET_LIST_DRAW,    &Server::processSetListDraw },
          { ACTION_DELETE_ALL_DRAWS, &Server::processDeleteAllDraws },

          // Markers defintion changes
          { ACTION_MARKERS_ADD,    &Server::processAddMarker },
          { ACTION_MARKERS_DELETE, &Server::processDeleteMarker },
          { ACTION_MARKERS_MOVE,   &Server::processMoveMaker },
          { ACTION_MARKERS_EDIT,   &Server::processEditMarker },

          // Special markers changes
          { ACTION_SET_ABSENT_MARKER,         &Server::processSetAbsentMarker },
          { ACTION_CHANGE_SINGLES_MANAGEMENT, &Server::processChangeSinglesManagement },

          // Finish registration
          { ACTION_FINISH_REGISTRATION, &Server::processFinishRegistration }

          // Ranking comparison
          // (Can't be requested by the server)

          // Score comparison
          // (Can't be requested by the server)
      }
{
}

void Server::setChecksum(const QString *checksum)
{
    m_checksum = checksum;
}

void Server::close()
{
    for (Connection *connection : std::as_const(m_connections)) {
        connection->close();
        removeConnection(connection);
    }
    QTcpServer::close();
}

void Server::incomingConnection(qintptr descriptor)
{
    // Create a new connection for the connection request
    Connection *connection = new Connection(this);
    if (! connection->setSocketDescriptor(descriptor)) {
        connection->deleteLater();
        return;
    }
    connect(connection, &Connection::disconnected,
            this, std::bind(&Server::removeConnection, this, connection));
    connect(connection, &Connection::handshakeResult,
            this, std::bind(&Server::processHandshake, this, connection, std::placeholders::_1));
    connect(connection, &Connection::clientRequest,
            this, std::bind(&Server::processClientRequest, this, connection,
                            std::placeholders::_1, std::placeholders::_2));
    connect(connection, &AbstractNetworkEngine::protocolError, this, &Server::protocolError);

    // Initiate a handshake

    m_pendingConnections.append(connection);
    emit message(tr("Verbindungsanfrage von %1, sende Handshake").arg(connection->ip()));

    // Send greeting message
    connection->sendData(HANDSHAKE_MESSAGE_INITIALIZE);

    // Start a concurrent timeout call
    QTimer::singleShot(m_connectTimeout, this,
                       std::bind(&Server::checkHandshakeTimeout, this, connection));
}

void Server::removeConnection(Connection *connection)
{
    if (m_pendingConnections.removeOne(connection)) {
        emit message(tr("Verbindung mit %1 beendet").arg(connection->ip()));
    } else if (m_connections.removeOne(connection)) {
        emit message(tr("Verbindung mit %1 beendet").arg(clientString(connection)));
        emit statusUpdate(tr("Verbindung mit %1 beendet").arg(clientString(connection)));
        emit clientCountChanged(m_connections.count());
    }
    connection->deleteLater();
}

void Server::checkHandshakeTimeout(Connection *connection)
{
    if (m_pendingConnections.contains(connection)) {
        emit message(tr("Keine Antwort von %1").arg(connection->ip()));
        removeConnection(connection);
    }
}

void Server::broadcastMessage(QJsonObject &jsonObject, int clientId, const QString &messageText)
{
    emit tx();
    jsonObject.insert(KEY_CHECKSUM, *m_checksum);
    jsonObject.insert(KEY_CLIENT_ID, clientId);
    for (Connection *connection : std::as_const(m_connections)) {
        connection->sendData(jsonObject);
    }
    emit message(messageText);
}

void Server::sendSingleReply(QJsonObject &jsonObject, Connection *connection,
                             const QString &messageText)
{
    jsonObject.insert(KEY_CHECKSUM, *m_checksum);
    connection->sendData(jsonObject);
    emit message(messageText);
}

int Server::clientCount() const
{
    return m_connections.count();
}

QString Server::clientString(Connection *connection) const
{
    return QStringLiteral("%1 (Client-ID %2)").arg(connection->ip(),
                                                   QString::number(connection->id()));
}

// =========
// Handshake
// =========

void Server::processHandshake(Connection *connection, const QString &phase)
{
    if (! m_pendingConnections.contains(connection)) {
        return;
    }

    if (phase == HANDSHAKE_PHASE_SYNCHRONIZE) {
        emit message(tr("Synchronisation von %1 angefragt, sende Daten").arg(connection->ip()));

        const QJsonObject settings {
            { HANDSHAKE_KEY_LOCALE,             m_locale },
            { HANDSHAKE_KEY_NAMES_SEPARATOR,    m_settings->namesSeparator() },
            { HANDSHAKE_KEY_TOURNAMENT_MODE,    static_cast<int>(
                                                    m_tournamentSettings->tournamentMode()) },
            { HANDSHAKE_KEY_BOOGER_SCORE,       m_tournamentSettings->boogerScore() },
            { HANDSHAKE_KEY_BOOGERS_PER_ROUND,  m_tournamentSettings->boogersPerRound() },
            { HANDSHAKE_KEY_TOURNAMENT_STARTED, m_db->networkTournamentStarted() } };

        const auto [ definedMarkers, success1 ] = m_db->markers();
        if (! success1) {
            return;
        }

        QJsonArray markers;
        for (const Markers::Marker &marker : definedMarkers) {
            markers.append(QJsonObject {
                { KEY_MARKER_ID,      marker.id },
                { KEY_MARKER_NAME,    marker.name },
                { KEY_MARKER_COLOR,   Connection::colorToString(marker.color) },
                { KEY_MARKER_SORTING, static_cast<int>(marker.sorting) }
            });
        }

        const auto [ playersData, success2 ] = m_db->playersData();
        if (! success2) {
            return;
        }

        QJsonArray players;
        const auto playersDataKeys = playersData.keys();
        for (const QString &name : playersDataKeys) {
            const Database::PlayersData data = playersData.value(name);
            QJsonObject jsonData;
            if (data.marker != 0) {
                jsonData.insert(KEY_MARKER_ID, data.marker);
            }
            if (data.assignedTable != 0) {
                jsonData.insert(KEY_PLAYERS_TABLE, data.assignedTable);
                jsonData.insert(KEY_PLAYERS_PAIR_NUMBER, data.assignedPairNumber);
            }
            players.append(QJsonObject { { name, jsonData } });
        }

        QJsonObject scoreChecksums;
        QMapIterator<int, QString> it(*m_scoreChecksums);
        while (it.hasNext()) {
            it.next();
            scoreChecksums.insert(QString::number(it.key()), it.value());
        }

        connection->sendData(QJsonObject {
            { KEY_ACTION,               ACTION_HANDSHAKE },
            { HANDSHAKE_KEY_PHASE,      HANDSHAKE_PHASE_SYNCHRONIZE },
            { HANDSHAKE_KEY_SETTINGS,   settings },
            { KEY_MARKERS_LIST,         markers },
            { KEY_MARKER_ABSENT,        m_tournamentSettings->marker(Markers::Absent) },
            { KEY_SINGLES_MANAGEMENT,   JsonHelper::singlesManagementToJson(
                                           m_tournamentSettings->singlesManagement()) },
            { KEY_PLAYERS_LIST,         players },
            { KEY_SCORE_CHECKSUMS_LIST, scoreChecksums },
            { KEY_CHECKSUM,             *m_checksum }
        });

    } else if (phase == HANDSHAKE_PHASE_ALLOCATE_ID) {
        m_clientIdCounter++;
        connection->setId(m_clientIdCounter);
        connection->sendData(QJsonObject { { KEY_ACTION,          ACTION_HANDSHAKE },
                                           { HANDSHAKE_KEY_PHASE, HANDSHAKE_PHASE_ALLOCATE_ID },
                                           { KEY_CLIENT_ID,       m_clientIdCounter } });
        m_pendingConnections.removeOne(connection);
        m_connections.append(connection);
        emit message(tr("Synchronisation mit %1 war erfolgreich, Verbindung hergestellt "
                        "(Client-ID %2)").arg(connection->ip(), QString::number(connection->id())));
        emit statusUpdate(tr("Verbindung mit %1 hergestellt (Client-ID %2)").arg(
                             connection->ip(), QString::number(connection->id())));
        emit clientCountChanged(m_connections.count());
    }
}

// ====================
// Changes broadcasting
// ====================

// Registration changes
// ====================

void Server::broadcastRegisterPlayers(const QString &name, int marker, int clientId)
{
    QJsonObject message { { KEY_ACTION,       ACTION_PLAYERS_REGISTER },
                          { KEY_PLAYERS_NAME, name },
                          { KEY_MARKER_ID,    marker } };
    broadcastMessage(message, clientId, tr("Sende Anmelden von „%1“").arg(name));
}

void Server::broadcastRenamePlayers(const QString &name, const QString &newName,
                                    int markerId, const QString &markerName, int clientId)
{
    QJsonObject message { { KEY_ACTION,           ACTION_PLAYERS_RENAME },
                          { KEY_PLAYERS_NAME,     name },
                          { KEY_PLAYERS_NEW_NAME, newName },
                          { KEY_MARKER_ID,        markerId },
                          { KEY_MARKER_NAME,      markerName } };

    QString markerChange;
    if (markerId != -1) {
        if (markerId == 0) {
            markerChange = tr(" und Entfernen der Markierung");
        } else {
            markerChange = tr(" und Setzen der Markierung auf „%1“").arg(markerName);
        }
    }

    broadcastMessage(message, clientId, tr("Sende Umbenennen von „%1“ zu „%2“%3").arg(
                                           name, newName, markerChange));
}

void Server::broadcastAssemblePair(const QString &player1, const QString &player2,
                                   const QString &separator, int assignedMarker, int clientId)
{
    QJsonObject message { { KEY_ACTION,       ACTION_PLAYERS_ASSEMBLE_PAIR },
                          { KEY_PLAYERS_NAME, QJsonArray { player1, player2, separator } },
                          { KEY_MARKER_ID,    assignedMarker } };
    broadcastMessage(message, clientId, tr("Sende Zusammenfügen von „%1“ und „%2“ zu einem "
                                           "Paar").arg(player1, player2));
}

void Server::broadcastDeletePlayers(const QString &name, int clientId)
{
    QJsonObject message { { KEY_ACTION,       ACTION_PLAYERS_DELETE },
                          { KEY_PLAYERS_NAME, name } };
    broadcastMessage(message, clientId, tr("Sende Löschen von „%1“").arg(name));
}

void Server::broadcastUpdateMarker(const QString &name, const QString &markerName, int markerId,
                                   int clientId)
{
    QJsonObject message { { KEY_ACTION,       ACTION_PLAYERS_UPDATE_MARKER },
                          { KEY_PLAYERS_NAME, name },
                          { KEY_MARKER_NAME,  markerName },
                          { KEY_MARKER_ID,    markerId } };
    if (markerId != 0) {
        broadcastMessage(message, clientId,
                        tr("Sende Markieren von „%1“ als „%2“").arg(name, markerName));
    } else {
        broadcastMessage(message, clientId,
                         tr("Sende Entfernen der Markierung von „%1“").arg(name));
    }
}

void Server::broadcastDeleteMarkedPlayers(const QString &markerName, int markerId, int clientId)
{
    QJsonObject message { { KEY_ACTION,      ACTION_PLAYERS_DELETE_MARKED },
                          { KEY_MARKER_NAME, markerName },
                          { KEY_MARKER_ID,   markerId } };
    broadcastMessage(message, clientId,
                     tr("Sende Löschen aller als „%1“ markierten %2").arg(
                        markerName, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                       : tr("Spieler")));
}

void Server::broadcastDeleteUnmarkedPlayers(int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_DELETE_UNMARKED } };
    broadcastMessage(message, clientId, tr("Sende Löschen aller unmarkierten %2").arg(
                                           m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                              : tr("Spieler")));
}

void Server::broadcastDeleteAllPlayers(int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_DELETE_ALL } };
    broadcastMessage(message, clientId, tr("Sende Löschen aller %2").arg(
                                           m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                              : tr("Spieler")));
}

void Server::broadcastListUpdateMarker(const QString &markerName, int markerId, int clientId)
{
    QJsonObject message { { KEY_ACTION,      ACTION_PLAYERS_LIST_UPDATE_MARKER },
                          { KEY_MARKER_NAME, markerName },
                          { KEY_MARKER_ID,   markerId } };

    if (markerId != 0) {
        broadcastMessage(message, clientId,
                         tr("Sende Markieren aller %1 als „%2“").arg(
                            m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                            markerName));
    } else {
        broadcastMessage(message, clientId, tr("Sende Entfernen aller Markierungen"));
    }
}

void Server::broadcastListRemoveMarker(const QString &markerName, int markerId, int clientId)
{
    QJsonObject message { { KEY_ACTION,      ACTION_PLAYERS_LIST_REMOVE_MARKER },
                          { KEY_MARKER_NAME, markerName },
                          { KEY_MARKER_ID,   markerId } };
    broadcastMessage(message, clientId,
                     tr("Sende Entfernen aller „%1“-Markierungen").arg(markerName));
}

// Draw changes
// ============

void Server::broadcastUpdateDraw(const QString &name, const Draw::Seat &seat, int marker,
                                 int clientId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_UPDATE_DRAW },
                          { KEY_PLAYERS_NAME,        name },
                          { KEY_PLAYERS_TABLE,       seat.table },
                          { KEY_PLAYERS_PAIR_NUMBER, seat.pair },
                          { KEY_MARKER_ID,           marker } };
    const QString text = seat.isEmpty() ? tr("Sende Löschen der Auslosung für „%1“").arg(name)
                                        : tr("Sende Speichern der Auslosung für „%1“").arg(name);
    broadcastMessage(message, clientId, text);
}

void Server::broadcastSetListDraw(const QVector<QString> &names, const QVector<Draw::Seat> &draw,
                                  int clientId)
{
    QJsonArray namesList;
    QJsonArray tables;
    QJsonArray pairs;
    for (int i = 0; i < names.count(); i++) {
        namesList.append(names.at(i));
        tables.append(draw.at(i).table);
        pairs.append(draw.at(i).pair);
    }

    QJsonObject message { { KEY_ACTION,            ACTION_SET_LIST_DRAW },
                          { KEY_PLAYERS_LIST,      namesList },
                          { KEY_TABLES_LIST,       tables },
                          { KEY_PAIR_NUMBERS_LIST, pairs } };

    broadcastMessage(message, clientId, tr("Sende Setzen der Auslosung aller Plätze"));
}

void Server::broadcastDeleteAllDraws(int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_DELETE_ALL_DRAWS } };
    broadcastMessage(message, clientId, tr("Sende Löschen aller Auslosungen"));
}

// Markers definition changes
// ==========================

void Server::broadcastAddMarker(const QString &name, const QColor &color,
                                Markers::MarkerSorting sorting, int clientId)
{
    QJsonObject message { { KEY_ACTION,         ACTION_MARKERS_ADD },
                          { KEY_MARKER_NAME,    name },
                          { KEY_MARKER_COLOR,   Connection::colorToString(color) },
                          { KEY_MARKER_SORTING, sorting } };
    broadcastMessage(message, clientId, tr("Sende Hinzufügen der Markierung „%1“").arg(name));
}

void Server::broadcastDeleteMarker(const QString &name, int id, int sequence, int clientId)
{
    QJsonObject message { { KEY_ACTION,          ACTION_MARKERS_DELETE },
                          { KEY_MARKER_NAME,     name },
                          { KEY_MARKER_ID,       id },
                          { KEY_MARKER_SEQUENCE, sequence } };
    broadcastMessage(message, clientId, tr("Sende Hinzufügen der Markierung „%1“").arg(name));
}

void Server::broadcastMoveMarker(const QString &name, int id, int sequence, int direction,
                                 int clientId)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_MOVE },
                          { KEY_MARKER_NAME,      name },
                          { KEY_MARKER_ID,        id },
                          { KEY_MARKER_SEQUENCE,  sequence },
                          { KEY_MARKER_DIRECTION, direction } };
    broadcastMessage(message, clientId,
                     tr("Sende Verschieben der Markierung „%1“ %2").arg(
                        name, direction == 1 ? tr("nach unten") : tr("nach oben")));
}

void Server::broadcastEditMarker(const QString &name, int id, const QString &newName,
                                 const QColor &color, Markers::MarkerSorting sorting,
                                 int clientId)
{
    QJsonObject message { { KEY_ACTION,          ACTION_MARKERS_EDIT },
                          { KEY_MARKER_NAME,     name },
                          { KEY_MARKER_ID,       id },
                          { KEY_MARKER_NEW_NAME, newName },
                          { KEY_MARKER_COLOR,    Connection::colorToString(color) },
                          { KEY_MARKER_SORTING,  sorting } };
    broadcastMessage(message, clientId, tr("Sende Änderung der Markierung „%1“").arg(name));
}

// Special markers changes
// =======================

void Server::broadcastSetAbsentMarker(const QString &name, int id, int clientId)
{
    QJsonObject message { { KEY_ACTION,      ACTION_SET_ABSENT_MARKER },
                          { KEY_MARKER_NAME, name },
                          { KEY_MARKER_ID,   id } };
    const QString text = id == -1
        ? tr("Sende Deaktivierung der Markierung für abwesende Anmeldungen")
        : tr("Sende Setzen der Markierung für abwesende Anmeldungen auf „%1“").arg(name);
    broadcastMessage(message, clientId, text);
}

void Server::broadcastSinglesManagementChange(const Markers::SinglesManagement &data,
                                              int clientId)
{
    QJsonObject message {
            { KEY_ACTION,             ACTION_CHANGE_SINGLES_MANAGEMENT },
            { KEY_SINGLES_MANAGEMENT, JsonHelper::singlesManagementToJson(data) }
    };
    broadcastMessage(message, clientId, tr("Sende Änderung für das Behandeln allein gekommener "
                                           "Spieler"));
}

// Finishing the registration
// ==========================

void Server::broadcastFinishRegistration(int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_FINISH_REGISTRATION } };
    broadcastMessage(message, clientId, tr("Sende Beenden der Anmeldung"));
}

// Ranking comparison
// ==================

// (Can't be requested by the server)

// Score chechsum changes
// ======================

void Server::scoreChecksumChanged(int round)
{
    QJsonObject message { { KEY_ACTION,         ACTION_SCORE_CHANGE },
                          { KEY_ROUND,          round },
                          { KEY_SCORE_CHECKSUM, m_scoreChecksums->value(round) } };
    broadcastMessage(message, 0, tr("Sende Änderung der Ergebnisse in Runde %1").arg(round));
}

// =======================
// Client request handlers
// =======================

void Server::processClientRequest(Connection *connection, const QString &action,
                                  const QJsonObject &jsonObject)
{
    emit rx();

    QApplication::setOverrideCursor(Qt::BusyCursor);

    const QString client = clientString(connection);
    const int clientId = jsonObject.value(KEY_CLIENT_ID).toInt();

    if (*m_checksum != jsonObject.value(KEY_CHECKSUM).toString()) {
        emit message(tr("Ungültige Anfrage von %1: Die Prüfsummen stimmen nicht überein").arg(
                        client));
        connection->sendData(QJsonObject {
            { KEY_ACTION,        ACTION_REJECTED },
            { KEY_REJECT_REASON, static_cast<int>(RejectReason::OutdatedRequest) },
            { KEY_CLIENT_ID,     clientId }
        });
        QApplication::restoreOverrideCursor();
        return;
    }

    // These functions have to be handled separately due to the different needed function arguments

    if (action == ACTION_RANKING_REQUEST) {
        processRankingRequest(connection, client, clientId);
        QApplication::restoreOverrideCursor();
        return;

    } else if (action == ACTION_SCORE_REQUEST) {
        const int round = jsonObject.value(KEY_ROUND).toInt();
        processScoreRequest(round, connection, client, clientId);
        QApplication::restoreOverrideCursor();
        return;
    }

    // All other ones can be handled via the request handlers map

    // Pass the request to the respective handler function
    if (m_requestHandlers.contains(action)) {
        (this->*m_requestHandlers.value(action))(jsonObject, client, clientId);
    } else {
        // This should not happen
        emit message(tr("Ungültige Anfrage von %1: Unbekannte Aktion angefragt").arg(client));
        connection->sendData(QJsonObject {
            { KEY_ACTION,        ACTION_REJECTED },
            { KEY_REJECT_REASON, static_cast<int>(RejectReason::UnknownAction) },
            { KEY_CLIENT_ID,     clientId }
        });
    }

    QApplication::restoreOverrideCursor();
}

// Registration changes
// ====================

void Server::processRegisterPlayers(const QJsonObject &jsonObject, const QString &client,
                                    int clientId)
{
    const QString name = jsonObject.value(KEY_PLAYERS_NAME).toString();
    const int marker = jsonObject.value(KEY_MARKER_ID).toInt();
    emit message(tr("Anfrage zum Anmelden von „%1“ von %2").arg(name, client));

    if (! m_db->registerPlayers(name, marker)) {
        return;
    }
    emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ angemeldet").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name));

    m_registrationPage->processNetworkChange();
    broadcastRegisterPlayers(name, marker, clientId);
}

void Server::processRenamePlayers(const QJsonObject &jsonObject, const QString &client,
                                  int clientId)
{
    const QString name = jsonObject.value(KEY_PLAYERS_NAME).toString();
    const QString newName = jsonObject.value(KEY_PLAYERS_NEW_NAME).toString();
    const int markerId = jsonObject.value(KEY_MARKER_ID).toInt();
    const QString markerName = jsonObject.value(KEY_MARKER_NAME).toString();

    QString markerChangeStart;
    QString markerChangeFinished;
    if (markerId != -1) {
        if (markerId == 0) {
            markerChangeStart = tr(" und Entfernen der Markierung");
            markerChangeFinished = tr(" und die Markierung entfernt");
        } else {
            markerChangeStart = tr(" und Setzen der Markierung auf „%1“").arg(markerName);
            markerChangeFinished = tr(" und die Markierung auf „%1“ gesetzt").arg(markerName);
        }
    }

    emit message(tr("Anfrage zum Umbenennen von „%1“ zu „%2“%3 von %4").arg(
                    name, newName, markerChangeStart, client));

    if (! m_db->renamePlayers(name, newName, markerId)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ zu „%3“ umbenannt%4").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                         name, newName, markerChangeFinished));

    m_registrationPage->processNetworkChange();
    m_registrationPage->checkUpdateDrawOverviewPage(false);
    emit nameEdited();
    broadcastRenamePlayers(name, newName, markerId, markerName, clientId);
}

void Server::processAssemblePair(const QJsonObject &jsonObject, const QString &client,
                                 int clientId)
{
    const QJsonArray namesData = jsonObject.value(KEY_PLAYERS_NAME).toArray();
    const QString player1 = namesData.at(0).toString();
    const QString player2 = namesData.at(1).toString();
    const QString separator = namesData.at(2).toString();
    const int marker = jsonObject.value(KEY_MARKER_ID).toInt();
    emit message(tr("Anfrage zum Zusammenfügen von „%1“ und „%2“ zu einem Paar von %3").arg(
                    player1, player2, client));

    if (! m_db->assemblePair(player1, player2, separator, marker)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: „%1“ und „%2“ zu einem Paar zusammengefügt").arg(
                         player1, player2));

    m_registrationPage->processNetworkChange();
    m_registrationPage->checkUpdateDrawOverviewPage(false);
    broadcastAssemblePair(player1, player2, separator, marker, clientId);
}

void Server::processDeletePlayers(const QJsonObject &jsonObject, const QString &client,
                                  int clientId)
{
    const QString name = jsonObject.value(KEY_PLAYERS_NAME).toString();
    emit message(tr("Anfrage zum Löschen von „%1“ von %2").arg(name, client));

    if (! m_db->deletePlayers(name)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ gelöscht").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name));

    m_registrationPage->processNetworkChange();
    m_registrationPage->checkUpdateDrawOverviewPage(false);
    broadcastDeletePlayers(name, clientId);
}

void Server::processUpdatePlayersMarker(const QJsonObject &jsonObject, const QString &client,
                                        int clientId)
{
    const QString name = jsonObject.value(KEY_PLAYERS_NAME).toString();
    const QString markerName = jsonObject.value(KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(KEY_MARKER_ID).toInt();
    if (markerId != 0) {
        emit message(tr("Anfrage zum Markieren von „%1“ als „%2“ von %3").arg(
                        name, markerName, client));
    } else {
        emit message(tr("Anfrage zum Entfernen der Markierung von „%1“ von %3").arg(
                        name, client));
    }

    if (! m_db->updateMarker(name, markerId)) {
        return;
    }

    if (markerId != 0) {
        emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ als „%3“ markiert").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                             name, markerName));
    } else {
        emit statusUpdate(tr("Netzwerkänderung: Markierung von %1 „%2“ entfernt").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                             name));
    }

    m_registrationPage->processNetworkChange();
    broadcastUpdateMarker(name, markerName, markerId, clientId);
}

void Server::processDeleteMarkedPlayers(const QJsonObject &jsonObject, const QString &client,
                                        int clientId)
{
    const QString markerName = jsonObject.value(KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(KEY_MARKER_ID).toInt();
    emit message(tr("Anfrage zum Löschen aller als „%1“ markierten %2 von %3").arg(
                    markerName, m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                    client));

    if (! m_db->deleteMarkedPlayers(markerId)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Alle als %1 markierten „%2“ gelöscht").arg(
                         markerName, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                        : tr("Spieler")));

    m_registrationPage->processNetworkChange();
    m_registrationPage->checkUpdateDrawOverviewPage(false);
    broadcastDeleteMarkedPlayers(markerName, markerId, clientId);
}

void Server::processDeleteUnmarkedPlayers(const QJsonObject &, const QString &client,
                                          int clientId)
{
    emit message(tr("Anfrage zum Löschen aller unmarkierten %1 von %2").arg(
                    m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"), client));

    if (! m_db->deleteUnmarkedPlayers()) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Alle unmarkierten %1 gelöscht").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    m_registrationPage->processNetworkChange();
    m_registrationPage->checkUpdateDrawOverviewPage(false);
    broadcastDeleteUnmarkedPlayers(clientId);
}

void Server::processDeleteAllPlayers(const QJsonObject &, const QString &client, int clientId)
{
    emit message(tr("Anfrage zum Löschen aller %1 von %2").arg(
                    m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"), client));

    if (! m_db->deleteAllPlayers()) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Alle %1 gelöscht").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    m_registrationPage->processNetworkChange();
    m_registrationPage->checkUpdateDrawOverviewPage(false);
    broadcastDeleteAllPlayers(clientId);
}

void Server::processUpdateListMarker(const QJsonObject &jsonObject, const QString &client,
                                     int clientId)
{
    const QString markerName = jsonObject.value(KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(KEY_MARKER_ID).toInt();
    if (markerId != 0) {
        emit message(tr("Anfrage zum Markieren aller %1 als „%2“ von %3").arg(
                        m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                        markerName, client));
    } else {
        emit message(tr("Anfrage zum Entfernen aller Markierungen von %1").arg(client));
    }

    if (! m_db->listUpdateMarker(markerId)) {
        return;
    }

    if (markerId != 0) {
        emit statusUpdate(tr("Netzwerkänderung: Alle %1 als „%2“ markiert").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                             markerName));
    } else {
        emit statusUpdate(tr("Netzwerkänderung: Alle Markierungen entfernt"));
    }

    m_registrationPage->processNetworkChange();
    broadcastListUpdateMarker(markerName, markerId, clientId);
}

void Server::processRemoveListMarker(const QJsonObject &jsonObject, const QString &client,
                                     int clientId)
{
    const QString markerName = jsonObject.value(KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(KEY_MARKER_ID).toInt();
    emit message(tr("Anfrage zum Entfernen aller „%1“-Markierungen von %2").arg(
                    markerName, client));

    if (! m_db->listRemoveMarker(markerId)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Alle „%1“-Markierungen entfernt").arg(markerName));

    m_registrationPage->processNetworkChange();
    broadcastListRemoveMarker(markerName, markerId, clientId);
}

// Draw changes
// ============

void Server::processUpdateDraw(const QJsonObject &jsonObject, const QString &client, int clientId)
{
    const QString name = jsonObject.value(KEY_PLAYERS_NAME).toString();
    const Draw::Seat seat = { jsonObject.value(KEY_PLAYERS_TABLE).toInt(),
                              jsonObject.value(KEY_PLAYERS_PAIR_NUMBER).toInt() };
    const int marker = jsonObject.value(KEY_MARKER_ID).toInt();
    if (seat.isEmpty()) {
        emit message(tr("Anfrage zum Löschen der Auslosung für „%1“ von %2").arg(name, client));
    } else {
        emit message(tr("Anfrage zum Speichern der Auslosung für „%1“ von %2").arg(name, client));
    }

    if (! m_db->updateDraw(name, seat, marker)) {
        return;
    }

    if (seat.isEmpty()) {
        emit statusUpdate(tr("Netzwerkänderung: Auslosung für „%1“ gelöscht").arg(name));
    } else {
        emit statusUpdate(tr("Netzwerkänderung: Auslosung für „%1“ gespeichert (Tisch %2 Paar "
                             "%3)").arg(name, QString::number(seat.table),
                                        QString::number(seat.pair)));
    }

    m_registrationPage->processNetworkChange();
    m_registrationPage->checkUpdateDrawOverviewPage(false);
    broadcastUpdateDraw(name, seat, marker, clientId);
}

void Server::processSetListDraw(const QJsonObject &jsonObject, const QString &client, int clientId)
{
    emit message(tr("Anfrage zum Setzen der Auslosung aller Plätze von %1").arg(client));

    QVector<QString> names;
    QVector<Draw::Seat> draw;
    const QJsonArray namesData = jsonObject.value(KEY_PLAYERS_LIST).toArray();
    const QJsonArray tablesData = jsonObject.value(KEY_TABLES_LIST).toArray();
    const QJsonArray pairsData = jsonObject.value(KEY_PAIR_NUMBERS_LIST).toArray();
    for (int i = 0; i < namesData.count(); i++) {
        names.append(namesData.at(i).toString());
        draw.append(Draw::Seat { tablesData.at(i).toInt(), pairsData.at(i).toInt() });
    }

    if (! m_db->setListDraw(names, draw)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Auslosung aller Plätze gesetzt"));

    m_registrationPage->processNetworkChange();
    m_registrationPage->checkUpdateDrawOverviewPage(false);
    broadcastSetListDraw(names, draw, clientId);
}

void Server::processDeleteAllDraws(const QJsonObject &, const QString &client, int clientId)
{
    emit message(tr("Anfrage zum Löschen aller Auslosungen von %1").arg(client));

    if (! m_db->deleteAllDraws()) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Alle Auslosungen gelöscht"));

    m_registrationPage->processNetworkChange();
    m_registrationPage->checkUpdateDrawOverviewPage(false);
    broadcastDeleteAllDraws(clientId);
}

// Markers defintion changes
// =========================

void Server::processAddMarker(const QJsonObject &jsonObject, const QString &client, int clientId)
{
    const QString name = jsonObject.value(KEY_MARKER_NAME).toString();
    const QColor color = Connection::colorFromValue(jsonObject.value(KEY_MARKER_COLOR));
    const Markers::MarkerSorting sorting
        = static_cast<Markers::MarkerSorting>(jsonObject.value(KEY_MARKER_SORTING).toInt());
    emit message(tr("Anfrage zum Hinzufügen der Markierung „%1“ von %2").arg(name, client));

    if (! m_db->addMarker(name, color, sorting)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ hinzugefügt").arg(name));

    m_markersWidget->processNetworkChange();
    broadcastAddMarker(name, color, sorting, clientId);
}

void Server::processDeleteMarker(const QJsonObject &jsonObject, const QString &client,
                                 int clientId)
{
    const QString name = jsonObject.value(KEY_MARKER_NAME).toString();
    const int id = jsonObject.value(KEY_MARKER_ID).toInt();
    const int sequence = jsonObject.value(KEY_MARKER_SEQUENCE).toInt();
    emit message(tr("Anfrage zum Löschen der Markierung „%1“ von %2").arg(name, client));

    if (! m_db->deleteMarker(id, sequence)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ gelöscht").arg(name));

    m_markersWidget->processNetworkChange();
    broadcastDeleteMarker(name, id, sequence, clientId);
}

void Server::processMoveMaker(const QJsonObject &jsonObject, const QString &client, int clientId)
{
    const QString name = jsonObject.value(KEY_MARKER_NAME).toString();
    const int id = jsonObject.value(KEY_MARKER_ID).toInt();
    const int sequence = jsonObject.value(KEY_MARKER_SEQUENCE).toInt();
    const int direction = jsonObject.value(KEY_MARKER_DIRECTION).toInt();
    emit message(tr("Anfrage zum Verschieben der Markierung „%1“ %2 von %3").arg(
                    name, direction == 1 ? tr("nach unten") : tr("nach oben"), client));

    if (! m_db->moveMarker(id, sequence, direction)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ %2 verschoben").arg(
                         name, direction == 1 ? tr("nach unten") : tr("nach oben")));

    m_markersWidget->processNetworkChange();
    broadcastMoveMarker(name, id, sequence, direction, clientId);
}

void Server::processEditMarker(const QJsonObject &jsonObject, const QString &client, int clientId)
{
    const QString name = jsonObject.value(KEY_MARKER_NAME).toString();
    const int id = jsonObject.value(KEY_MARKER_ID).toInt();
    const QString newName = jsonObject.value(KEY_MARKER_NEW_NAME).toString();
    const QColor color = Connection::colorFromValue(jsonObject.value(KEY_MARKER_COLOR));
    const Markers::MarkerSorting sorting = static_cast<Markers::MarkerSorting>(
                                               jsonObject.value(KEY_MARKER_SORTING).toInt());
    emit message(tr("Anfrage zum Ändern der Markierung „%1“ von %2").arg(name, client));

    if (! m_db->editMarker(id, newName, color, sorting)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ geändert").arg(name));

    m_markersWidget->processNetworkChange();
    broadcastEditMarker(name, id, newName, color, sorting, clientId);
}

// Special markers changes
// =======================

void Server::processSetAbsentMarker(const QJsonObject &jsonObject, const QString &client,
                                    int clientId)
{
    const QString name = jsonObject.value(KEY_MARKER_NAME).toString();
    const int id = jsonObject.value(KEY_MARKER_ID).toInt();
    emit message(id == -1 ? tr("Anfrage zum Deaktivieren der Markierung für abwesende Anmeldungen "
                               "von %1").arg(client)
                          : tr("Anfrage zum Setzen der Markierung für abwesende Anmeldungen auf "
                               "„%1“ von %2").arg(name, client));

    m_markersWidget->setCheckableMarker(Markers::Absent, id);
    m_tournamentSettings->setMarker(Markers::Absent, id);
    emit statusUpdate(id == -1
        ? tr("Netzwerkänderung: Markierung für abwesende Anmeldungen deaktiviert")
        : tr("Netzwerkänderung: Markierung für abwesende Anmeldungen auf „%1“ gesetzt").arg(name));

    m_markersWidget->processNetworkChange();
    broadcastSetAbsentMarker(name, id, clientId);
}

void Server::processChangeSinglesManagement(const QJsonObject &jsonObject, const QString &client,
                                            int clientId)
{
    const Markers::SinglesManagement singlesManagement = JsonHelper::parseSinglesManagement(
        jsonObject.value(KEY_SINGLES_MANAGEMENT).toObject());
    emit message(tr("Anfrage zum Ändern der Einstellungen zum Behandeln allein gekommner Spieler "
                    "von %1").arg(client));

    m_markersWidget->setSinglesManagement(singlesManagement);
    m_tournamentSettings->setSinglesManagement(singlesManagement);
    emit statusUpdate(tr("Netzwerkänderung: Änderungen der Einstellungen zum Behandeln allein "
                         "gekommener Spieler übernommen"));

    m_markersWidget->processNetworkChange();
    broadcastSinglesManagementChange(singlesManagement, clientId);
}

// Finishing the registration
// ==========================

void Server::processFinishRegistration(const QJsonObject &, const QString &client, int clientId)
{
    emit message(tr("Anfrage zum Beenden der Anmeldung von %1").arg(client));

    m_registrationPage->networkFinishRegistration(clientId);
    emit statusUpdate(tr("Netzwerkänderung: Anmeldung beendet"));

    broadcastFinishRegistration(clientId);
}

// Ranking comparison
// ==================

void Server::processRankingRequest(Connection *connection, const QString &client, int clientId)
{
    emit message(tr("Rangliste von %1 angefragt").arg(client));

    if (! m_db->localTournamentStarted()) {
        connection->sendData(QJsonObject { { KEY_ACTION,            ACTION_RANKING_REQUEST },
                                           { KEY_CLIENT_ID,         clientId },
                                           { KEY_RANKING_AVAILABLE, false } });
        emit message(tr("Es ist noch keine Rangliste verfügbar"));
        return;
    }

    QJsonArray names;
    QJsonArray boogers;
    QJsonArray goals;
    QJsonArray opponentGoals;
    QJsonArray playedBoogers;

    const auto [ ranking, success ] = m_db->ranking(-1);
    if (! success) {
        return;
    }

    for (const Database::RankData &rank : ranking) {
        names.append(QJsonValue(rank.name));
        boogers.append(QJsonValue(rank.boogers));
        goals.append(QJsonValue(rank.goals));
        opponentGoals.append(QJsonValue(rank.opponentGoals));
        playedBoogers.append(QJsonValue(rank.playedBoogers));
    }

    QJsonObject data { { KEY_ACTION,              ACTION_RANKING_REQUEST },
                       { KEY_CLIENT_ID,           clientId },
                       { KEY_RANKING_AVAILABLE,   true },
                       { KEY_PLAYERS_LIST,        names },
                       { KEY_BOOGERS_LIST,        boogers },
                       { KEY_GOALS_LIST,          goals },
                       { KEY_OPPONENT_GOALS_LIST, opponentGoals },
                       { KEY_PLAYED_BOOGERS_LIST, playedBoogers } };

    sendSingleReply(data, connection,
                    tr("Rangliste an %1 gesendet").arg(client));
}

// Score comparison
// ================

void Server::processScoreRequest(int round, Connection *connection, const QString &client,
                                 int clientId)
{
    emit message(tr("Ergebnisse für Runde %1 von %2 angefragt").arg(
                    QString::number(round), client));

    const auto [ score, success ] = m_db->scores(round);
    if (! success) {
        return;
    }

    QJsonArray tables;
    for (const auto table : score.tables) {
        tables.append(table);
    }

    QJsonArray pairs;
    for (const auto &pair : score.pairs) {
        pairs.append(pair);
    }

    QJsonArray scores;
    for (const auto &pairResults : score.scores) {
        QJsonArray singleScore;
        for (const auto singleResult : pairResults) {
            singleScore.append(singleResult);
        }
        scores.append(singleScore);
    }

    QJsonObject data { { KEY_ACTION,       ACTION_SCORE_REQUEST },
                       { KEY_CLIENT_ID,    clientId },
                       { KEY_TABLES_LIST,  tables },
                       { KEY_PLAYERS_LIST, pairs },
                       { KEY_SCORES_LIST,  scores } };

    sendSingleReply(data, connection,
                    tr("Ergebnisse für Runde %1 an %2 gesendet").arg(
                    QString::number(round), client));
}
