/* Copyright (c) 2019-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes

#include "MarkersTemplatePage.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/Settings.h"
#include "shared/TournamentSettings.h"
#include "shared/TableDelegate.h"
#include "shared/SharedStyles.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QLabel>
#include <QTableWidget>
#include <QPushButton>
#include <QHeaderView>
#include <QMessageBox>

// C++ includes
#include <utility>
#include <algorithm>

MarkersTemplatePage::MarkersTemplatePage(SharedObjects *sharedObjects, QWidget *parent)
    : AbstractSettingsPage(tr("Vorlage Markierungen"), parent),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings())
{
    m_markersTemplate = m_settings->markersTemplate();

    layout()->addWidget(new QLabel(tr("Automatisch angelegte Markierungen:")));

    m_markersTemplateDisplay = new QTableWidget;
    m_markersTemplateDisplay->verticalHeader()->hide();
    m_markersTemplateDisplay->setShowGrid(false);
    m_markersTemplateDisplay->setSelectionMode(QTableWidget::NoSelection);
    m_markersTemplateDisplay->setFocusPolicy(Qt::NoFocus);
    m_markersTemplateDisplay->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_markersTemplateDisplay->setItemDelegate(new TableDelegate(m_markersTemplateDisplay));
    m_markersTemplateDisplay->setColumnCount(2);
    m_markersTemplateDisplay->setRowCount(m_markersTemplate.count());
    m_markersTemplateDisplay->setHorizontalHeaderLabels({ tr("Name/Farbe"), tr("Sortierung") });

    // First, we set the approximate wanted maximum height
    m_markersTemplateDisplay->setMaximumHeight(
        m_markersTemplateDisplay->horizontalHeader()->height()
        + m_markersTemplateDisplay->rowHeight(0) * std::min(5, m_markersTemplate.count()));

    // Then we add it to the layout and show it so that the geometry is correct
    layout()->addWidget(m_markersTemplateDisplay);
    m_markersTemplateDisplay->show();

    // Then, we add additional pixels needed by the style
    // (height - header height - viewport height)
    m_markersTemplateDisplay->setMaximumHeight(
        2 * m_markersTemplateDisplay->height()
        - m_markersTemplateDisplay->horizontalHeader()->height()
        - m_markersTemplateDisplay->viewport()->height());

    layout()->addWidget(new QLabel(tr("Vorauswahl Markierung für:")));

    m_settingsDisplay = new QTableWidget;
    m_settingsDisplay->horizontalHeader()->hide();
    m_settingsDisplay->verticalHeader()->hide();
    m_settingsDisplay->setShowGrid(false);
    m_settingsDisplay->setSelectionMode(QTableWidget::NoSelection);
    m_settingsDisplay->setFocusPolicy(Qt::NoFocus);
    m_settingsDisplay->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_settingsDisplay->setItemDelegate(new TableDelegate(m_settingsDisplay));
    m_settingsDisplay->setColumnCount(2);
    m_settingsDisplay->setRowCount(SettingsRow::Count);

    m_settingsDisplay->setMaximumHeight(m_settingsDisplay->rowHeight(0) * SettingsRow::Count);
    layout()->addWidget(m_settingsDisplay);
    m_settingsDisplay->show();
    m_settingsDisplay->setMaximumHeight(2 * m_settingsDisplay->height()
                                        - m_settingsDisplay->horizontalHeader()->height()
                                        - m_settingsDisplay->viewport()->height());


    m_settingsDisplay->setItem(SettingsRow::Default, 0,
                               new QTableWidgetItem(tr("Neue Anmeldungen:")));

    m_settingsDisplay->setItem(SettingsRow::Drawn, 0,
                               new QTableWidgetItem(tr("Ausgeloste Paare/Spieler:")));

    m_settingsDisplay->setItem(SettingsRow::Absent, 0,
                               new QTableWidgetItem(tr("Abwesende Anmeldungen:")));

    m_settingsDisplay->resizeColumnsToContents();

    m_noProcessSingles = new QLabel(tr("Allein gekommene Spieler im Paarmodus werden nicht "
                                       "gesondert berücksichtigt"));
    m_noProcessSingles->setWordWrap(true);
    m_noProcessSingles->hide();
    layout()->addWidget(m_noProcessSingles);

    m_processSingles = new QGroupBox(tr("Allein gekommene Spieler"));
    m_processSingles->setStyleSheet(SharedStyles::GROUPBOX_NORMAL_TITLE);
    auto *processSingleslayout = new QVBoxLayout(m_processSingles);
    m_singlesMarker = new QLabel;
    processSingleslayout->addWidget(m_singlesMarker);
    m_automaticMarking = new QLabel;
    processSingleslayout->addWidget(m_automaticMarking);
    m_assignedMarker = new QLabel;
    processSingleslayout->addWidget(m_assignedMarker);
    layout()->addWidget(m_processSingles);

    layout()->addStretch();

    auto *controlLayout = new QHBoxLayout;
    layout()->addLayout(controlLayout);
    controlLayout->addStretch();

    auto *adoptCurrentSettingsButton = new QPushButton(tr("Aktuelle Einstellungen übernehmen"));
    connect(adoptCurrentSettingsButton, &QPushButton::clicked,
            this, &MarkersTemplatePage::adoptCurrentSettings);
    controlLayout->addWidget(adoptCurrentSettingsButton);
    adoptCurrentSettingsButton->setEnabled(! m_db->dbFile().isEmpty());

    auto *restoreTemplateButton = new QPushButton(tr("Zurücksetzen"));
    connect(restoreTemplateButton, &QPushButton::clicked,
            this, &MarkersTemplatePage::restoreTemplate);
    controlLayout->addWidget(restoreTemplateButton);

    m_singlesManagementTemplate = m_settings->singlesManagementTemplate();
    m_defaultMarker = m_settings->marker(Markers::Default);
    m_drawnMarker = m_settings->marker(Markers::Drawn);
    m_absentMarker = m_settings->marker(Markers::Absent);

    updateTemplateDisplay();
}

const Markers::Marker &MarkersTemplatePage::findMarker(int id) const
{
    for (const auto &marker : std::as_const(m_markersTemplate)) {
        if (marker.id == id) {
            return marker;
        }
    }

    return findMarker(0);
}

QString MarkersTemplatePage::markerName(int id) const
{
    const auto &marker = findMarker(id);
    return marker.name.isEmpty() ? tr("(unmarkiert)")
                                 : tr("<span style=\"color: %1\">%2</span>").arg(
                                      marker.color.name(), marker.name);
}

QTableWidgetItem *MarkersTemplatePage::markerItem(int id) const
{
    auto *item = new QTableWidgetItem;

    if (id == -1) {
        item->setText(tr("nicht gesetzt"));
        auto font = item->font();
        font.setItalic(true);
        item->setFont(font);
        return item;
    }

    const auto &marker = findMarker(id);
    item->setText(marker.name.isEmpty() ? tr("(unmarkiert)") : marker.name);
    item->setForeground(marker.color);
    return item;
}

void MarkersTemplatePage::updateTemplateDisplay()
{
    m_markersTemplateDisplay->setRowCount(m_markersTemplate.count());
    for (int i = 0; i < m_markersTemplate.count(); i++) {
        QTableWidgetItem *item = new QTableWidgetItem;
        m_markersTemplateDisplay->setItem(i, 0, item);
        item->setText(m_markersTemplate.at(i).name.isEmpty() ? tr("(unmarkiert)")
                                                             : m_markersTemplate.at(i).name);
        item->setForeground(m_markersTemplate.at(i).color);
        m_markersTemplateDisplay->setItem(i, 1, new QTableWidgetItem(
            m_markersTemplate.at(i).sorting == Markers::MarkerSorting::NewBlock
                                                   ? tr("Neuer Block") : tr("Fortlaufend")));
    }
    m_markersTemplateDisplay->resizeColumnsToContents();

    m_settingsDisplay->setItem(SettingsRow::Default, 1, markerItem(m_defaultMarker));
    m_settingsDisplay->setItem(SettingsRow::Drawn, 1, markerItem(m_drawnMarker));
    m_settingsDisplay->setItem(SettingsRow::Absent, 1, markerItem(m_absentMarker));
    m_settingsDisplay->resizeColumnsToContents();

    if (m_singlesManagementTemplate.enabled) {
        m_noProcessSingles->hide();
        m_processSingles->show();
        m_singlesMarker->setText(tr("Markierung für „Allein da“: %1").arg(
                                    markerName(m_singlesManagementTemplate.singleMarker)));
        if (m_singlesManagementTemplate.markSingles) {
            m_automaticMarking->setText(
                tr("Automatisch setzen, wenn der Name „%1“ %2").arg(
                    m_singlesManagementTemplate.conditionString,
                    m_singlesManagementTemplate.condition == 0 ? tr("nicht enthält")
                                                               : tr("enthält")));
            m_automaticMarking->show();
        } else {
            m_automaticMarking->hide();
        }
        m_assignedMarker->setText(tr("Markierung nach Zuweisung: %1").arg(
                                     markerName(m_singlesManagementTemplate.assignedMarker)));
    } else {
        m_noProcessSingles->show();
        m_processSingles->hide();
    }
}

void MarkersTemplatePage::restoreTemplate()
{
    m_defaultMarker = 0;
    m_drawnMarker = -1;
    m_absentMarker = -1;
    m_markersTemplate = m_settings->defaultMarkersTemplate();
    m_singlesManagementTemplate = m_settings->defaultSinglesManagementTemplate();
    updateTemplateDisplay();
}

void MarkersTemplatePage::adoptCurrentSettings()
{
    const auto [ markers, success ] = m_db->markers();
    if (! success) {
        return;
    }
    m_markersTemplate = markers;

    m_singlesManagementTemplate = m_tournamentSettings->singlesManagement();
    m_defaultMarker = m_tournamentSettings->marker(Markers::Default);
    m_drawnMarker = m_tournamentSettings->marker(Markers::Drawn);
    m_absentMarker = m_tournamentSettings->marker(Markers::Absent);
    updateTemplateDisplay();
}

void MarkersTemplatePage::saveSettings()
{
    m_settings->saveMarker(Markers::Default, m_defaultMarker);
    m_settings->saveMarker(Markers::Drawn, m_drawnMarker);
    m_settings->saveMarker(Markers::Absent, m_absentMarker);
    m_settings->saveMarkersTemplate(m_markersTemplate);
    m_settings->saveSinglesManagementTemplate(m_singlesManagementTemplate);
}
