/* Copyright (c) 2010-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes

#include "MainWindow.h"
#ifdef Q_OS_MACOS
#include "MacOSApplication.h"
#endif

#include "shared/SharedObjects.h"
#include "shared/ResourceFinder.h"

// Qt includes
#include <QApplication>
#include <QDebug>
#include <QTranslator>
#include <QLibraryInfo>
#include <QFileInfo>
#include <QStyle>
#ifndef Q_OS_MACOS
#include <QStringList>
#endif

int main(int argc, char *argv[])
{
    // Create the core application
#ifndef Q_OS_MACOS
    // On Linux and Windows, we can use a default QApplication:
    QApplication application(argc, argv);
#else
    // On macOS, "Open with" does not simply pass the file as a command line argument, but triggers
    // a QFileOpenEvent that we have to process. Thus, we need a custom QApplication on Mac:
    MacOSApplication application(argc, argv);
#endif

    // Set application-wide styles
    if (application.style()->objectName() != QLatin1String("fusion")) {
        application.setStyleSheet(QStringLiteral(
            "QGroupBox[checkable=false] { "
                "font-weight: bold; "
            "}"));
    } else {
        application.setStyleSheet(QStringLiteral(
            // This is a workaround for Bug #44056 (https://bugreports.qt.io/browse/QTBUG-44056)
            // which hopefully can be removed once the bug is actually fixed, along with all
            // setProperty("notitle", true) calls on QGroupBoxes.
            "QGroupBox { "
                "border: 1px solid palette(Mid); "
                "border-radius: 2px; "
                "padding-top: 1.2em; "
            "} "
            "QGroupBox[notitle=true] { "
                "padding-top: 0; "
            "} "
            "QGroupBox[checkable=false] { "
                "font-weight: bold; "
            "} "
            "QGroupBox::title { "
                "background-color: transparent; "
                "subcontrol-origin: margin; "
                "subcontrol-position: top center; "
                "padding: 0.4em 0 0; "
            "} "
            // This prevents HLine and VLine QFrames to be a black line
            "QFrame[frameShape=\"4\"],"
            "QFrame[frameShape=\"5\"] { "
                "color: palette(Mid); "
            "}"));
    }

    // Create the application-wide shared objects
    SharedObjects sharedObjects;
    ResourceFinder *resourceFinder = sharedObjects.resourceFinder();

    // Install the translations

    // Nobody outside of Germany will use this, and nobody will ever translate
    // it to any other language. So let's just load the de locale stuff :-P

    // Add the Qt localizations
    QTranslator qtTranslator;
    // Look for a system-wide Qt locale file
    if (! qtTranslator.load(QStringLiteral("qt_de"),
                            QLibraryInfo::location(QLibraryInfo::TranslationsPath))) {

        // Look for a resource file
        const QString path = resourceFinder->find(QStringLiteral("qt_de.qm"));
        if (! path.isEmpty()) {
            qtTranslator.load(QStringLiteral("qt_de"), QFileInfo(path).absolutePath());
        }
    }
    if (! qtTranslator.isEmpty()) {
        application.installTranslator(&qtTranslator);
    }

    // Add the Muckturnier localizations (plural forms)
    QTranslator muckturnierTranslator;
    const QString path = resourceFinder->find(QStringLiteral("muckturnier_de.qm"));
    if (! path.isEmpty()) {
        muckturnierTranslator.load(QStringLiteral("muckturnier_de"),
                                   QFileInfo(path).absolutePath());
    }
    if (! muckturnierTranslator.isEmpty()) {
        application.installTranslator(&muckturnierTranslator);
    }

    // Create the main window
    MainWindow mainWindow(&sharedObjects);
    mainWindow.show();

    // Handle "Open with ..." requests
#ifndef Q_OS_MACOS
    // On Linux and Windows, "Open with" simply passes the file to be opened as a command line
    // argument that we handle here:
    const QStringList arguments(application.arguments());
    if (arguments.count() > 1) {
        mainWindow.setEnabled(false);
        QApplication::processEvents();
        mainWindow.openTournament(arguments.at(1));
        mainWindow.setEnabled(true);
    }
#else
    // On macOS, we listen for a QFileOpenEvent in the custom QApplication, catch the file name to
    // be opened and pass it to the main window via the fileOpenRequested signal:
    QObject::connect(&application, &MacOSApplication::fileOpenRequested,
                     &mainWindow, &MainWindow::openTournament);
#endif

    return application.exec();
}
