/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SERVER_PROTOCOL_H
#define SERVER_PROTOCOL_H

#include <QJsonObject>
#include <QDataStream>

namespace ServerProtocol
{

// Basic protocol data
constexpr int PROTOCOL_VERSION = 3;
const QDataStream::Version STREAM_VERSION = QDataStream::Qt_5_11;
const QLatin1String SERVER_SIGNATURE("muckturnier server");

// Keys used in all requests and answers
const QLatin1String KEY_ACTION   ("action");
const QLatin1String KEY_CLIENT_ID("client id");
const QLatin1String KEY_CHECKSUM ("checksum");

// Invalid request handling
const QLatin1String ACTION_REJECTED  ("request rejected");
const QLatin1String KEY_REJECT_REASON("reject reason");
enum RejectReason {
    Unknown,
    OutdatedRequest,
    UnknownAction
};

// Data types
// ==========

// Markers
// -------

const QLatin1String KEY_MARKER_NAME     ("marker name");
const QLatin1String KEY_MARKER_ID       ("marker id");
const QLatin1String KEY_MARKER_COLOR    ("marker color");
const QLatin1String KEY_MARKER_SORTING  ("marker sorting");
const QLatin1String KEY_MARKER_NEW_NAME ("new marker name");
const QLatin1String KEY_MARKER_SEQUENCE ("marker sequence");
const QLatin1String KEY_MARKER_DIRECTION("marker direction");
const QLatin1String KEY_MARKER_ABSENT   ("absent marker");

// Players
// -------

const QLatin1String KEY_PLAYERS_NAME       ("players name");
const QLatin1String KEY_PLAYERS_NEW_NAME   ("new players name");
const QLatin1String KEY_PLAYERS_TABLE      ("players table number");
const QLatin1String KEY_PLAYERS_PAIR_NUMBER("players pair number");

// Data lists
// ----------

const QLatin1String KEY_PLAYERS_LIST        ("players list");
const QLatin1String KEY_MARKERS_LIST        ("markers list");
const QLatin1String KEY_TABLES_LIST         ("tables list");
const QLatin1String KEY_PAIR_NUMBERS_LIST   ("pair numbers list");
const QLatin1String KEY_BOOGERS_LIST        ("boogers list");
const QLatin1String KEY_GOALS_LIST          ("goals list");
const QLatin1String KEY_OPPONENT_GOALS_LIST ("opponent goals list");
const QLatin1String KEY_PLAYED_BOOGERS_LIST ("played boogers list");
const QLatin1String KEY_SCORES_LIST         ("scores list");
const QLatin1String KEY_SCORE_CHECKSUMS_LIST("score checksums");

// Other (request specific) keys
// -----------------------------

const QLatin1String KEY_SINGLES_MANAGEMENT("singles management");
const QLatin1String KEY_RANKING_AVAILABLE ("ranking available");
const QLatin1String KEY_ROUND             ("round");
const QLatin1String KEY_SCORE_CHECKSUM    ("score checksum");

// Handshake
// =========

const QLatin1String ACTION_HANDSHAKE("handshake");

const QLatin1String HANDSHAKE_KEY_PHASE   ("phase");
const QLatin1String HANDSHAKE_PHASE_FAILED("failed");

// Welcome message
const QLatin1String HANDSHAKE_PHASE_INITIALIZATION("initialization");
const QLatin1String HANDSHAKE_KEY_SERVER_SIGNATURE("server signature");
const QLatin1String HANDSHAKE_KEY_PROTOCOL_VERSION("protocol version");
const QJsonObject HANDSHAKE_MESSAGE_INITIALIZE {
    { KEY_ACTION,                     ACTION_HANDSHAKE },
    { HANDSHAKE_KEY_PHASE,            HANDSHAKE_PHASE_INITIALIZATION },
    { HANDSHAKE_KEY_SERVER_SIGNATURE, SERVER_SIGNATURE },
    { HANDSHAKE_KEY_PROTOCOL_VERSION, PROTOCOL_VERSION }
};

// Synchronization request
const QLatin1String HANDSHAKE_PHASE_SYNCHRONIZE("synchronization");
const QJsonObject HANDSHAKE_MESSAGE_SYNCHRONIZE {
    { KEY_ACTION,          ACTION_HANDSHAKE },
    { HANDSHAKE_KEY_PHASE, HANDSHAKE_PHASE_SYNCHRONIZE }
};

// Synchronization answer

const QLatin1String HANDSHAKE_KEY_SETTINGS          ("settings");
const QLatin1String HANDSHAKE_KEY_LOCALE            ("locale");
const QLatin1String HANDSHAKE_KEY_NAMES_SEPARATOR   ("names separator");
const QLatin1String HANDSHAKE_KEY_TOURNAMENT_MODE   ("tournament mode");
const QLatin1String HANDSHAKE_KEY_BOOGER_SCORE      ("booger score");
const QLatin1String HANDSHAKE_KEY_BOOGERS_PER_ROUND ("boogers per round");
const QLatin1String HANDSHAKE_KEY_TOURNAMENT_STARTED("tournament started");

// Synchronization success, request/allocate an id
const QLatin1String HANDSHAKE_PHASE_ALLOCATE_ID("id allocation");
const QJsonObject HANDSHAKE_MESSAGE_ALLOCATE_ID {
    { KEY_ACTION,          ACTION_HANDSHAKE },
    { HANDSHAKE_KEY_PHASE, HANDSHAKE_PHASE_ALLOCATE_ID }
};

// Actions
// =======

// Registration changes

const QLatin1String ACTION_PLAYERS_REGISTER     ("register players");
const QLatin1String ACTION_PLAYERS_RENAME       ("rename players");
const QLatin1String ACTION_PLAYERS_ASSEMBLE_PAIR("assemble pair");
const QLatin1String ACTION_PLAYERS_DELETE       ("delete players");
const QLatin1String ACTION_PLAYERS_UPDATE_MARKER("update players marker");

const QLatin1String ACTION_PLAYERS_DELETE_MARKED  ("delete marked players");
const QLatin1String ACTION_PLAYERS_DELETE_UNMARKED("delete unmarked players");
const QLatin1String ACTION_PLAYERS_DELETE_ALL     ("delete all players");

const QLatin1String ACTION_PLAYERS_LIST_UPDATE_MARKER("update players list marker");
const QLatin1String ACTION_PLAYERS_LIST_REMOVE_MARKER("remove players list marker");

// Draw changes
const QLatin1String ACTION_UPDATE_DRAW     ("update draw");
const QLatin1String ACTION_SET_LIST_DRAW   ("set list draw");
const QLatin1String ACTION_DELETE_ALL_DRAWS("delete all draws");

// Markers changes
const QLatin1String ACTION_MARKERS_ADD   ("add marker");
const QLatin1String ACTION_MARKERS_EDIT  ("edit marker");
const QLatin1String ACTION_MARKERS_DELETE("delete marker");
const QLatin1String ACTION_MARKERS_MOVE  ("move marker");

// Special markers definitions
const QLatin1String ACTION_SET_ABSENT_MARKER        ("set absent marker");
const QLatin1String ACTION_CHANGE_SINGLES_MANAGEMENT("singles management change");

// Finish registration
const QLatin1String ACTION_FINISH_REGISTRATION("finish registration");

// Score change
const QLatin1String ACTION_SCORE_CHANGE("score change");

// Score request
const QLatin1String ACTION_SCORE_REQUEST("score request");

// Ranking request
const QLatin1String ACTION_RANKING_REQUEST("ranking request");

}

#endif // SERVER_PROTOCOL_H
