/* Copyright (c) 2019-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes
#include "JsonHelper.h"

// Qt includes
#include <QString>
#include <QJsonObject>
#include <QJsonArray>
#include <QJsonDocument>
#include <QDebug>

namespace JsonHelper
{

QString serializeJson(const QJsonObject &json)
{
    return QString::fromUtf8(QJsonDocument(json).toJson(QJsonDocument::Compact));
}

QString serializeJson(const QJsonArray &json)
{
    return QString::fromUtf8(QJsonDocument(json).toJson(QJsonDocument::Compact));
}

QJsonObject jsonObjectFromString(const QString &data)
{
    return QJsonDocument::fromJson(data.toUtf8()).object();
}

QJsonArray jsonArrayFromString(const QString &data)
{
    return QJsonDocument::fromJson(data.toUtf8()).array();
}

// ===================================================================================
// If anything is changed here, don't forget to bump the server protocol version!

const QString SINGLES_MANAGEMENT_ENABLED          = QStringLiteral("enabled");
const QString SINGLES_MANAGEMENT_MARKER_SINGLE    = QStringLiteral("single marker");
const QString SINGLES_MANAGEMENT_MARK_SINGLES     = QStringLiteral("mark singles");
const QString SINGLES_MANAGEMENT_CONDITION_STRING = QStringLiteral("condition string");
const QString SINGLES_MANAGEMENT_CONDITION        = QStringLiteral("condition");
const QString SINGLES_MANAGEMENT_MARKER_ASSIGNED  = QStringLiteral("assigned marker");

// ====================================================================================

QJsonObject singlesManagementToJson(const Markers::SinglesManagement &data)
{
    return QJsonObject {
        { SINGLES_MANAGEMENT_ENABLED,          data.enabled },
        { SINGLES_MANAGEMENT_MARKER_SINGLE,    data.singleMarker },
        { SINGLES_MANAGEMENT_MARK_SINGLES,     data.markSingles },
        { SINGLES_MANAGEMENT_CONDITION_STRING, data.conditionString },
        { SINGLES_MANAGEMENT_CONDITION,        data.condition },
        { SINGLES_MANAGEMENT_MARKER_ASSIGNED,  data.assignedMarker }
    };
}

QString singlesManagementToString(const Markers::SinglesManagement &data)
{
    return QString::fromUtf8(
        QJsonDocument(singlesManagementToJson(data)).toJson(QJsonDocument::Compact));
}

Markers::SinglesManagement parseSinglesManagement(const QString &data)
{
    return parseSinglesManagement(jsonObjectFromString(data));
}

Markers::SinglesManagement parseSinglesManagement(const QJsonObject &jsonObject)
{
    Markers::SinglesManagement data;
    QJsonValue value;

    value = jsonObject.value(SINGLES_MANAGEMENT_ENABLED);
    data.enabled = value.type() == QJsonValue::Bool
        ? value.toBool() : Markers::DEFAULT_SINGLES_MANAGEMENT.enabled;

    value = jsonObject.value(SINGLES_MANAGEMENT_MARKER_SINGLE);
    data.singleMarker = value.type() == QJsonValue::Double
        ? value.toInt() : Markers::DEFAULT_SINGLES_MANAGEMENT.singleMarker;

    value = jsonObject.value(SINGLES_MANAGEMENT_MARK_SINGLES);
    data.markSingles = value.type() == QJsonValue::Bool
        ? value.toBool() : Markers::DEFAULT_SINGLES_MANAGEMENT.markSingles;

    value = jsonObject.value(SINGLES_MANAGEMENT_CONDITION_STRING);
    data.conditionString = value.type() == QJsonValue::String
        ? value.toString() : Markers::DEFAULT_SINGLES_MANAGEMENT.conditionString;

    value = jsonObject.value(SINGLES_MANAGEMENT_CONDITION);
    data.condition = value.type() == QJsonValue::Double
        ? value.toInt() : Markers::DEFAULT_SINGLES_MANAGEMENT.condition;

    value = jsonObject.value(SINGLES_MANAGEMENT_MARKER_ASSIGNED);
    data.assignedMarker = value.type() == QJsonValue::Double
        ? value.toInt() : Markers::DEFAULT_SINGLES_MANAGEMENT.assignedMarker;

    return data;
}

QString registrationColumnsToString(const QVector<RegistrationColumns::Column> &columns)
{
    return serializeJson(registrationColumnsToJson(columns));
}

QJsonArray registrationColumnsToJson(const QVector<RegistrationColumns::Column> &columns)
{
    QJsonArray data;
    for (const RegistrationColumns::Column column : RegistrationColumns::ALL_COLUMNS) {
        if (columns.contains(column)) {
            data.append(RegistrationColumns::COLUMN_ID.value(column));
        }
    }
    return data;
}

QVector<RegistrationColumns::Column> parseRegistrationColumns(const QString &data)
{
    if (! data.isEmpty()) {
        return parseRegistrationColumns(jsonArrayFromString(data));
    } else {
        return RegistrationColumns::DEFAULT_COLUMNS;
    }
}

QVector<RegistrationColumns::Column> parseRegistrationColumns(const QJsonArray &jsonArray)
{
    QVector<RegistrationColumns::Column> columns;
    for (const RegistrationColumns::Column column : RegistrationColumns::ALL_COLUMNS) {
        if (jsonArray.contains(RegistrationColumns::COLUMN_ID.value(column))) {
            columns.append(column);
        }
    }
    return columns;
}

const QString DRAW_CONSECUTIVE      = QStringLiteral("consecutive");
const QString DRAW_TABLES_LIMIT     = QStringLiteral("tables limit");
const QString DRAW_RANDOM_DRAW      = QStringLiteral("random draw");
const QString DRAW_WINDOW_DRAW      = QStringLiteral("window draw");
const QString DRAW_CONSECUTIVE_DRAW = QStringLiteral("consecutive draw");

QString drawSettingsToString(const Draw::Settings &settings)
{
    return serializeJson(drawSettingsToJson(settings));
}

QJsonObject drawSettingsToJson(const Draw::Settings &settings)
{
    return QJsonObject {
        { DRAW_CONSECUTIVE,      settings.consecutive },
        { DRAW_TABLES_LIMIT,     settings.tablesLimit },
        { DRAW_RANDOM_DRAW,      settings.randomDraw },
        { DRAW_WINDOW_DRAW,      settings.windowDraw },
        { DRAW_CONSECUTIVE_DRAW, settings.consecutiveDraw }
    };
}

Draw::Settings parseDrawSettings(const QString &data)
{
    return parseDrawSettings(jsonObjectFromString(data));
}

Draw::Settings parseDrawSettings(const QJsonObject &jsonObject)
{
    const bool consecutive = jsonObject.value(DRAW_CONSECUTIVE).toBool();
    const int tablesLimit = jsonObject.value(DRAW_TABLES_LIMIT).toInt();
    const int randomDraw = jsonObject.value(DRAW_RANDOM_DRAW).toInt();
    const int windowDraw = jsonObject.value(DRAW_WINDOW_DRAW).toInt();
    const int consecutiveDraw = jsonObject.value(DRAW_CONSECUTIVE_DRAW).toInt();

    if (! consecutive
        && ((tablesLimit != -1 && (randomDraw > tablesLimit || consecutiveDraw > tablesLimit))
            || randomDraw >= consecutiveDraw
            || (randomDraw != -1 && consecutiveDraw != -1
                && consecutiveDraw == randomDraw + 1
                && windowDraw != -1))) {

        return Draw::DEFAULT_SETTINGS;
    }

    return Draw::Settings {
        consecutive,
        tablesLimit,
        randomDraw,
        windowDraw,
        consecutiveDraw
    };
}

const QString OPTIONAL_PAGES_DRAW_OVERVIEW = QStringLiteral("draw_overview");

const QHash<OptionalPages::Page, QString>  OPTIONAL_PAGES_ID {
    { OptionalPages::Page::DrawOverview, OPTIONAL_PAGES_DRAW_OVERVIEW }
};

const QHash<QString, OptionalPages::Page>  OPTIONAL_PAGES_ID_MAP {
    { OPTIONAL_PAGES_DRAW_OVERVIEW, OptionalPages::Page::DrawOverview }
};

QString optionalPagesToString(const QVector<OptionalPages::Page> &pages)
{
    return serializeJson(optionalPagesToJson(pages));
}

QJsonArray optionalPagesToJson(const QVector<OptionalPages::Page> &pages)
{
    QJsonArray ids;
    for (const OptionalPages::Page page : pages) {
        ids.append(OPTIONAL_PAGES_ID.value(page));
    }
    return ids;
}

QVector<OptionalPages::Page> parseOptionalPages(const QString &data)
{
    return parseOptionalPages(jsonArrayFromString(data));
}

QVector<OptionalPages::Page> parseOptionalPages(const QJsonArray &json)
{
    QVector<OptionalPages::Page> pages;
    for (const QJsonValue &value : json) {
        const QString id = value.toString();
        if (OPTIONAL_PAGES_ID_MAP.contains(id)) {
            pages.append(OPTIONAL_PAGES_ID_MAP.value(id));
        }
    }
    return pages;
}

// Legacy functions needed for updates
// ===================================

QVector<RegistrationColumns::Column> parsePlayersPageColumns(const QString &data)
{
    if (data.isEmpty()) {
        return RegistrationColumns::DEFAULT_COLUMNS;
    } else {
        return parsePlayersPageColumns(jsonObjectFromString(data));
    }
}

QVector<RegistrationColumns::Column> parsePlayersPageColumns(const QJsonObject &jsonObject)
{
    QVector<RegistrationColumns::Column> columns;
    for (const RegistrationColumns::Column column : RegistrationColumns::ALL_COLUMNS) {
        QJsonValue jsonValue = jsonObject.value(RegistrationColumns::COLUMN_ID.value(column));
        const bool visible = jsonValue.type() == QJsonValue::Bool
                                 ? jsonValue.toBool()
                                 : RegistrationColumns::DEFAULT_COLUMNS.contains(column);
        if (visible) {
            columns.append(column);
        }
    }
    return columns;
}

}
