/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes
#include "ResourceFinder.h"

// Qt includes
#include <QDebug>
#include <QFileInfo>
#include <QStandardPaths>
#include <QCoreApplication>
#include <QUrl>
#include <QDir>

// C++ includes
#include <utility>

namespace
{
const QLatin1String SLASH("/");
}

ResourceFinder::ResourceFinder(QObject *parent)
    : QObject(parent)
{
    QVector<QString> possibleLocations;
    const auto appDir = QCoreApplication::applicationDirPath();

#ifdef Q_OS_LINUX
    const QLatin1String resourcesPath("muckturnier/");

    // Add a possibly prefixed directory tree
    const auto prefixed = QUrl(appDir + SLASH).resolved(
                              QUrl(QStringLiteral("../share/"))).toString();
    possibleLocations.append(prefixed);
    possibleLocations.append(prefixed + resourcesPath);

    // Add system-wide locations
    const auto genericDataLocation = QStandardPaths::standardLocations(
                                         QStandardPaths::GenericDataLocation);
    for (const QString &path : genericDataLocation) {
        possibleLocations.append(path + SLASH);
        possibleLocations.append(path + SLASH + resourcesPath);
    }
#endif

#ifdef Q_OS_WIN
    possibleLocations.append(appDir + QLatin1String("/res/"));
#endif

#ifdef Q_OS_MACOS
    possibleLocations.append(QUrl(appDir + SLASH).resolved(
                                 QUrl(QStringLiteral("../Resources/"))).toString());
#endif

    // Assemble the list
    const QDir dir;
    for (const QString &location : std::as_const(possibleLocations)) {
        if (dir.exists(location) && ! m_possibleLocations.contains(location)) {
            m_possibleLocations.append(location);
        }
    }
}

QString ResourceFinder::find(const QString &file)
{
    // Check if we already have it
    if (m_path.contains(file)) {
        return m_path.value(file);
    }

    // ... then we check all other possible data locations
    for (const QString &path : std::as_const(m_possibleLocations)) {
        const QString filePath = path + file;
        if (QFileInfo::exists(filePath)) {
            m_path[file] = filePath;
            return filePath;
        }
    }

    // This will also return QString() if nothing is found
    return m_path[file];
}
