/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes
#include "SearchEngine.h"
#include "StringTools.h"

// Qt includes
#include <QRegularExpression>
#include <QDebug>

namespace
{

const QVector<QChar> UMLAUTS_TO_REPLACE {
    QChar(0x00e4), // ä
    QChar(0x00f6), // ö
    QChar(0x00fc), // ü
    QChar(0x00E6), // æ
    QChar(0x0153), // œ
    QChar(0x00F8)  // ø
};

const QVector<QString> UMLAUT_REPLACEMENTS {
    QStringLiteral("ae"),
    QStringLiteral("oe"),
    QStringLiteral("ue"),
    QStringLiteral("ae"),
    QStringLiteral("oe"),
    QStringLiteral("oe")
};

const QVector<QString> REPLACED_UMLAUTS {
    QStringLiteral("ae"),
    QStringLiteral("oe"),
    QStringLiteral("ue"),
};

const QVector<QString> REPLACED_UMLAUTS_MAPPING {
    QStringLiteral("a"),
    QStringLiteral("o"),
    QStringLiteral("u"),
};

const QChar SPACE(QLatin1Char(' '));

const QRegularExpression SEQUENTIAL_CHARACTERS(QStringLiteral("(.)\\1+"));

}

namespace Koeln
{

const QChar ZERO (QLatin1Char('0'));
const QChar ONE  (QLatin1Char('1'));
const QChar TWO  (QLatin1Char('2'));
const QChar THREE(QLatin1Char('3'));
const QChar FOUR (QLatin1Char('4'));
const QChar FIVE (QLatin1Char('5'));
const QChar SIX  (QLatin1Char('6'));
const QChar SEVEN(QLatin1Char('7'));
const QChar EIGHT(QLatin1Char('8'));

const QLatin1String FOUR_EIGHT("48");

const QChar A(QLatin1Char('a'));
const QChar B(QLatin1Char('b'));
const QChar C(QLatin1Char('c'));
const QChar D(QLatin1Char('d'));
const QChar E(QLatin1Char('e'));
const QChar F(QLatin1Char('f'));
const QChar G(QLatin1Char('g'));
const QChar H(QLatin1Char('h'));
const QChar I(QLatin1Char('i'));
const QChar J(QLatin1Char('j'));
const QChar K(QLatin1Char('k'));
const QChar L(QLatin1Char('l'));
const QChar M(QLatin1Char('m'));
const QChar N(QLatin1Char('n'));
const QChar O(QLatin1Char('o'));
const QChar P(QLatin1Char('p'));
const QChar Q(QLatin1Char('q'));
const QChar R(QLatin1Char('r'));
const QChar S(QLatin1Char('s'));
const QChar T(QLatin1Char('t'));
const QChar U(QLatin1Char('u'));
const QChar V(QLatin1Char('v'));
const QChar W(QLatin1Char('w'));
const QChar X(QLatin1Char('x'));
const QChar Y(QLatin1Char('y'));
const QChar Z(QLatin1Char('z'));

}

SearchEngine::SearchEngine(QObject *parent, StringTools *stringTools)
    : QObject(parent),
      m_stringTools(stringTools),
      m_variantsCache {
          { SearchType::Default,         &m_defaultVariants },
          { SearchType::DrawnPairsScore, &m_drawnPairsScoreVariants },
          { SearchType::DrawnPairsDraw,  &m_drawnPairsDrawVariants }
      }
{
}

QVector<QString> SearchEngine::searchVariants(const QString &text)
{
    QVector<QString> variants;

    // We only work on a lower-case, simplified version of the string with "ß" replaced by "s" in
    // each case (the replacement with "ss" would be removed by compressing the string anyway):
    const QString workText = m_stringTools->toLower(text.simplified().replace(QStringLiteral("ß"),
                                                                              QStringLiteral("s")));

    // Index 0 has to be the phonetic representation.
    // We calculate the Kölner Phonetik representation of the name and add it. We use the
    // simplified version with "ß" replaced for this but additionally strip diacritical marks.
    // We need the un-compressed version for the algorithm to work properly.
    variants.append(koelnerPhonetik(stripDiacritis(workText)));

    // From here on, we only work on the compressed version
    // (with sequential characters replaced by one character)
    const QString compressed = removeSequentialCharacters(workText);

    // Generate the version with diacritial marks stripped
    const QString diacriticsStripped = stripDiacritis(compressed);
    variants.append(diacriticsStripped);

    // Generate the version with Umlauts replaced by two characters
    // and the remainung diacritial marks stripped afterwards
    QString umlautsReplaced = compressed;
    for (int i = 0; i < UMLAUTS_TO_REPLACE.size(); i++) {
        umlautsReplaced.replace(UMLAUTS_TO_REPLACE.at(i), UMLAUT_REPLACEMENTS.at(i));
    }
    umlautsReplaced = stripDiacritis(umlautsReplaced);
    if (! variants.contains(umlautsReplaced)) {
        variants.append(umlautsReplaced);
    }

    // ... and the same for the opposite direction: Two characters replaced by the corresponding
    // Umlaut and all diacritial marks stripped afterwards.
    // We need this so that if we have e. g. "Bär / Rödel", a search for "bar roed" will also match.
    QString replacementsToUmlautMapping = compressed;
    for (int i = 0; i < REPLACED_UMLAUTS.size(); i++) {
        replacementsToUmlautMapping.replace(REPLACED_UMLAUTS.at(i), REPLACED_UMLAUTS_MAPPING.at(i));
    }
    replacementsToUmlautMapping = stripDiacritis(replacementsToUmlautMapping);
    if (! variants.contains(replacementsToUmlautMapping)) {
        variants.append(replacementsToUmlautMapping);
    }

    return variants;
}

QString SearchEngine::removeSequentialCharacters(QString text) const
{
    QRegularExpressionMatch match = SEQUENTIAL_CHARACTERS.match(text);
    while (match.hasMatch()) {
        text.replace(match.captured(0), match.captured(1));
        match = SEQUENTIAL_CHARACTERS.match(text);
    }
    return text;
}

QString SearchEngine::stripDiacritis(const QString &text) const
{
    // Here, we strip out all diacritical marks and accents by first separating the combinig
    // characters and then removing them, cf. http://en.wikipedia.org/wiki/Combining_character
    QString replaced = text.normalized(QString::NormalizationForm_D);
    for (int i = 0x0300; i <= 0x036F; i++) {
        replaced.remove(QChar(i));
    }
    return replaced;
}

bool SearchEngine::checkMatch(const QString &text, SearchType searchType, bool phoneticSearch) const
{
    if (text.isEmpty()) {
        return false;
    }

    return checkMatch(m_variantsCache.value(searchType)->value(text), phoneticSearch);
}

bool SearchEngine::checkMatch(const QVector<QString> &variantList, bool phoneticSearch) const
{
    const int variantListCount = variantList.count();
    const int start = phoneticSearch ? 0 : 1;

    for (int searchTermIndex = start; searchTermIndex < m_searchTerm.count(); searchTermIndex++) {
        const auto wordsCount = m_searchTerm.at(searchTermIndex);
        const auto words = wordsCount.keys();

        for (int variantsIndex = start; variantsIndex < variantListCount; variantsIndex++) {
            const QString &variant = variantList.at(variantsIndex);

            bool match = true;
            for (const QString &word : words) {
                if (variant.count(word) < wordsCount.value(word)) {
                    match = false;
                    break;
                }
            }

            if (match) {
                return true;
            }
        }
    }

    return false;
}

void SearchEngine::clearSearchCache(SearchType searchType)
{
    (*m_variantsCache.value(searchType)).clear();
}

void SearchEngine::updateSearchCache(const QString &name, SearchType searchType)
{
    (*m_variantsCache.value(searchType)).insert(name, searchVariants(name));
}

void SearchEngine::setSearchTerm(const QString &text)
{
    setSearchTerm(searchVariants(text));
}

void SearchEngine::setSearchTerm(const QVector<QString> &variantsList)
{
    m_searchTerm.clear();
    for (const QString &variant : variantsList) {
        QHash<QString, int> variantWords;
        const auto splitted = variant.split(SPACE);
        for (const QString &word : splitted) {
            variantWords[word]++;
        }
        m_searchTerm.append(variantWords);
    }
}

QString SearchEngine::koelnerPhonetik(const QString &text) const
{
    // Calculate the Kölner Phonetik representation of a string
    // cf. https://de.wikipedia.org/wiki/Kölner_Phonetik

    // WARNING
    // This function expects text to be lower-case, "ß" replaced by "s", Umlaut diacritical marks
    // stripped and all words only separated by one space.
    // If it's used in some other context, be sure to do the proper preparation here!

    QStringList phonetic;

    const auto words = text.split(SPACE);
    for (QString word : words) {
        // Step 1: Translate the string using the translation table
        QString translated;
        const int count = word.count();

        for (int i = 0; i < count; i++) {
            const QChar current = word.at(i);
            const QChar before = i > 0 ? word.at(i - 1) : QChar();
            const QChar after = i < count - 1 ? word.at(i + 1) : QChar();

            // Keep spaces between words
            if (current == SPACE) {
                translated += SPACE;

            // a, e, i, j, o, u, y
            } else if (current == Koeln::A
                    || current == Koeln::E
                    || current == Koeln::I
                    || current == Koeln::J
                    || current == Koeln::O
                    || current == Koeln::U
                    || current == Koeln::Y) {

                translated += Koeln::ZERO;

            // b
            } else if (current == Koeln::B) {
                translated += Koeln::ONE;

            // p, but not before h
            } else if (current == Koeln::P
                    && after != Koeln::H) {

                translated += Koeln::ONE;

            // d, t, but not before c, s, z
            } else if ((current == Koeln::D
                        || current == Koeln::T)
                    && (   after != Koeln::C
                        && after != Koeln::S
                        && after != Koeln::Z)) {

                translated += Koeln::TWO;

            // f, v, w
            } else if (current == Koeln::F
                    || current == Koeln::V
                    || current == Koeln::W) {

                translated += Koeln::THREE;

            // p before h
            } else if (current == Koeln::P
                    && after == Koeln::H) {

                translated += Koeln::THREE;

            // g, k, q
            } else if (current == Koeln::G
                    || current == Koeln::K
                    || current == Koeln::Q) {

                translated += Koeln::FOUR;

            // c at the beginning before a, h, k, l, o, q, r, u, x
            } else if (i == 0
                    && current == Koeln::C
                    && (after == Koeln::A
                        || after == Koeln::H
                        || after == Koeln::K
                        || after == Koeln::L
                        || after == Koeln::O
                        || after == Koeln::Q
                        || after == Koeln::R
                        || after == Koeln::U
                        || after == Koeln::X)) {

                translated += Koeln::FOUR;

            // c after the beginning before a, h, k, o, q, u, x but not after s, z
            } else if (i > 0
                    && current == Koeln::C
                    && (before != Koeln::S
                        && before != Koeln::Z)
                    && (after == Koeln::A
                        || after == Koeln::H
                        || after == Koeln::K
                        || after == Koeln::O
                        || after == Koeln::Q
                        || after == Koeln::U
                        || after == Koeln::X)) {

                translated += Koeln::FOUR;

            // x but not after c, k, q
            } else if (current == Koeln::X
                    && (before != Koeln::C
                        && before != Koeln::K
                        && before != Koeln::Q)) {

                translated += Koeln::FOUR_EIGHT;

            // l
            } else if (current == Koeln::L) {
                translated += Koeln::FIVE;

            // m, n
            } else if (current == Koeln::M
                    || current == Koeln::N) {

                translated += Koeln::SIX;

            // r
            } else if (current == Koeln::R) {
                translated += Koeln::SEVEN;

            // s, z
            } else if (current == Koeln::S
                    || current == Koeln::Z) {

                translated += Koeln::EIGHT;

            // c after s, z
            } else if (current == Koeln::C
                    && (before == Koeln::S
                        || before == Koeln::Z)) {

                translated += Koeln::EIGHT;

            // c at the beginning, but not before a, h, k, l, o, q, r, u, x
            } else if (i == 0
                    && current == Koeln::C
                    && (after != Koeln::A
                        && after != Koeln::H
                        && after != Koeln::K
                        && after != Koeln::L
                        && after != Koeln::O
                        && after != Koeln::Q
                        && after != Koeln::R
                        && after != Koeln::U
                        && after != Koeln::X)) {

                translated += Koeln::EIGHT;

            // c, but not before a, h, k, o, q, u, x
            } else if (current == Koeln::C
                    && (after != Koeln::A
                        && after != Koeln::H
                        && after != Koeln::K
                        && after != Koeln::O
                        && after != Koeln::Q
                        && after != Koeln::U
                        && after != Koeln::X)) {

                translated += Koeln::EIGHT;

            // d, t before c, s, z
            } else if ((current == Koeln::D
                        || current == Koeln::T)
                    && (after == Koeln::C
                        || after == Koeln::S
                        || after == Koeln::Z)) {

                translated += Koeln::EIGHT;

            // x after c, k, q
            } else if (current == Koeln::X
                    && (before == Koeln::C
                        || before == Koeln::K
                        || before == Koeln::Q)) {

                translated += Koeln::EIGHT;
            }

        }

        // Step 2: Replace sequences of numbers with only one
        translated = removeSequentialCharacters(translated);

        // Step 3: Remove all occurences of 0 that are not at the start
        // This has to be done in an extra step: Sequences of numbers can appear after doing this.
        // If we don't add the zeroes in the first place, those would be eliminated by step 2.
        QString zeroesStripped;
        for (int i = 0; i < translated.count(); i++) {
            const QChar current = translated.at(i);
            if (current != Koeln::ZERO || i == 0) {
                zeroesStripped.append(current);
            }
        }

        // Add the result to the phonetic representation word list (if it was a translatable word)
        if (! zeroesStripped.isEmpty()) {
            phonetic.append(zeroesStripped);
        }
    }

    // Join the list and return it
    return phonetic.join(SPACE);
}

QHash<QString, QStringList> SearchEngine::checkListForPossibleDuplicates(const QStringList &list,
                                                                         bool phoneticSearch)
{
    QHash<QString, QStringList> possibleDuplicates;
    const int count = list.count();

    if (count == 1) {
        return possibleDuplicates;
    }

    QHash<QString, QVector<QString>> variantsCache;
    for (const QString &name : list) {
        variantsCache.insert(name, searchVariants(name));
    }

    // We actually do have to compare each entry to each other, because if we "only" do an upper
    // triangular matrix comparison, we lose potentional duplicates of entries that are on a higher
    // position in the list, because they aren't compared to the lower indices anymore.

    for (int searchIndex = 0; searchIndex < count; searchIndex++) {
        setSearchTerm(variantsCache.value(list.at(searchIndex)));

        for (int compareIndex = 0; compareIndex < count; compareIndex++) {
            if (searchIndex == compareIndex) {
                continue;
            }

            if (checkMatch(variantsCache.value(list.at(compareIndex).simplified()),
                phoneticSearch)) {

                possibleDuplicates[list.at(searchIndex)].append(list.at(compareIndex));
            }
        }
    }

    return possibleDuplicates;
}
