/* Copyright (c) 2010-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes
#include "SearchWidget.h"
#include "SearchLineEdit.h"
#include "PhoneticSearchButton.h"
#include "SharedStyles.h"

// Qt includes
#include <QHBoxLayout>
#include <QLabel>
#include <QDebug>

SearchWidget::SearchWidget(const QString &searchId, SharedObjects *sharedObjects,
                           QWidget *parent) : QWidget(parent)
{
    QHBoxLayout *layout = new QHBoxLayout(this);

    m_searchLabel = new QLabel(tr("Suche:"));
    layout->addWidget(m_searchLabel);

    m_search = new SearchLineEdit;
    m_search->setClearButtonEnabled(true);
    connect(m_search, &QLineEdit::textChanged, this, &SearchWidget::processInput);
    layout->addWidget(m_search);

    m_phoneticSearch = new PhoneticSearchButton(searchId, sharedObjects);
    connect(m_phoneticSearch, &QToolButton::toggled,
            this, [this] { emit searchChanged(m_search->text()); });
    layout->addWidget(m_phoneticSearch);

    m_matches = new QLabel;
    m_matches->hide();
    layout->addWidget(m_matches);
}

void SearchWidget::updateMatches(int matches, int hiddenMatches)
{
    m_matches->show();

    QString text = matches == 0 ? tr("<span style=\"%1\">Keine Treffer</span>").arg(
                                     SharedStyles::RED_TEXT)
                                : tr("%1 Treffer").arg(matches);
    if (hiddenMatches > 0) {
        text.append(tr(" – <span style=\"%1\">%2 ausgeblendet</span>").arg(
                       SharedStyles::RED_TEXT, QString::number(hiddenMatches)));
    }

    m_matches->setText(text);
}

void SearchWidget::processInput(const QString &text)
{
    emit searchChanged(text);
    if (text.isEmpty()) {
        m_matches->hide();
    }
}

void SearchWidget::clear()
{
    m_search->clear();
    m_savedSearch.clear();
}

void SearchWidget::saveSearch()
{
    if (! m_savedSearch.isEmpty()) {
        return;
    }

    m_savedSearch = m_search->text();
    m_search->clear();
}

void SearchWidget::restoreSearch()
{
    if (m_savedSearch.isEmpty()) {
        return;
    }

    m_search->setText(m_savedSearch);
    m_savedSearch.clear();
}

bool SearchWidget::phoneticSearch() const
{
    return m_phoneticSearch->isChecked();
}

QString SearchWidget::searchText() const
{
    return m_search->text();
}
