/* Copyright (c) 2019-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes
#include "Settings.h"
#include "DefaultValues.h"
#include "JsonHelper.h"
#include "OptionalPages.h"

// Qt includes
#include <QDebug>
#include <QAction>
#include <QJsonObject>
#include <QJsonArray>
#include <QStandardPaths>
#include <QDir>

// C++ includes
#include <utility>

// Config file string literals
// ===========================

namespace
{

// Options check state group
const QLatin1String SAVE_ON_CLOSE("save_on_close/");
const QString SAVE_ON_CLOSE_MAINWINDOW_GEOMETRY
    = SAVE_ON_CLOSE + QLatin1String("mainwindow_geometry");
const QString SAVE_ON_CLOSE_MAINWINDOW_STATE
    = SAVE_ON_CLOSE + QLatin1String("mainwindow_state");
const QString SAVE_ON_CLOSE_LAST_DB_PATH
    = SAVE_ON_CLOSE + QLatin1String("last_db_path");
const QString SAVE_ON_CLOSE_STOPWATCH_STATE
    = SAVE_ON_CLOSE + QLatin1String("stopwatch_state");
const QString SAVE_ON_CLOSE_PHONETIC_SEARCH
    = SAVE_ON_CLOSE + QLatin1String("phonetic_search");

// Session data
const QString SESSION_DATA = QStringLiteral("session_data/");
const QString SESSION_DATA_TAB_POSITION
    = SESSION_DATA + QLatin1String("tab_position");
const QString SESSION_DATA_MAINWINDOW_GEOMETRY
    = SESSION_DATA + QLatin1String("mainwindow_geometry");
const QString SESSION_DATA_MAINWINDOW_STATE
    = SESSION_DATA + QLatin1String("mainwindow_state");
const QString SESSION_DATA_LAST_DB_PATH
    = SESSION_DATA + QLatin1String("last_db_path");
const QString NEW_TOURNAMENT_SHOW_EXTENDED_OPTIONS
    = SESSION_DATA + QLatin1String("new_tournament_show_extended_options");

// Stop watch state
const QLatin1String STOPWATCH_STATE("stopwatch_state/");
const QString STOPWATCH_STATE_VERSION
    = STOPWATCH_STATE + QLatin1String("version");
const QString STOPWATCH_STATE_HOURS
    = STOPWATCH_STATE + QLatin1String("hours");
const QString STOPWATCH_STATE_MINUTES
    = STOPWATCH_STATE + QLatin1String("minutes");
const QString STOPWATCH_STATE_SECONDS
    = STOPWATCH_STATE + QLatin1String("seconds");
const QString STOPWATCH_STATE_TITLE
    = STOPWATCH_STATE + QLatin1String("title");
const QString STOPWATCH_STATE_GEOMETRY
    = STOPWATCH_STATE + QLatin1String("geometry");

// Phonetic search state
const QString PHONETIC_SEARCH = QStringLiteral("phonetic_search");

// Strings
const QLatin1String STRINGS("strings/");
const QString STRINGS_BOOGER_SYMBOL = STRINGS + QLatin1String("booger_symbol");
const QString STRINGS_NAMES_SEPARATOR = STRINGS + QLatin1String("names_separator");

// Tournament template

const QString TOURNAMENT_TEMPLATE = QStringLiteral("tournament_template");
const QString TOURNAMENT_TEMPLATE_VERSION
    = TOURNAMENT_TEMPLATE + QLatin1String("/version");
const QString TOURNAMENT_TEMPLATE_DB_FILE_NAME_TEMPLATE
    = TOURNAMENT_TEMPLATE + QLatin1String("/db_file_name_template");
const QString TOURNAMENT_TEMPLATE_TOURNAMENT_MODE
    = TOURNAMENT_TEMPLATE + QLatin1String("/tournament_mode");
const QString TOURNAMENT_TEMPLATE_SCORE_TYPE
    = TOURNAMENT_TEMPLATE + QLatin1String("/score_type");
const QString TOURNAMENT_TEMPLATE_BOOGER_SCORE
    = TOURNAMENT_TEMPLATE + QLatin1String("/booger_score");
const QString TOURNAMENT_TEMPLATE_BOOGERS_PER_ROUND
    = TOURNAMENT_TEMPLATE + QLatin1String("/boogers_per_round");
const QString TOURNAMENT_TEMPLATE_REGISTRATION_COLUMNS
    = TOURNAMENT_TEMPLATE + QLatin1String("/registration_columns");
const QString TOURNAMENT_TEMPLATE_REGISTRATION_LIST_HEADER_HIDDEN
    = TOURNAMENT_TEMPLATE + QLatin1String("/registration_list_header_hidden");
const QString TOURNAMENT_TEMPLATE_AUTO_SELECT_PAIRS
    = TOURNAMENT_TEMPLATE + QLatin1String("/auto_select_pairs");
const QString TOURNAMENT_TEMPLATE_SELECT_BY_LAST_ROUND
    = TOURNAMENT_TEMPLATE + QLatin1String("/select_by_last_round");
const QString TOURNAMENT_TEMPLATE_NO_TABLE_NUMBERS
    = TOURNAMENT_TEMPLATE + QLatin1String("/no_table_numbers");
const QString TOURNAMENT_TEMPLATE_INCLUDE_OPPONENT_GOALS
    = TOURNAMENT_TEMPLATE + QLatin1String("/include_opponent_goals");
const QString TOURNAMENT_TEMPLATE_GRAY_OUT_UNIMPORTANT_GOALS
    = TOURNAMENT_TEMPLATE + QLatin1String("/gray_out_unimportant_goals");
const QString TOURNAMENT_TEMPLATE_OPTIONAL_PAGES
    = TOURNAMENT_TEMPLATE + QLatin1String("/optional_pages");

// Markers template

constexpr int VERSION_TEMPLATE_MARKERS = 2;

const QString MARKERS_TEMPLATE = QStringLiteral("markers_template");

const QString MARKERS_TEMPLATE_VERSION = MARKERS_TEMPLATE + QLatin1String("/version");
const QString MARKERS_TEMPLATE_MARKERS = MARKERS_TEMPLATE + QLatin1String("/markers");
const QLatin1String MARKERS_TEMPLATE_MARKERS_JSON_NAME   ("n");
const QLatin1String MARKERS_TEMPLATE_MARKERS_JSON_COLOR  ("c");
const QLatin1String MARKERS_TEMPLATE_MARKERS_JSON_SORTING("s");

const QHash<Markers::Type, QString> MARKERS_TEMPLATE_MARKER_KEY {
    { Markers::Default, MARKERS_TEMPLATE + QLatin1String("/default_marker") },
    { Markers::Drawn,   MARKERS_TEMPLATE + QLatin1String("/drawn_marker") },
    { Markers::Absent,  MARKERS_TEMPLATE + QLatin1String("/absent_marker") }
};

const QHash<Markers::Type, int> MARKERS_TEMPLATE_MARKER_DEFAULT_VALUE {
    { Markers::Default, 0 },
    { Markers::Drawn, -1 },
    { Markers::Absent,  -1 }
};

// Singles Management

const QString SINGLES_MANAGEMENT_TEMPLATE = QStringLiteral("singles_management_template");

const QString SINGLES_MANAGEMENT_TEMPLATE_VERSION
    = SINGLES_MANAGEMENT_TEMPLATE + QLatin1String("/version");
const QString SINGLES_MANAGEMENT_TEMPLATE_DATA
    = SINGLES_MANAGEMENT_TEMPLATE + QLatin1String("/data");

// Network settings

const QString NETWORK = QStringLiteral("network");

const QString NETWORK_PREFER_IPV6       = NETWORK + QLatin1String("/prefer_ipv6");

const QString NETWORK_SERVER_PORT       = NETWORK + QLatin1String("/server_port");
const QString NETWORK_CONNECT_TIMEOUT   = NETWORK + QLatin1String("/connect_timeout");
const QString NETWORK_REQUEST_TIMEOUT   = NETWORK + QLatin1String("/request_timeout");

const QString NETWORK_DISCOVER_PORT     = NETWORK + QLatin1String("/discover_port");
const QString NETWORK_DISCOVER_TIMEOUT  = NETWORK + QLatin1String("/discover_timeout");
const QString NETWORK_DISCOVER_ATTEMPTS = NETWORK + QLatin1String("/discover_attempts");

// QTabWidget::TabPosition map
const QVector<QString> POSSIBLE_TAB_POSITIONS {
    QStringLiteral("north"),
    QStringLiteral("south"),
    QStringLiteral("west"),
    QStringLiteral("east")
};

// Draw settings
const QString DRAW_SETTINGS_TEMPLATE = QStringLiteral("draw_settings_template");
const QString DRAW_SETTINGS_TEMPLATE_VERSION = DRAW_SETTINGS_TEMPLATE + QLatin1String("/version");
const QString DRAW_SETTINGS_TEMPLATE_DATA    = DRAW_SETTINGS_TEMPLATE + QLatin1String("/data");

}

// Shared strings
// ==============

const QString Settings::registrationPhoneticId = QStringLiteral("registration_page_register");

// Settings functions
// ==================

Settings::Settings(QObject *parent)
    : QSettings(QStringLiteral("muckturnier.org"), QStringLiteral("muckturnier"), parent)
{
    m_boogerSymbol = value(STRINGS_BOOGER_SYMBOL, DefaultValues::BOOGER_SYMBOL).toString();
    if (m_boogerSymbol.left(1).simplified().isEmpty()) {
        m_boogerSymbol = DefaultValues::BOOGER_SYMBOL;
    }

    m_namesSeparator = value(STRINGS_NAMES_SEPARATOR, DefaultValues::NAMES_SEPARATOR).toString();
    m_namesSeparator.remove(QLatin1Char('\t'));
}

void Settings::setSaveOnCloseValue(const QString &key, const QString &dataSet, bool state)
{
    setValue(key, state);
    if (! state) {
        remove(dataSet);
    }
}

bool Settings::saveOnCloseValue(const QString &key) const
{
    const QVariant data = value(key);
    return data.isValid() ? data.toBool() : true;
}

// Tab position
// ------------

void Settings::setTabPosition(const QString &position)
{
    setValue(SESSION_DATA_TAB_POSITION, position);
}

QString Settings::tabPosition() const
{
    const QString position = value(SESSION_DATA_TAB_POSITION, QStringLiteral("south")).toString();
    if (POSSIBLE_TAB_POSITIONS.contains(position)) {
        return position;
    } else {
        return QStringLiteral("south");
    }
}

// Main window geometry
// --------------------

bool Settings::saveMainWindowGeometryEnabled() const
{
    return saveOnCloseValue(SAVE_ON_CLOSE_MAINWINDOW_GEOMETRY);
}

void Settings::enableSaveMainWindowGeometry(bool state)
{
    setSaveOnCloseValue(SAVE_ON_CLOSE_MAINWINDOW_GEOMETRY, SESSION_DATA_MAINWINDOW_GEOMETRY, state);
}

void Settings::saveMainWindowGeometry(const QByteArray &data)
{
    setValue(SESSION_DATA_MAINWINDOW_GEOMETRY, data);
}

QByteArray Settings::mainWindowGeometry() const
{
    return value(SESSION_DATA_MAINWINDOW_GEOMETRY, QByteArray()).toByteArray();
}

// Main window state
// -----------------

bool Settings::saveMainWindowStateEnabled() const
{
    return saveOnCloseValue(SAVE_ON_CLOSE_MAINWINDOW_STATE);
}

void Settings::enableSaveMainWindowState(bool state)
{
    setSaveOnCloseValue(SAVE_ON_CLOSE_MAINWINDOW_STATE, SESSION_DATA_MAINWINDOW_STATE, state);
}

void Settings::saveMainWindowState(const QByteArray &data)
{
    if (! data.isEmpty()) {
        setValue(SESSION_DATA_MAINWINDOW_STATE, data);
    } else {
        remove(SESSION_DATA_MAINWINDOW_STATE);
    }
}

QByteArray Settings::mainWindowState() const
{
    return value(SESSION_DATA_MAINWINDOW_STATE, QByteArray()).toByteArray();
}

// Last database path
// ------------------

bool Settings::saveLastDbPathEnabled() const
{
    return saveOnCloseValue(SAVE_ON_CLOSE_LAST_DB_PATH);
}

void Settings::enableSaveLastDbPath(bool state)
{
    setSaveOnCloseValue(SAVE_ON_CLOSE_LAST_DB_PATH, SESSION_DATA_LAST_DB_PATH, state);
}

void Settings::saveLastDbPath(const QString &dbPath)
{
    setValue(SESSION_DATA_LAST_DB_PATH, dbPath);
}

QString Settings::lastDbPath() const
{
    const QString documentsLocation
        = QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation);
    const QString path = value(SESSION_DATA_LAST_DB_PATH, documentsLocation).toString();
    return QDir(path).exists() ? path : documentsLocation;
}

// Stop watch state
// ----------------

void Settings::enableSaveStopWatchState(bool state)
{
    setSaveOnCloseValue(SAVE_ON_CLOSE_STOPWATCH_STATE, STOPWATCH_STATE, state);
}

bool Settings::saveStopWatchStateEnabled() const
{
    return saveOnCloseValue(SAVE_ON_CLOSE_STOPWATCH_STATE);
}

void Settings::saveStopWatchState(StopWatchState::State state)
{
    setValue(STOPWATCH_STATE_VERSION,  StopWatchState::VERSION);
    setValue(STOPWATCH_STATE_HOURS,    state.hours);
    setValue(STOPWATCH_STATE_MINUTES,  state.minutes);
    setValue(STOPWATCH_STATE_SECONDS,  state.seconds);
    setValue(STOPWATCH_STATE_TITLE,    state.title);
    setValue(STOPWATCH_STATE_GEOMETRY, state.geometry);
}

StopWatchState::State Settings::stopWatchState() const
{
    if (! saveStopWatchStateEnabled()
        || value(STOPWATCH_STATE_VERSION, 0).toInt() != StopWatchState::VERSION) {

        return StopWatchState::DEFAULT_STATE;
    }

    int hours = value(STOPWATCH_STATE_HOURS, 0).toInt();
    if (hours < 0 || hours > 99) {
        hours = 0;
    }

    int minutes = value(STOPWATCH_STATE_MINUTES, 0).toInt();
    if (minutes < 0 || minutes > 59) {
        minutes = 0;
    }

    int seconds = value(STOPWATCH_STATE_SECONDS, 0).toInt();
    if (seconds < 0 || seconds > 59) {
        seconds = 0;
    }

    return StopWatchState::State {
        hours,
        minutes,
        seconds,
        value(STOPWATCH_STATE_TITLE, 0).toString(),
        value(STOPWATCH_STATE_GEOMETRY, QByteArray()).toByteArray()
    };
}

// "Extended options" of the "New tournament" dialog

bool Settings::newTournamentShowExtendedOptions() const
{
    return value(NEW_TOURNAMENT_SHOW_EXTENDED_OPTIONS, false).toBool();
}

void Settings::setNewTournamentShowExtendedOptions(bool state)
{
    setValue(NEW_TOURNAMENT_SHOW_EXTENDED_OPTIONS, state);
}

// Phonetic search state

void Settings::enableSavePhoneticSearchState(bool state)
{
    setSaveOnCloseValue(SAVE_ON_CLOSE_PHONETIC_SEARCH, PHONETIC_SEARCH, state);
}

bool Settings::savePhoneticSearchStateEnabled() const
{
    return saveOnCloseValue(SAVE_ON_CLOSE_PHONETIC_SEARCH);
}

bool Settings::phoneticSearchState(const QString &searchId) const
{
    const bool defaultValue = searchId == Settings::registrationPhoneticId
                                  ? true : false;
    return value(PHONETIC_SEARCH + QStringLiteral("/") + searchId, defaultValue).toBool();
}

void Settings::savePhoneticSearchState(const QString &searchId, bool state)
{
    if (savePhoneticSearchStateEnabled()) {
        setValue(PHONETIC_SEARCH + QStringLiteral("/") + searchId, state);
    }
}

// Tournament template
// -------------------

void Settings::saveTournamentTemplate(const TournamentTemplate::Settings &tournamentTemplate)
{
    const int version = value(TOURNAMENT_TEMPLATE_VERSION, -1).toInt();
    if (version < 3) {
        // Clean up the old "players page" category
        const QString playersPage = QStringLiteral("players_page");

        // The "columns" key has been moved to "tournament_template/players_page_columns"
        remove(playersPage + QLatin1String("/columns"));

        // Check if there are other keys in "players_page" (maybe added by later versions)
        beginGroup(playersPage);
        const bool otherKeysExist = allKeys().count() > 0;
        endGroup();

        // If not, remove the "players_page" category
        if (! otherKeysExist) {
            remove(playersPage);
        }
    }

    setValue(TOURNAMENT_TEMPLATE_VERSION, TournamentTemplate::VERSION);
    setValue(TOURNAMENT_TEMPLATE_TOURNAMENT_MODE,
             Tournament::MODE_MAP.value(tournamentTemplate.tournamentMode));
    setValue(TOURNAMENT_TEMPLATE_SCORE_TYPE,
             Tournament::SCORE_TYPE_MAP.value(tournamentTemplate.scoreType));
    setValue(TOURNAMENT_TEMPLATE_BOOGER_SCORE, tournamentTemplate.boogerScore);
    setValue(TOURNAMENT_TEMPLATE_BOOGERS_PER_ROUND, tournamentTemplate.boogersPerRound);
    setValue(TOURNAMENT_TEMPLATE_REGISTRATION_COLUMNS,
             JsonHelper::registrationColumnsToString(tournamentTemplate.registrationColumns));
    setValue(TOURNAMENT_TEMPLATE_REGISTRATION_LIST_HEADER_HIDDEN,
             tournamentTemplate.registrationListHeaderHidden);
    setValue(TOURNAMENT_TEMPLATE_AUTO_SELECT_PAIRS, tournamentTemplate.autoSelectPairs);
    setValue(TOURNAMENT_TEMPLATE_SELECT_BY_LAST_ROUND, tournamentTemplate.selectByLastRound);
    setValue(TOURNAMENT_TEMPLATE_NO_TABLE_NUMBERS, tournamentTemplate.noTableNumbers);
    setValue(TOURNAMENT_TEMPLATE_INCLUDE_OPPONENT_GOALS, tournamentTemplate.includeOpponentGoals);
    setValue(TOURNAMENT_TEMPLATE_GRAY_OUT_UNIMPORTANT_GOALS,
             tournamentTemplate.grayOutUnimportantGoals);
    setValue(TOURNAMENT_TEMPLATE_OPTIONAL_PAGES,
             JsonHelper::optionalPagesToString(tournamentTemplate.optionalPages));

    saveDrawSettingsTemplate(tournamentTemplate.drawSettings);
}

TournamentTemplate::Settings Settings::tournamentTemplate() const
{
    const int version = value(TOURNAMENT_TEMPLATE_VERSION, -1).toInt();
    if (version <= 0 || version > TournamentTemplate::VERSION) {
        return TournamentTemplate::DEFAULT_SETTINGS;
    }

    // Get the tournament mode
    Tournament::Mode tournamentMode;
    if (version >= 2) {
        // Check for a correct value of TOURNAMENT_TEMPLATE_TOURNAMENT_MODE
        const QString tournamentModeString
            = value(TOURNAMENT_TEMPLATE_TOURNAMENT_MODE,
                    Tournament::MODE_FIXED_PAIRS).toString();
        if (! Tournament::MODE_MAP.values().contains(tournamentModeString)) {
            return TournamentTemplate::DEFAULT_SETTINGS;
        } else {
            tournamentMode = Tournament::MODE_MAP.key(tournamentModeString);
        }
    } else if (version < 2) {
        // Read the old bool "pair_mode" value
        tournamentMode = value(TOURNAMENT_TEMPLATE + QLatin1String("/pair_mode"), true).toBool()
                             ? Tournament::Mode::FixedPairs : Tournament::Mode::SinglePlayers;
    }

    // Get the score type
    Tournament::ScoreType scoreType;
    const QString scoreTypeString = value(TOURNAMENT_TEMPLATE_SCORE_TYPE,
                                          Tournament::SCORE_TYPE_HORIZONTAL).toString();
    if (! Tournament::SCORE_TYPE_MAP.values().contains(scoreTypeString)) {
        return TournamentTemplate::DEFAULT_SETTINGS;
    } else {
        scoreType = Tournament::SCORE_TYPE_MAP.key(scoreTypeString);
    }

    // Check for correct booger score and boogers per round
    const int boogerScore = value(TOURNAMENT_TEMPLATE_BOOGER_SCORE,
                                  DefaultValues::BOOGER_SCORE).toInt();
    const int boogersPerRound = value(TOURNAMENT_TEMPLATE_BOOGERS_PER_ROUND,
                                      DefaultValues::BOOGERS_PER_ROUND).toInt();
    if (boogerScore < 1 || boogersPerRound < 1) {
        return TournamentTemplate::DEFAULT_SETTINGS;
    }

    // Get the players page columns
    QVector<RegistrationColumns::Column> registrationColumns;
    if (version >= 3) {
        registrationColumns = JsonHelper::parseRegistrationColumns(
                                  value(TOURNAMENT_TEMPLATE_REGISTRATION_COLUMNS).toString());
    } else if (version < 3) {
        registrationColumns = JsonHelper::parsePlayersPageColumns(
                                  value(QStringLiteral("players_page/columns")).toString());
    }

    // Assemble the template and return it
    return TournamentTemplate::Settings {
        tournamentMode,
        scoreType,
        boogerScore,
        boogersPerRound,
        drawSettingsTemplate(),
        registrationColumns,
        value(TOURNAMENT_TEMPLATE_REGISTRATION_LIST_HEADER_HIDDEN, false).toBool(),
        value(TOURNAMENT_TEMPLATE_AUTO_SELECT_PAIRS, true).toBool(),
        value(TOURNAMENT_TEMPLATE_SELECT_BY_LAST_ROUND, true).toBool(),
        value(TOURNAMENT_TEMPLATE_NO_TABLE_NUMBERS, false).toBool(),
        value(TOURNAMENT_TEMPLATE_INCLUDE_OPPONENT_GOALS, true).toBool(),
        value(TOURNAMENT_TEMPLATE_GRAY_OUT_UNIMPORTANT_GOALS, true).toBool(),
        JsonHelper::parseOptionalPages(value(TOURNAMENT_TEMPLATE_OPTIONAL_PAGES).toString())
    };
}

// Database file name template
// ---------------------------

void Settings::saveDbFileNameTemplate(const QString &text)
{
    setValue(TOURNAMENT_TEMPLATE_DB_FILE_NAME_TEMPLATE, text);
}

QString Settings::dbFileNameTemplate() const
{
    return value(TOURNAMENT_TEMPLATE_DB_FILE_NAME_TEMPLATE,
                 DefaultValues::DB_FILE_NAME_TEMPLATE).toString();
}

// Markers template
// ----------------

void Settings::saveMarkersTemplate(const QVector<Markers::Marker> &markers)
{
    QJsonArray json;
    for (const Markers::Marker &marker : markers) {
        json.append(QJsonObject {
            { MARKERS_TEMPLATE_MARKERS_JSON_NAME,    marker.name },
            { MARKERS_TEMPLATE_MARKERS_JSON_COLOR,   QString::number(marker.color.rgb()) },
            { MARKERS_TEMPLATE_MARKERS_JSON_SORTING, static_cast<int>(marker.sorting) } });
    }

    setValue(MARKERS_TEMPLATE_VERSION, VERSION_TEMPLATE_MARKERS);
    setValue(MARKERS_TEMPLATE_MARKERS, JsonHelper::serializeJson(json));
}

bool Settings::parseMarkerData(int &id, const QJsonValue &data,
                               QVector<Markers::Marker> &markers) const
{
    Markers::Marker marker;
    const QJsonObject markerData = data.toObject();
    QJsonValue value;

    value = markerData.value(MARKERS_TEMPLATE_MARKERS_JSON_NAME);
    if (value.type() != QJsonValue::String) {
        return false;
    } else {
        marker.name = value.toString();
    }

    value = markerData.value(MARKERS_TEMPLATE_MARKERS_JSON_COLOR);
    if (value.type() != QJsonValue::String) {
        return false;
    } else {
        marker.color = QColor::fromRgb(value.toString().toUInt());
    }

    value = markerData.value(MARKERS_TEMPLATE_MARKERS_JSON_SORTING);
    if (value.type() != QJsonValue::Double) {
        return false;
    } else {
        marker.sorting = static_cast<Markers::MarkerSorting>(value.toInt());
    }

    if (marker.name.isEmpty()) {
        marker.id = 0;
    } else {
        marker.id = ++id;
    }
    markers.append(marker);

    return true;
}

void Settings::appendMarker(const QString &name, const QColor &color,
                            Markers::MarkerSorting sorting, int id,
                            QVector<Markers::Marker> &markers) const
{
    Markers::Marker marker;
    marker.name = name;
    marker.color = color;
    marker.sorting = sorting;
    marker.id = id;
    markers.append(marker);
}

QVector<Markers::Marker> Settings::defaultMarkersTemplate() const
{
    QVector<Markers::Marker> markers;
    markers.reserve(4);
    appendMarker(tr("Rot"),  QColor(255, 0,   0,   255),
                 Markers::MarkerSorting::NewBlock, 1, markers);
    appendMarker(tr("Blau"), QColor(0,   0,   255, 255),
                 Markers::MarkerSorting::NewBlock, 2, markers);
    appendMarker(QString(),  QColor(0,   0,   0,   255),
                 Markers::MarkerSorting::NewBlock, 0, markers);
    appendMarker(tr("Grün"), QColor(0,   128, 0,   255),
                 Markers::MarkerSorting::Merged,   3, markers);
    return markers;
}

QVector<Markers::Marker> Settings::markersTemplate() const
{
    const int version = value(MARKERS_TEMPLATE_VERSION, 0).toInt();

    QJsonArray data;
    if (version == 2) {
        data = JsonHelper::jsonArrayFromString(value(MARKERS_TEMPLATE_MARKERS,
                                                     QString()).toString());
    } else if (version == 1) {
        data = JsonHelper::jsonObjectFromString(QString::fromUtf8(
            value(QStringLiteral("markers_template/data"), QByteArray()).toByteArray())).value(
                QStringLiteral("d")).toArray();
    } else {
        return defaultMarkersTemplate();
    }

    if (data.isEmpty()) {
        return defaultMarkersTemplate();
    }

    QVector<Markers::Marker> markers;

    bool templateParsed = true;
    int id = 0;

    for (const QJsonValue &value : std::as_const(data)) {
        if (! parseMarkerData(id, value, markers)) {
            templateParsed = false;
            break;
        }
    }

    // Be sure to have an "(unmarked)" marker with ID 0 and no name
    if (templateParsed) {
        templateParsed = false;
        for (const Markers::Marker &marker : std::as_const(markers)) {
            if (marker.id == 0 && marker.name.isEmpty()) {
                templateParsed = true;
                break;
            }
        }
    }

    if (! templateParsed) {
        return defaultMarkersTemplate();
    }

    return markers;
}

void Settings::saveMarker(Markers::Type type, int id)
{
    setValue(MARKERS_TEMPLATE_MARKER_KEY.value(type), id);
}

int Settings::marker(Markers::Type type) const
{
    return value(MARKERS_TEMPLATE_MARKER_KEY.value(type),
                 MARKERS_TEMPLATE_MARKER_DEFAULT_VALUE.value(type)).toInt();
}

// Singles Management
// ------------------

void Settings::saveSinglesManagementTemplate(const Markers::SinglesManagement &data)
{
    setValue(SINGLES_MANAGEMENT_TEMPLATE_VERSION, Markers::SINGLES_MANAGEMENT_VERSION);
    setValue(SINGLES_MANAGEMENT_TEMPLATE_DATA,    JsonHelper::singlesManagementToString(data));
}

Markers::SinglesManagement Settings::defaultSinglesManagementTemplate() const
{
    Markers::SinglesManagement data = Markers::DEFAULT_SINGLES_MANAGEMENT;
    data.conditionString = namesSeparator().simplified();
    return data;
}

Markers::SinglesManagement Settings::singlesManagementTemplate() const
{
    if (value(SINGLES_MANAGEMENT_TEMPLATE_VERSION, 0).toInt()
        != Markers::SINGLES_MANAGEMENT_VERSION) {

        return defaultSinglesManagementTemplate();
    }

    return JsonHelper::parseSinglesManagement(value(SINGLES_MANAGEMENT_TEMPLATE_DATA).toString());
}

// Strings
// -------

void Settings::saveBoogerSymbol(const QString &text)
{
    setValue(STRINGS_BOOGER_SYMBOL, text);
    m_boogerSymbol = text;
}

const QString &Settings::boogerSymbol() const
{
    return m_boogerSymbol;
}

void Settings::saveNamesSeparator(const QString &text)
{
    setValue(STRINGS_NAMES_SEPARATOR, text);
    m_namesSeparator = text;
}

void Settings::setTransientNamesSeparator(const QString &text)
{
    if (m_namesSeparator != text) {
        m_originalNamesSeparator = m_namesSeparator;
        m_namesSeparator = text;
    } else {
        m_originalNamesSeparator.clear();
    }
}

bool Settings::resetNamesSeparator()
{
    if (! m_originalNamesSeparator.isEmpty()) {
        m_namesSeparator = m_originalNamesSeparator;
        m_originalNamesSeparator.clear();
        return true;
    } else {
        return false;
    }
}

const QString &Settings::namesSeparator() const
{
    return m_namesSeparator;
}

// Network
// -------

void Settings::saveServerPort(int port)
{
    setValue(NETWORK_SERVER_PORT, port);
}

int Settings::serverPort() const
{
    const int port = value(NETWORK_SERVER_PORT, DefaultValues::SERVER_PORT).toInt();
    return (port >= 1024 && port <= 65535) ? port : DefaultValues::SERVER_PORT;
}

void Settings::saveConnectTimeout(int timeout)
{
    setValue(NETWORK_CONNECT_TIMEOUT, timeout);
}

int Settings::connectTimeout() const
{
    return value(NETWORK_CONNECT_TIMEOUT, DefaultValues::CONNECT_TIMEOUT).toInt();
}

void Settings::saveRequestTimeout(int timeout)
{
    setValue(NETWORK_REQUEST_TIMEOUT, timeout);
}

int Settings::requestTimeout() const
{
    return value(NETWORK_REQUEST_TIMEOUT, DefaultValues::REQUEST_TIMEOUT).toInt();
}

void Settings::saveDiscoverPort(int port)
{
    setValue(NETWORK_DISCOVER_PORT, port);
}

int Settings::discoverPort() const
{
    const int port = value(NETWORK_DISCOVER_PORT, DefaultValues::DISCOVER_SERVER_PORT).toInt();
    return (port >= 1024 && port <= 65535) ? port : DefaultValues::DISCOVER_SERVER_PORT;
}

void Settings::saveDiscoverTimeout(int timeout)
{
    setValue(NETWORK_DISCOVER_TIMEOUT, timeout);
}

int Settings::discoverTimeout() const
{
    return value(NETWORK_DISCOVER_TIMEOUT, DefaultValues::DISCOVER_TIMEOUT).toInt();
}

void Settings::saveDiscoverAttempts(int attempts)
{
    setValue(NETWORK_DISCOVER_ATTEMPTS, attempts);
}

int Settings::discoverAttempts() const
{
    return value(NETWORK_DISCOVER_ATTEMPTS, DefaultValues::DISCOVER_ATTEMPTS).toInt();
}

void Settings::savePreferIpv6(bool state)
{
    setValue(NETWORK_PREFER_IPV6, state);
}

bool Settings::preferIpv6() const
{
    return value(NETWORK_PREFER_IPV6, false).toBool();
}

// Draw settings

void Settings::saveDrawSettingsTemplate(const Draw::Settings &data)
{
    setValue(DRAW_SETTINGS_TEMPLATE_VERSION, Draw::SETTINGS_VERSION);
    setValue(DRAW_SETTINGS_TEMPLATE_DATA,    JsonHelper::drawSettingsToString(data));
}

Draw::Settings Settings::drawSettingsTemplate() const
{
    if (value(DRAW_SETTINGS_TEMPLATE_VERSION, 0).toInt() != Draw::SETTINGS_VERSION) {
        return Draw::DEFAULT_SETTINGS;
    }

    return JsonHelper::parseDrawSettings(value(DRAW_SETTINGS_TEMPLATE_DATA).toString());
}
