/* Copyright (c) 2019-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Local includes
#include "TournamentSettings.h"
#include "JsonHelper.h"

// Qt includes
#include <QDebug>
#include <QJsonObject>
#include <QJsonArray>

// C++ includes
#include <utility>

namespace
{

// The revision we use
constexpr int TOURNAMENT_SETTINGS_VERSION = 2;

// JSON Keys for all settings

const QString MUCKTURNIER_VERSION = QStringLiteral("muckturnier version");
const QString TOURNAMENT_MODE     = QStringLiteral("tournament mode");
const QString BOOGER_SCORE        = QStringLiteral("booger score");
const QString BOOGERS_PER_ROUND   = QStringLiteral("boogers per round");

const QString TOURNAMENT_OPEN = QStringLiteral("tournament open");

const QString SCORE_TYPE                 = QStringLiteral("score type");
const QString NO_TABLE_NUMBERS           = QStringLiteral("no table numbers");
const QString AUTOSELECT_PAIRS           = QStringLiteral("autoselect pairs");
const QString SELECT_BY_LAST_ROUND       = QStringLiteral("select by last round");
const QString INCLUDE_OPPONENT_GOALS     = QStringLiteral("include opponent goals");
const QString GRAY_OUT_UNIMPORTANT_GOALS = QStringLiteral("gray out unimportant goals");

const QString DEFAULT_MARKER             = QStringLiteral("default marker");
const QString DRAWN_MARKER               = QStringLiteral("drawn marker");
const QString ABSENT_MARKER              = QStringLiteral("absent marker");
const QString SINGLES_MANAGEMENT_VERSION = QStringLiteral("singles management version");
const QString SINGLES_MANAGEMENT         = QStringLiteral("singles management");

const QString REGISTRATION_COLUMNS            = QStringLiteral("registration columns");
const QString REGISTRATION_LIST_HEADER_HIDDEN = QStringLiteral("registration list header hidden");

const QString DRAW_SETTINGS_VERSION = QStringLiteral("draw settings version");
const QString DRAW_SETTINGS         = QStringLiteral("draw settings");

const QString OPTIONAL_PAGES = QStringLiteral("optional pages");

}

TournamentSettings::TournamentSettings(QObject *parent) : QObject(parent)
{
}

int TournamentSettings::version() const
{
    return TOURNAMENT_SETTINGS_VERSION;
}

bool TournamentSettings::loadingDb() const
{
    return m_loadingDb;
}

void TournamentSettings::setLoadingDb(bool state)
{
    m_loadingDb = state;
}

QString TournamentSettings::updateSettingsFromV8(const QString &data) const
{
    // Move the "players page columns" hash to the "registration columns" array

    const QString playersPageColumns = QStringLiteral("players page columns");
    QJsonObject settings = JsonHelper::jsonObjectFromString(data);
    const auto columns = JsonHelper::parsePlayersPageColumns(
                             settings.value(playersPageColumns).toObject());

    settings.insert(QStringLiteral("registration columns"),
                    JsonHelper::registrationColumnsToJson(columns));
    settings.remove(playersPageColumns);

    return JsonHelper::serializeJson(settings);
}

const QVector<OptionalPages::Page> &TournamentSettings::optionalPages() const
{
    return m_optionalPages;
}

void TournamentSettings::setOptionalPageShown(OptionalPages::Page page, bool state)
{
    if (m_loadingDb) {
        return;
    }

    if (state) {
        m_optionalPages.append(page);
    } else {
        m_optionalPages.removeOne(page);
    }
}

void TournamentSettings::setDbSettingsVersion(int version)
{
    m_dbSettingsVersion = version;
}

int TournamentSettings::dbSettingsVersion() const
{
    return m_dbSettingsVersion;
}

bool TournamentSettings::load(const QString &data)
{
    const QJsonObject json = JsonHelper::jsonObjectFromString(data);

    // There are some essential values that have to be present and valid. We parse them here:

    const QString tournamentMode = json.value(TOURNAMENT_MODE).toString();
    bool tournamentModeValid = false;
    QHashIterator<Tournament::Mode, QString> it(Tournament::MODE_MAP);
    while (it.hasNext()) {
        it.next();
        if (it.value() == tournamentMode) {
            m_tournamentMode = it.key();
            tournamentModeValid = true;
            break;
        }
    }
    if (! tournamentModeValid) {
        return false;
    }

    QJsonValue value;

    value = json.value(BOOGER_SCORE);
    if (value.type() == QJsonValue::Double) {
        m_boogerScore = value.toInt();
    } else {
        return false;
    }

    value = json.value(BOOGERS_PER_ROUND);
    if (value.type() == QJsonValue::Double) {
        m_boogersPerRound = value.toInt();
    } else {
        return false;
    }

    // For all other values, we can fall back to a default setting

    const QJsonValue tournamentOpenValue(json.value(TOURNAMENT_OPEN));
    m_tournamentOpen = tournamentOpenValue.type() == QJsonValue::Bool
                           ? tournamentOpenValue.toBool() : true;
    m_originalTournamentOpen = m_tournamentOpen;

    const QString muckturnierVersion = json.value(MUCKTURNIER_VERSION).toString();
    m_muckturnierVersion = muckturnierVersion.isEmpty() ? tr("unbekannt") : muckturnierVersion;

    m_noTableNumbers = json.value(NO_TABLE_NUMBERS).toBool();

    m_autoSelectPairs = json.value(AUTOSELECT_PAIRS).toBool();
    m_selectByLastRound = json.value(SELECT_BY_LAST_ROUND).toBool();

    m_includeOpponentGoals = json.value(INCLUDE_OPPONENT_GOALS).toBool();

    m_grayOutUnimportantGoals = json.value(GRAY_OUT_UNIMPORTANT_GOALS).toBool();

    m_marker[Markers::Default] = json.value(DEFAULT_MARKER).toInt();

    const auto drawnMarkerValue = json.value(DRAWN_MARKER);
    m_marker[Markers::Drawn] = drawnMarkerValue.isUndefined() ? -1 : drawnMarkerValue.toInt();

    m_marker[Markers::Absent] = json.value(ABSENT_MARKER).toInt();

    const bool singlesManagementParsed = json.value(SINGLES_MANAGEMENT_VERSION).toInt()
                                         == Markers::SINGLES_MANAGEMENT_VERSION;
    m_singlesManagement = singlesManagementParsed
        ? JsonHelper::parseSinglesManagement(json.value(SINGLES_MANAGEMENT).toObject())
        : Markers::DEFAULT_SINGLES_MANAGEMENT;

    m_registrationColumns = JsonHelper::parseRegistrationColumns(
                                json.value(REGISTRATION_COLUMNS).toArray());

    m_registrationListHeaderHidden = json.value(REGISTRATION_LIST_HEADER_HIDDEN).toBool();

    m_scoreType = Tournament::SCORE_TYPE_MAP.key(json.value(SCORE_TYPE).toString(),
                                                 Tournament::ScoreType::Horizontal);

    const bool drawSettingsParsed = json.value(DRAW_SETTINGS_VERSION).toInt()
                                        == Draw::SETTINGS_VERSION;
    m_drawSettings = drawSettingsParsed
        ? JsonHelper::parseDrawSettings(json.value(DRAW_SETTINGS).toObject())
        : Draw::DEFAULT_SETTINGS;

    m_optionalPages = JsonHelper::parseOptionalPages(json.value(OPTIONAL_PAGES).toArray());

    m_unhandledValues.clear();
    if (m_dbSettingsVersion > TOURNAMENT_SETTINGS_VERSION) {
        // The settings were created by a newer version of Muckturnier. Some keys are not handled by
        // this version yet. To keep them when closing and saving the currrent settings, we cache
        // them here so that they can be re-added later.

        auto unhandledKeys = json.keys();
        unhandledKeys.removeOne(MUCKTURNIER_VERSION);
        unhandledKeys.removeOne(TOURNAMENT_MODE);
        unhandledKeys.removeOne(TOURNAMENT_OPEN);
        unhandledKeys.removeOne(SCORE_TYPE);
        unhandledKeys.removeOne(BOOGER_SCORE);
        unhandledKeys.removeOne(BOOGERS_PER_ROUND);
        unhandledKeys.removeOne(NO_TABLE_NUMBERS);
        unhandledKeys.removeOne(AUTOSELECT_PAIRS);
        unhandledKeys.removeOne(SELECT_BY_LAST_ROUND);
        unhandledKeys.removeOne(INCLUDE_OPPONENT_GOALS);
        unhandledKeys.removeOne(GRAY_OUT_UNIMPORTANT_GOALS);
        unhandledKeys.removeOne(DEFAULT_MARKER);
        unhandledKeys.removeOne(DRAWN_MARKER);
        unhandledKeys.removeOne(ABSENT_MARKER);
        if (singlesManagementParsed) {
            unhandledKeys.removeOne(SINGLES_MANAGEMENT_VERSION);
            unhandledKeys.removeOne(SINGLES_MANAGEMENT);
        }
        unhandledKeys.removeOne(REGISTRATION_COLUMNS);
        unhandledKeys.removeOne(REGISTRATION_LIST_HEADER_HIDDEN);
        if (drawSettingsParsed) {
            unhandledKeys.removeOne(DRAW_SETTINGS_VERSION);
            unhandledKeys.removeOne(DRAW_SETTINGS);
        }
        unhandledKeys.removeOne(OPTIONAL_PAGES);

        for (const QString &key : std::as_const(unhandledKeys)) {
            m_unhandledValues.insert(key, json.value(key));
        }
    }

    return true;
}

QString TournamentSettings::toString() const
{
    QJsonObject settings {
        { MUCKTURNIER_VERSION,             m_muckturnierVersion },
        { TOURNAMENT_MODE,                 Tournament::MODE_MAP.value(m_tournamentMode) },
        { TOURNAMENT_OPEN,                 m_tournamentOpen },
        { SCORE_TYPE,                      Tournament::SCORE_TYPE_MAP.value(m_scoreType) },
        { BOOGER_SCORE,                    m_boogerScore },
        { BOOGERS_PER_ROUND,               m_boogersPerRound },
        { NO_TABLE_NUMBERS,                m_noTableNumbers },
        { AUTOSELECT_PAIRS,                m_autoSelectPairs },
        { SELECT_BY_LAST_ROUND,            m_selectByLastRound },
        { INCLUDE_OPPONENT_GOALS,          m_includeOpponentGoals },
        { GRAY_OUT_UNIMPORTANT_GOALS,      m_grayOutUnimportantGoals },
        { DEFAULT_MARKER,                  m_marker.value(Markers::Default) },
        { DRAWN_MARKER,                    m_marker.value(Markers::Drawn) },
        { ABSENT_MARKER,                   m_marker.value(Markers::Absent) },
        { SINGLES_MANAGEMENT_VERSION,      Markers::SINGLES_MANAGEMENT_VERSION },
        { SINGLES_MANAGEMENT,              JsonHelper::singlesManagementToJson(
                                               m_singlesManagement) },
        { REGISTRATION_COLUMNS,            JsonHelper::registrationColumnsToJson(
                                               m_registrationColumns) },
        { REGISTRATION_LIST_HEADER_HIDDEN, m_registrationListHeaderHidden },
        { DRAW_SETTINGS_VERSION,           Draw::SETTINGS_VERSION },
        { DRAW_SETTINGS,                   JsonHelper::drawSettingsToJson(m_drawSettings) },
        { OPTIONAL_PAGES,                  JsonHelper::optionalPagesToJson(m_optionalPages) }
    };

    // If we have unhandled values, add them as well so that they aren't lost
    if (! m_unhandledValues.isEmpty()) {
        QHashIterator<QString, QJsonValue> it(m_unhandledValues);
        while (it.hasNext()) {
            it.next();
            settings.insert(it.key(), it.value());
        }
    }

    return JsonHelper::serializeJson(settings);
}

void TournamentSettings::setMuckturnierVersion(const QString &text)
{
    m_muckturnierVersion = text;
}

const QString &TournamentSettings::muckturnierVersion() const
{
    return m_muckturnierVersion;
}

void TournamentSettings::setTournamentMode(Tournament::Mode tournamentMode)
{
    m_tournamentMode = tournamentMode;
}

Tournament::Mode TournamentSettings::tournamentMode() const
{
    return m_tournamentMode;
}

bool TournamentSettings::isPairMode() const
{
    return m_tournamentMode == Tournament::Mode::FixedPairs;
}

bool TournamentSettings::isSinglePlayerMode() const
{
    return m_tournamentMode == Tournament::Mode::SinglePlayers;
}

void TournamentSettings::setTournamentOpen(bool state)
{
    m_tournamentOpen = state;
}

bool TournamentSettings::tournamentOpen() const
{
    return m_tournamentOpen;
}

bool TournamentSettings::tournamentOpenChanged() const
{
    return m_tournamentOpen != m_originalTournamentOpen;
}

void TournamentSettings::setBoogerScore(int score)
{
    m_boogerScore = score;
}

const int &TournamentSettings::boogerScore() const
{
    return m_boogerScore;
}

void TournamentSettings::setBoogersPerRound(int count)
{
    m_boogersPerRound = count;
}

int TournamentSettings::boogersPerRound() const
{
    return m_boogersPerRound;
}

void TournamentSettings::setNoTableNumbers(bool state)
{
    m_noTableNumbers = state;
}

bool TournamentSettings::noTableNumbers() const
{
    return m_noTableNumbers;
}

void TournamentSettings::setAutoSelectPairs(bool state)
{
    m_autoSelectPairs = state;
}

bool TournamentSettings::autoSelectPairs() const
{
    return m_autoSelectPairs;
}

void TournamentSettings::setSelectByLastRound(bool state)
{
    m_selectByLastRound = state;
}

bool TournamentSettings::selectByLastRound() const
{
    return m_selectByLastRound;
}

void TournamentSettings::setIncludeOpponentGoals(bool state)
{
    m_includeOpponentGoals = state;
}

bool TournamentSettings::includeOpponentGoals() const
{
    return m_includeOpponentGoals;
}

void TournamentSettings::setGrayOutUnimportantGoals(bool state)
{
    m_grayOutUnimportantGoals = state;
}

bool TournamentSettings::grayOutUnimportantGoals() const
{
    return m_grayOutUnimportantGoals;
}

void TournamentSettings::setMarker(Markers::Type type, int id)
{
    m_marker[type] = id;
}

int TournamentSettings::marker(Markers::Type type) const
{
    return m_marker.value(type);
}

void TournamentSettings::setSinglesManagement(const Markers::SinglesManagement &data)
{
    m_singlesManagement = data;
}

const Markers::SinglesManagement &TournamentSettings::singlesManagement() const
{
    return m_singlesManagement;
}

void TournamentSettings::setRegistrationColumns(const QVector<RegistrationColumns::Column> &data)
{
    m_registrationColumns = data;
}

const QVector<RegistrationColumns::Column> &TournamentSettings::registrationColumns() const
{
    return m_registrationColumns;
}

void TournamentSettings::setRegistrationListHeaderHidden(bool state)
{
    m_registrationListHeaderHidden = state;
}

bool TournamentSettings::registrationListHeaderHidden() const
{
    return m_registrationListHeaderHidden;
}

void TournamentSettings::setScoreType(Tournament::ScoreType type)
{
    m_scoreType = type;
}

Tournament::ScoreType TournamentSettings::scoreType() const
{
    return m_scoreType;
}

void TournamentSettings::setDrawSettings(const Draw::Settings &settings)
{
    m_drawSettings = settings;
}

const Draw::Settings &TournamentSettings::drawSettings() const
{
    return m_drawSettings;
}
